package at.ac.tuwien.asgaard.delta.gui.html;

import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import javax.swing.undo.*;
import java.util.Enumeration;

final class ExtendedHTMLDocument extends HTMLDocument {
	private ExtendedHTMLEditorKit kit;

	/**
	 * Constructs an HTML document with the default content
	 * storage implementation and the given style/attribute
	 * storage mechanism.
	 *
	 * @param styles  the styles
	 */
	public ExtendedHTMLDocument(StyleSheet styles, ExtendedHTMLEditorKit kit) {
		super(styles);
		this.kit = kit;
	}

	public String getTitle() {
		return (String)getProperty(Document.TitleProperty);
	}

	void setLinkElementAttributes(int startOffset, int length, AttributeSet aTagAttrSet, int newDELTALinkID, boolean useLinkSeparator) {
		try {
			writeLock();

			DefaultDocumentEvent changes = new DefaultDocumentEvent(startOffset, length, DocumentEvent.EventType.CHANGE);
			buffer.change(startOffset, length, changes);

			int thisEnd;
			int endOffset = (startOffset + length);
			for (int pos = startOffset; pos < endOffset; pos = thisEnd) {
				Element element = getCharacterElement(pos);
				thisEnd = element.getEndOffset();
				if (pos == thisEnd)
					break;
				MutableAttributeSet attr = (MutableAttributeSet) element.getAttributes();
				MutableAttributeSet aTagAttrSetCopy = (MutableAttributeSet) aTagAttrSet.copyAttributes();

				if (useLinkSeparator)
					if ((pos == startOffset) && (thisEnd == endOffset))
						aTagAttrSetCopy.addAttribute(HTML.Attribute.CLASS, HTMLView.LINK_SEPARATOR_CLASS_ATTR);
					else if (pos == startOffset)
							aTagAttrSetCopy.addAttribute(HTML.Attribute.CLASS, HTMLView.LINK_SEPARATOR_CLASS_ATTR_START);
					else if (thisEnd == endOffset)
						aTagAttrSetCopy.addAttribute(HTML.Attribute.CLASS, HTMLView.LINK_SEPARATOR_CLASS_ATTR_END);

				SimpleAttributeSet aTagSet = new SimpleAttributeSet();
				aTagSet.addAttribute(HTML.Tag.A, aTagAttrSetCopy);

				changes.addEdit(new AttributeUndoableEdit(element, aTagSet, false));
				attr.addAttributes(aTagSet);
				kit.deltaLinkMap.add(newDELTALinkID, element);
			}
			changes.end();
			fireChangedUpdate(changes);
		// fireUndoableEditUpdate(new UndoableEditEvent(this, changes));
		} finally {
			writeUnlock();
		}
	}

	void applyCharacterChanges(int startOffset, int endOffset, DefaultDocumentEvent changes) {
		int length = endOffset - startOffset;
		if (changes == null)
			changes = new DefaultDocumentEvent(startOffset, length, DocumentEvent.EventType.CHANGE);
		buffer.change(startOffset, length, changes);
		changes.end();
	    fireChangedUpdate(changes);
	//    fireUndoableEditUpdate(new UndoableEditEvent(this, changes));
	}

    /**
     * Removes attributes for some part of the document.
     * A write lock is held by this operation while changes
     * are being made, and a DocumentEvent is sent to the listeners
     * after the change has been successfully completed.
     * <p>
     * This method is thread safe, although most Swing methods
     * are not. Please see
     * <A HREF="http://java.sun.com/products/jfc/swingdoc-archive/threads.html">Threads
     * and Swing</A> for more information.
     *
     * @param offset the offset in the document >= 0
     * @param length the length >= 0
     * @param s the attributes
     */
	void removeCharacterAttributes(int offset, int length, AttributeSet s) {
		try {
			writeLock();
			DefaultDocumentEvent changes =
					new DefaultDocumentEvent(offset, length, DocumentEvent.EventType.CHANGE);
			buffer.change(offset, length, changes);
			int index = offset, end = offset + length;

			while (index < end) {
				Element e = getCharacterElement(index);
				MutableAttributeSet a
						= (MutableAttributeSet) e.getAttributes();
				changes.addEdit(new AttributeRemoveEdit(e, s));
				for (Enumeration f = s.getAttributeNames(); f.hasMoreElements();)
					a.removeAttribute(f.nextElement());
				index = Math.max(index + 1, endOffset(e));
			}

			changes.end();
			fireChangedUpdate(changes);
		// fireUndoableEditUpdate(new UndoableEditEvent(this, changes));
		} finally {
			writeUnlock();
		}
	}

    /**
     * Removes attributes for a paragraph.
     * <p>
     * This method is thread safe, although most Swing methods
     * are not. Please see
     * <A HREF="http://java.sun.com/products/jfc/swingdoc-archive/threads.html">Threads
     * and Swing</A> for more information.
     *
     * @param offset the offset into the paragraph >= 0
     * @param length the number of characters affected >= 0
     * @param s the attributes
     */
/* void removeParagraphAttributes(int offset, int length, AttributeSet s) {
		try {
			writeLock();
			DefaultDocumentEvent changes =
					new DefaultDocumentEvent(offset, length, DocumentEvent.EventType.CHANGE);
			int index = offset, end = offset + length;

			while (index < end) {
				Element e = getParagraphElement(index);
				MutableAttributeSet a
						= (MutableAttributeSet) e.getAttributes();
				changes.addEdit(new AttributeRemoveEdit(e, s));
				for (Enumeration f = s.getAttributeNames(); f.hasMoreElements();)
					a.removeAttribute(f.nextElement());
				index = Math.max(index + 1, endOffset(e));
			}

			// Cannot call updateBidi(changes);
			changes.end();
			fireChangedUpdate(changes);
			fireUndoableEditUpdate(new UndoableEditEvent(this, changes));
		} finally {
			writeUnlock();
		}
	}*/

	static final class AttributeRemoveEdit extends AbstractUndoableEdit {
		// Attributes that should be removed with their old values
		private final MutableAttributeSet old;
		private final Element element;

		AttributeRemoveEdit(Element e, AttributeSet toRemove) {
			element = e;
			old = new SimpleAttributeSet();
			AttributeSet a = e.getAttributes();
			for (Enumeration f = toRemove.getAttributeNames(); f.hasMoreElements();) {
				Object key = f.nextElement();
				if (a.isDefined(key))
					old.addAttribute(key, a.getAttribute(key));
			}
		}

		public void redo() throws CannotRedoException {
			super.redo();
			MutableAttributeSet a
					= (MutableAttributeSet) element.getAttributes();
			for (Enumeration f = old.getAttributeNames(); f.hasMoreElements();) {
				a.removeAttribute(f.nextElement());
			}
		}

		public void undo() throws CannotUndoException {
			super.undo();
			MutableAttributeSet a
					= (MutableAttributeSet) element.getAttributes();
			a.addAttributes(old);
		}
	}

	protected int endOffset(Element e) {
		return Math.min(getLength(), e.getEndOffset());
	}


}
