package at.ac.tuwien.asgaard.delta.gui.html;

import at.ac.tuwien.asgaard.delta.util.*;
import at.ac.tuwien.asgaard.delta.gui.LinkIDConstants;
import at.ac.tuwien.asgaard.util.SortedListMap;
import at.ac.tuwien.asgaard.util.Log;
import at.ac.tuwien.asgaard.swing.SwingUtils;

import javax.swing.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelListener;
import java.awt.event.MouseWheelEvent;
import java.awt.*;
import java.io.*;
import java.util.*;

final class ExtendedHTMLEditorKit extends HTMLEditorKit implements ViewFactory {
	int maxLinkID;
	final SortedListMap<Integer,Element> deltaLinkMap;
	boolean showLinks = true;

	private boolean addDELTALinksOnLoad, removeDELTALinksOnSave;
	private final Preferences properties;
	private final HTMLView htmlView;
	private boolean loading;
	private final ExtendedLinkController linkHandler;
	static StyleSheet defaultStyleSheet;

	private Color prefLinkHTMLColor, prefLinkDELTAColor, prefLinkOverlappingColor;

	ExtendedHTMLEditorKit(Preferences properties, HTMLView htmlView) {
		this.properties = properties;
		this.htmlView = htmlView;
		setProperties();
		deltaLinkMap = new SortedListMap<Integer, Element>(new Comparator<Element>() {
			public int compare(Element element1, Element element2) {
				int startDiff = element1.getStartOffset() - element2.getStartOffset();
				return (startDiff == 0 ? (element1.getEndOffset() - element2.getEndOffset()) : startDiff);
			}
		});
		maxLinkID = LinkIDConstants.FIRST_ID - 1;
		linkHandler = new ExtendedLinkController(htmlView, this);
		// imageMap = new HashMap();
	}

	public ExtendedHTMLDocument createDocument() {
		StyleSheet styles = getStyleSheet();
		StyleSheet ss = new StyleSheet();
		ss.addStyleSheet(styles);
		ExtendedHTMLDocument doc = new ExtendedHTMLDocument(ss, this);
		doc.setParser(getParser());
		doc.setTokenThreshold(100);
		return doc;
	}

	void setProperties() {
		addDELTALinksOnLoad = removeDELTALinksOnSave = properties.getBoolean(Preferences.KEY_HTML_REMOVE_DELTA_LINKS_ON_SAVE);
		prefLinkHTMLColor = properties.getColor(Preferences.KEY_COLOR_HTML_LINK_FG);
		prefLinkDELTAColor = properties.getColor(Preferences.KEY_COLOR_LINK_FG);
		prefLinkOverlappingColor = properties.getColor(Preferences.KEY_COLOR_LINK_HTML_OVERLAPPING_FG);
	}

	public StyleSheet getStyleSheet() {
		if (defaultStyleSheet == null) {
			defaultStyleSheet = new StyleSheet() {
				protected int getCompressionThreshold() {
					return 0; // to have unique attributesets!
				}
			};
			try {
				InputStream is = HTMLEditorKit.class.getResourceAsStream(DEFAULT_CSS);
				Reader r = new BufferedReader(new InputStreamReader(is));
				defaultStyleSheet.loadRules(r, null);
				r.close();
			} catch (Throwable err) {
				// on error we simply have no styles... the html will look mighty wrong but still function.
				Log.error("Couldn't load 'default.css'!");
				err.printStackTrace();
			}
		//	defaultStyleSheet.addRule("a { color: green }");
		}
		return defaultStyleSheet;
	}

	public ViewFactory getViewFactory() {
		return this;
	}

/* public final ExtendedImageView getImageView(Element elem) {
		return (ExtendedImageView)imageMap.get(elem);
	}*/

	public void install(JEditorPane c) {
		// c.setContentType("text/html");
		c.addMouseListener(linkHandler);
		c.addMouseMotionListener(linkHandler);
	}
/* public void deinstall(JEditorPane c) {
		c.removeMouseListener(linkHandler);
		c.removeMouseMotionListener(linkHandler);
	}*/

	Point getLastClickedLocation() {
		return linkHandler.lastClickedLocation;
	}

	private final class ExtendedLinkController extends LinkController implements MouseWheelListener {
		private final JEditorPane editor;
		private final HTMLEditorKit kit;
		private int oldPos = -1;
		Point lastClickedLocation;

		public ExtendedLinkController(JEditorPane editor, HTMLEditorKit kit) {
			this.editor = editor;
			this.kit = kit;
		}

		public void mouseClicked(MouseEvent evt) {
			if (editor.isEnabled())
				try {
					if (SwingUtilities.isLeftMouseButton(evt)) {
						lastClickedLocation = evt.getPoint();
						if (!evt.isControlDown() && !evt.isAltDown())   // deactivates links on <Ctrl>
							super.mouseClicked(evt);
					}
				} catch (Throwable t) {  // supress unpredictable swing/html errors
					System.err.println("should not happen: "+t);
					t.printStackTrace();
				}
		}

		public void mouseMoved(MouseEvent evt) {
			if (editor.isEnabled())
				try {
					if (showLinks && !evt.isAltDown())  // do not deactivate mouse-over-links on <Ctrl>
						super.mouseMoved(evt);
					// not really nice, but the easiest way to change the cursor to TEXT_CURSOR over text
					int pos = editor.viewToModel(new Point(evt.getX(), evt.getY()));
					if (pos != oldPos) {
						boolean overText = false;
						Element elem = ((HTMLDocument) editor.getDocument()).getCharacterElement(pos);
						if (elem != null)
							overText = ((elem.getAttributes().getAttribute(StyleConstants.NameAttribute) == HTML.Tag.CONTENT) &&
								(editor.getDocument().getText(elem.getStartOffset(), elem.getEndOffset() - elem.getStartOffset()).trim().length() > 0));

						if (overText) {
							if (editor.getCursor() != kit.getLinkCursor())
								editor.setCursor(Cursor.getPredefinedCursor(Cursor.TEXT_CURSOR));
						} else if (editor.getCursor() != kit.getLinkCursor())
							editor.setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
						oldPos = pos;
					}
				} catch (Throwable t) {  // supress unpredictable swing/html errors
					System.err.println("should not happen: " + t);
					t.printStackTrace();
				}
		}

		public void mouseWheelMoved(MouseWheelEvent evt) {
			mouseMoved(new MouseEvent(htmlView, evt.getID(), evt.getWhen(), evt.getModifiers(),
									  evt.getX(), evt.getY(), evt.getClickCount(), evt.isPopupTrigger(), evt.getButton()));
		}
	}

	Color bgColor;

	public View create(Element elem) {
		Object o = elem.getAttributes().getAttribute(StyleConstants.NameAttribute);
		if (o instanceof HTML.Tag) {
			HTML.Tag kind = (HTML.Tag) o;
			if (kind == HTML.Tag.CONTENT) {
				AttributeSet attrs = (AttributeSet) elem.getAttributes().getAttribute(HTML.Tag.A);
				if (attrs != null) {
					Set deltaLinkIDs = HTMLView.getDELTALinkIDs(attrs);
					if (!deltaLinkIDs.isEmpty()) {
						if (loading)  // to prevent doing this every time this method is called
							for (Iterator iterator = deltaLinkIDs.iterator(); iterator.hasNext();) {
								Integer integer = (Integer) iterator.next();
								deltaLinkMap.add(integer.intValue(), elem);
							}
				//		return new LinkInlineView(elem, true, deltaLinkIDs.size() > 1);
					}
				//	return new LinkInlineView(elem, false, false);
				}
			} /* else
			if (kind == HTML.Tag.IMG) {
				View view = new ExtendedImageView(elem);
				imageMap.put(elem, view);
				return view;
			}*/
		}
//		if (elem.getAttributes().getAttribute(HTML.Tag.A) != null)
//			return new LinkInlineView(elem);
		return super.getViewFactory().create(elem);
	}

	public void read(Reader in, Document doc, int pos)
			throws IOException, BadLocationException {
		if (bgColor != Color.WHITE)
			htmlView.getParent().getParent().addMouseWheelListener(linkHandler);

		maxLinkID = LinkIDConstants.FIRST_ID - 1;
		deltaLinkMap.clear();
		loading = true;
		bgColor = Color.WHITE;
		htmlView.setBackground(Color.WHITE);

		HTMLDocument hdoc = (HTMLDocument) doc;
		ParserCallback receiver = hdoc.new HTMLReader(pos, 0, 0, null) {
			public void handleStartTag(HTML.Tag tag, MutableAttributeSet attrSet, int pos) {
				if (tag == HTML.Tag.A) {
					String idStr = (String) attrSet.getAttribute(HTML.Attribute.ID);
					try {
						SortedSet idSet = HTMLView.parseIDList(idStr);
						for (Iterator iterator = idSet.iterator(); iterator.hasNext();) {
							Integer id = (Integer) iterator.next();
							maxLinkID = Math.max(maxLinkID, id.intValue());
						}
						if (addDELTALinksOnLoad) {
							attrSet.addAttribute(HTML.Attribute.HREF, idStr);
							if (idSet.size() == 1)
								attrSet.addAttribute(HTML.Attribute.COLOR, htmlView.prefLinkTextColor);
							else
								attrSet.addAttribute(HTML.Attribute.COLOR, htmlView.prefLinkSetTextColor);
						}
					} catch (NumberFormatException exc) {
					//	attrSet.addAttribute(HTML.Attribute.COLOR, htmlView.prefLinkTextColor);
						htmlView.checkURL((String) attrSet.getAttribute(HTML.Attribute.HREF));
					}
				} else if (tag == HTML.Tag.BODY) {
					String colorStr = (String) attrSet.getAttribute(HTML.Attribute.BGCOLOR);
					if (colorStr != null) {
						bgColor = defaultStyleSheet.stringToColor(colorStr);
						htmlView.setBackground(bgColor);
					}
				}
				try {
					Object attribute = attrSet.getAttribute("color");
					if ((attribute != null) && (SwingUtils.stringToColor((String) attribute) == null))
						attrSet.addAttribute("color", "Black");
					super.handleStartTag(tag, attrSet, pos);
				} catch (Exception exc) {	// don't trust HTMLEditorKit; e.g. a NPE occur on an undefined color-name
					System.err.println("Exception by Java's HTMLEditorKit (tag="+ tag +",attrSet="+ attrSet +"): "+exc);
				}
			}
		};
		getParser().parse(in, receiver, true);
		receiver.flush();
		loading = false;
	}

	public void write(Writer out, Document doc, int pos, int len)
			throws IOException, BadLocationException {
		new ExtendedHTMLWriter(out, (HTMLDocument) doc, pos, len) {
			protected void writeAttributes(AttributeSet attrSet) throws IOException {
				if (removeDELTALinksOnSave && (attrSet instanceof SimpleAttributeSet) &&
						(HTMLView.getDELTALinkIDs(attrSet) != HTMLView.BAD_LINK_IDS)) {
					MutableAttributeSet newAttrSet = (MutableAttributeSet) attrSet.copyAttributes();
					newAttrSet.removeAttribute(HTML.Attribute.HREF);
					newAttrSet.removeAttribute(HTML.Attribute.NAME);
					newAttrSet.removeAttribute(HTML.Attribute.COLOR);
					super.writeAttributes(newAttrSet);
				} else
					super.writeAttributes(attrSet);
			}
		}.write();
	}

	static class ExtendedHTMLWriter extends HTMLWriter {
		private HTMLDocument doc;
		ExtendedHTMLWriter(Writer w, HTMLDocument doc) {
			this(w, doc, 0, doc.getLength());
		}

		ExtendedHTMLWriter(Writer w, HTMLDocument doc, int pos, int len) {
			super(w, doc, pos, len);
			this.doc = doc;
		}

		protected void endTag(Element elem) throws IOException {
			if (matchNameAttribute(elem.getAttributes(), HTML.Tag.HEAD)) {
				StyleSheet styleSheet = doc.getStyleSheet();
				styleSheet.removeStyleSheet(defaultStyleSheet);
				writeStyles(styleSheet);
			}
			super.endTag(elem);
		// writeLineSeparator();
		}

		/**
		 * Outputs the styles as a single element. Styles are not stored as
		 * elements, but part of the document. For the time being styles are
		 * written out as a comment, inside a style tag.
		 */
		void writeStyles(StyleSheet sheet) throws IOException {
			if (sheet != null) {
				Enumeration styles = sheet.getStyleNames();
				if (styles != null) {
					boolean outputStyle = false;
					while (styles.hasMoreElements()) {
						String name = (String) styles.nextElement();
						// Don't write out the default style.
						if (!StyleContext.DEFAULT_STYLE.equals(name)
								&& writeStyle(name, sheet.getStyle(name), outputStyle))
							outputStyle = true;
					}
					if (outputStyle) {
						decrIndent();
						indent();
						super.write("-->");
						writeLineSeparator();
						decrIndent();
						indent();
						super.write("</style>");
						writeLineSeparator();
						indent();
					}
				}
			}
		}

		/**
		 * Outputs the named style. <code>outputStyle</code> indicates
		 * whether or not a style has been output yet. This will return
		 * true if a style is written.
		 */
		boolean writeStyle(String name, Style style, boolean outputStyle) throws IOException {
			boolean didOutputStyle = false;
			Enumeration attributes = style.getAttributeNames();
			if (attributes != null) {
				while (attributes.hasMoreElements()) {
					Object attribute = attributes.nextElement();
					if (attribute instanceof CSS.Attribute) {
						String value = style.getAttribute(attribute).toString();
						if (value != null) {
							if (!outputStyle) {
								indent();
								super.write("<style type=\"text/css\">");
								incrIndent();
								writeLineSeparator();
								indent();
								super.write("<!--");
								incrIndent();
								writeLineSeparator();
								outputStyle = true;
							}
							if (!didOutputStyle) {
								didOutputStyle = true;
								indent();
								super.write(name);
								super.write(" {");
							} else
								super.write(";");
							super.write(' ');
							super.write(attribute.toString());
							super.write(": ");
							super.write(value);
						}
					}
				}
			}
			if (didOutputStyle) {
				super.write(" }");
				writeLineSeparator();
			}
			return didOutputStyle;
		}
	}

	private class LinkInlineView extends InlineView {
		private final boolean deltaLink, overlappingDeltaLinks;
		public LinkInlineView(Element elem, boolean deltaLink, boolean overlappingDeltaLinks) {
			super(elem);
			this.deltaLink = deltaLink;
			this.overlappingDeltaLinks = overlappingDeltaLinks;
		}

		@Override public Color getForeground() {
			if (showLinks)
				if (deltaLink)
					return overlappingDeltaLinks ? prefLinkOverlappingColor : prefLinkDELTAColor;
				else
					return prefLinkHTMLColor;
			return Color.BLACK;
		}

		@Override public boolean isUnderline() {
			if (showLinks)
				return super.isUnderline();
			return false;
		}
	}
}
