package at.ac.tuwien.asgaard.delta.gui.macros;

import at.ac.tuwien.asgaard.delta.gui.macros.images.MacrosImages;
import at.ac.tuwien.asgaard.util.MiscUtils;
import org.w3c.dom.NodeList;

import javax.swing.*;
import java.util.*;

public class MacrosListItem {
	public static final	int	TYPE_ELEMENT	= 0;
	public static final	int	TYPE_MACRO		= 1;
	public static final	int	TYPE_ATTR_MACRO	= 2;
	public static final	int	TYPE_GROUP		= 3;
	public static final	int	TYPE_GROUP_LINK	= 4;
	public static final	int	TYPE_LABEL		= 5;
	public static final	int	TYPE_SEPARATOR	= 6;

	private static final int MAX_TOOLTIP_LINE_WIDTH = 75;
	protected String name;
	protected final String description;
	protected String toolTipText;
	protected String toolTipTextVerbose;
	protected final int type;
	protected final org.w3c.dom.Element xmlElement;
	protected final Group parent;
	protected final int level;

	public static final MacrosListItem SEPARATOR = new MacrosListItem(TYPE_SEPARATOR, null, null, null);

	protected MacrosListItem(int type, String name, org.w3c.dom.Element xmlDataElement, Group parent) {
		this.type = type;
		this.name = name;
		xmlElement = xmlDataElement;
		this.parent = parent;
		description = (xmlElement == null) ? null : xmlElement.getAttribute("description").trim();
		level = (parent == null) ? 0 : parent.macrosListModel.getLevel();
	}

	public final String getName() {
		return name;
	}

	public final int getType() {
		return type;
	}

	public final org.w3c.dom.Element getXMLElement() {
		return xmlElement;
	}

	public final String getDescription() {
		return description;
	}

	public String getToolTipText() {
		return toolTipText;
	}

	public String getToolTipTextVerbose() {
		return toolTipTextVerbose;
	}

	public String getShortDescription() {
		return "";
	}

	// shown in root-node of SimpleXMLTree
	public String getShortHTMLDescription() {
		return "";
	}

	public Icon getIcon() { return null;}

	public final boolean isElement() {
		return (type == TYPE_ELEMENT);
	}

	public final boolean isMacro() {
		return (type == TYPE_MACRO);
	}

	public final boolean isGroup() {
		return (type == TYPE_GROUP);
	}

	public final boolean isLabel() {
		return (type >= TYPE_LABEL);
	}

	public boolean isInsertable() {
		return false;
	}

	public final Group getParent() {
		return parent;
	}

	public final String toString() {
		return name;
	}

	// for MacrosListModel#indexOf
	public final boolean equals(Object o) {
		if (o == null) return false;
		if (!(o instanceof MacrosListItem)) return false;
		MacrosListItem otherItem = (MacrosListItem) o;
		if (type != otherItem.type) return false;
		if (type < TYPE_SEPARATOR) {
			if (!name.equals(otherItem.name)) return false;
			if (type < TYPE_LABEL) {
				if (((parent == null) && (otherItem.parent != null)) ||
					((parent != null) && (otherItem.parent == null)))
					return false;
				if (parent == null) return true;
				return parent.equals(otherItem.parent);
			}
		}
		return true;
	}

	public final int hashCode() {
		int result;
		result = name.hashCode();
		result = 29 * result + (description != null ? description.hashCode() : 0);
		result = 29 * result + (toolTipText != null ? toolTipText.hashCode() : 0);
		result = 29 * result + type;
		result = 29 * result + xmlElement.hashCode();
		result = 29 * result + (parent != null ? parent.hashCode() : 0);
		return result;
	}

	static final class ListItemComparator implements Comparator {
		private final boolean sortByType;
		private final boolean ascending;
		private final boolean ignoreCase;

		ListItemComparator(boolean sortByType, boolean ascending, boolean ignoreCase) {
			this.sortByType = sortByType;
			this.ascending = ascending;
			this.ignoreCase = ignoreCase;
		}

		public int compare(Object o1, Object o2) {
			if (o1 == o2) return 0;
			MacrosListItem item1 = (MacrosListItem) o1, item2 = (MacrosListItem) o2;
			String name1 = item1.getName(), name2 = item2.getName();
			int comp = 0;
			if (sortByType)
				comp = item1.getType() - item2.getType();
			if (comp == 0)
				if (ignoreCase)
					comp = name1.compareToIgnoreCase(name2);
				else
					comp = name1.compareTo(name2);
			if (comp == 0) comp = -1; // should never return 0 for unequal ListItems
			return (ascending ? 1 : -1) * comp;
		}
	}

	public static class Element extends MacrosListItem {
		protected int insertablePosition;
		protected final boolean standalone;
		public Element(String name, org.w3c.dom.Element xmlDataElement, int insertablePosition, Group parent) {
			this(name, TYPE_ELEMENT, xmlDataElement, insertablePosition, parent);
		}

		protected Element(String name, int type, org.w3c.dom.Element xmlDataElement, int insertablePosition, Group parent) {
			super(type, name, xmlDataElement, parent);
			this.insertablePosition = insertablePosition;
			standalone = Boolean.valueOf(xmlDataElement.getAttribute("standalone")).booleanValue();
			if (description.length() == 0) {
				toolTipText = null;
				toolTipTextVerbose = "<html>&nbsp;<b>Element:</b> <i>" + name + "</i>"
									 + "<br>&nbsp;<b>is standalone:</b> "
									 + (standalone ? "yes" : "<u>no</u> (needs additional elements/attributes)")
									 + "<br>&nbsp;<b>is insertable:</b> ";
			} else {
				toolTipText = "<html>&nbsp;<b>Element:</b> <i>" + name + "</i>"
							  + "<br>&nbsp;<b>Description:</b> "
							  + MiscUtils.wrapString(description, "<br>&nbsp;&nbsp;&nbsp;", MAX_TOOLTIP_LINE_WIDTH, 13);
				toolTipTextVerbose = toolTipText + "<br>&nbsp;<b>is standalone:</b> "
									 + (standalone ? "yes" : "<u>no</u> (needs additional elements/attributes)")
									 + "<br>&nbsp;<b>is insertable:</b> ";
			}
		}

		public final boolean isInsertable() {
			return (insertablePosition >= 0);
		}

		public final int getInsertablePosition() {
			return insertablePosition;
		}

		final void setInsertablePosition(int insertablePosition) {
			this.insertablePosition = insertablePosition;
		}

		public String getToolTipTextVerbose() {
			return toolTipTextVerbose + (isInsertable() ? "yes" : "no");
		}

		public final boolean isStandalone() {
			return standalone;
		}

		public String getShortDescription() {
			return "Element: " + name;
		}

		// shown in root-node of SimpleXMLTree
		public String getShortHTMLDescription() {
			if (insertablePosition == -1)
				return "<b>Element: <s><i>" + name + "<i></s></b>";
			return "<b>Element: <i>" + name + "</i></b>";
		}

		public Icon getIcon() {
			if (insertablePosition >= 0) return MacrosImages.ELEMENT_ICON;
			return MacrosImages.ELEMENT_NOT_INSERTABLE_ICON;
		}
	}

	public static final class Macro extends Element {
		private final boolean hasLink;
		private final boolean takesSelection;
		private final boolean inputRequired;
		private Map<String,String> constantDefMap;

		public Macro(String name, org.w3c.dom.Element xmlDataElement, int insertablePosition, Group parent) {
			super(name, TYPE_MACRO, xmlDataElement, insertablePosition, parent);
			hasLink = Boolean.valueOf(xmlDataElement.getAttribute("hasLink")).booleanValue();
			takesSelection = hasLink || Boolean.valueOf(xmlDataElement.getAttribute("takesSelection")).booleanValue();
			inputRequired = Boolean.valueOf(xmlDataElement.getAttribute("inputRequired")).booleanValue();
			if (description.length() == 0) {
				toolTipText = null;
				toolTipTextVerbose = "<html>&nbsp;<b>Element:</b> <i>" + name + "</i>"
									 + "<br>&nbsp;<b>is standalone:</b> "
									 + (standalone ? "yes" : "<u>no</u> (needs additional elements/attributes)")
									 + "<br>&nbsp;<b>has link:</b> " + (hasLink ? "yes" : "no") + "<br>&nbsp;" +
									 "<b>takes selection:</b> " + (takesSelection ? "yes" : "no") + "<br>&nbsp;" +
									 "<b>needs input:</b> " + (inputRequired ? "yes" : "no")
									 + "<br>&nbsp;<b>is insertable:</b> ";
			} else {
				toolTipText = "<html>&nbsp;<b>Element:</b> <i>" + name + "</i>"
							  + "<br>&nbsp;<b>Description:</b> "
							  + MiscUtils.wrapString(description, "<br>&nbsp;&nbsp;&nbsp;", MAX_TOOLTIP_LINE_WIDTH, 13);
				toolTipTextVerbose = toolTipText + "<br>&nbsp;<b>is standalone:</b> "
									 + (standalone ? "yes" : "<u>no</u> (needs additional elements/attributes)")
									 + "<br>&nbsp;<b>contains link:</b> " + (hasLink ? "yes" : "no") + "<br>&nbsp;" +
									 "<b>takes selection:</b> " + (takesSelection ? "yes" : "no") + "<br>&nbsp;" +
									 "<b>needs input:</b> " + (inputRequired ? "yes" : "no") 
									 + "<br>&nbsp;<b>is insertable:</b> ";
			}

			NodeList constantDefs = xmlDataElement.getElementsByTagName("DELTA_constant-def");
			int size = constantDefs.getLength();
			if (size > 0) {
				constantDefMap = new HashMap<String, String>(size);
				for (int i = 0; i < size; i++) {
					org.w3c.dom.Element element = (org.w3c.dom.Element) constantDefs.item(0);
					constantDefMap.put(element.getAttribute("name"), element.getAttribute("value"));
					element.getParentNode().removeChild(element);
				}
			}
		}

		public boolean hasLink() {
			return hasLink;
		}

		public boolean isTakesSelection() {
			return takesSelection;
		}

		public boolean isInputRequired() {
			return inputRequired;
		}

		public String getShortDescription() {
			return "Macro: " + name;
		}

		// shown in root-node of SimpleXMLTree
		public String getShortHTMLDescription() {
			if (isInsertable())
				return "<b>Macro: <i>" + name + "</i></b>";
			return "<b>Macro: <s><i>" + name + "<i></s></b>";
		}

		public Icon getIcon() {
			if (insertablePosition >= 0)
				if (hasLink)
					return MacrosImages.MACRO_LINK_ICON;
				else
					return MacrosImages.MACRO_ICON;
			if (hasLink) return MacrosImages.MACRO_LINK_NOT_INSERTABLE_ICON;
			return MacrosImages.MACRO_NOT_INSERTABLE_ICON;
		}

		public Map<String, String> getConstantDefMap() {
			return constantDefMap;
		}
	}

	public static final class Group extends MacrosListItem {
		final MacrosListModel macrosListModel;
		private final boolean groupRef;

		public Group(String name, org.w3c.dom.Element xmlDataElement, MacrosListModel mlm, Group parent) {
			super(TYPE_GROUP, name, xmlDataElement, parent);
			macrosListModel = mlm;
			groupRef = xmlElement.hasAttribute("ref");
			if (description.length() == 0) {
				toolTipText = null;
				toolTipTextVerbose = "<html>&nbsp;<b>Group:</b> <i>" + name + "</i>";
			} else {
				toolTipText = toolTipTextVerbose = "<html>&nbsp;<b>Group:</b> <i>" + name + "</i>" +
					"<br>&nbsp;<b>Description:</b> " + MiscUtils.wrapString(description, "<br>&nbsp;&nbsp;&nbsp;", MAX_TOOLTIP_LINE_WIDTH, 13);
			}
		}

		public MacrosListModel getMacrosListModel() {
			return macrosListModel;
		}

		public boolean isGroupRef() {
			return groupRef;
		}

		public Icon getIcon() {
			return MacrosImages.OPENED_GROUP_ICON;
		}
	}

	public static final class GroupLink extends MacrosListItem {
		private Group linkToGroup;
		private final String linkToGroupName;

		public GroupLink(String name, org.w3c.dom.Element xmlElement, Group parent) {
			super(TYPE_GROUP_LINK, name, xmlElement, parent);
			
			if (description.length() == 0) {
				toolTipText = null;
				toolTipTextVerbose = "<html>&nbsp;<b>Link</b> to Group \"<i>" + name + "</i>\"";
			} else {
				toolTipText = toolTipTextVerbose = "<html>&nbsp;<b>Link</b> to Group \"<i>" + name + "</i>\"" +
					"<br>&nbsp;<b>Description:</b> " + MiscUtils.wrapString(description, "<br>&nbsp;&nbsp;&nbsp;", MAX_TOOLTIP_LINE_WIDTH, 13);
			}
			
			linkToGroupName = name;
			String label = xmlElement.getAttribute("label");
			if (!label.equals(""))
				this.name = label;
		}

		String getLinkToGroupName() {
			return linkToGroupName;
		}

		void setLinkToGroup(Group linkToGroup) {
			this.linkToGroup = linkToGroup;
		}

		public Group getLinkToGroup() {
			return linkToGroup;
		}

		public Icon getIcon() {
			return MacrosImages.GROUP_LINK_ICON;
		}
	}

// labels (and subclasses are not selectable)
	public static class Label extends MacrosListItem {
		public Label(org.w3c.dom.Element xmlElement, String text) {
			super(TYPE_LABEL, text, xmlElement, null);
			if (description.length() != 0)
				toolTipText = MiscUtils.wrapString(description, "<br>&nbsp;&nbsp;&nbsp;", MAX_TOOLTIP_LINE_WIDTH, 13);
		}
	}
}