package at.ac.tuwien.asgaard.delta.gui.macros;

import at.ac.tuwien.asgaard.util.xml.dtd.DTDHelper;
import org.w3c.dom.*;

import javax.swing.*;
import java.util.*;

/**
 * MacrosListModel
 *
 * @version 2.7.3
 * @author Peter Votruba
 * @since JDK 1.3.1
 */
final class MacrosListModel extends AbstractListModel {
	private final List<MacrosListItem> listData;
	private MacrosListItem[] items;
	private final int level;
    private int itemCount;
	private boolean hasGroup;

	/** Creates new MacrosListModel */
	private MacrosListModel(int level) {
        this.level = level;
		listData = new ArrayList<MacrosListItem>();
		itemCount = 0;
		hasGroup = false;
	}

	MacrosListModel(DTDHelper dtdHelper, Node groupRootNode, Map<String, Element> macrosMap, Collection<String> undefinedMacros) {
		this(0);
		Map<String, MacrosListItem> groupItems = new HashMap<String, MacrosListItem>();
		List<MacrosListItem> groupLinkItems = new LinkedList<MacrosListItem>();
		traverseStructureXMLNodes(dtdHelper, groupRootNode, macrosMap, null, groupItems, groupLinkItems, level, undefinedMacros);
		for (Iterator iter = groupLinkItems.iterator(); iter.hasNext(); ) {
			MacrosListItem.GroupLink groupLinkItem = (MacrosListItem.GroupLink)iter.next();
			MacrosListItem.Group groupItem = (MacrosListItem.Group)groupItems.get(groupLinkItem.getLinkToGroupName());
			groupLinkItem.setLinkToGroup(groupItem);
		}
	}
	private void traverseStructureXMLNodes(DTDHelper dtdHelper, Node groupRootNode, Map<String, Element> macrosMap,
										   MacrosListItem.Group parentGroup, Map<String, MacrosListItem> groupItems, List<MacrosListItem> groupLinkItems,
										   int level, Collection<String> undefinedMacros) {
		if (groupRootNode == null) return;
		listData.clear();
		for (Node childNode = groupRootNode.getFirstChild(); childNode != null; childNode = childNode.getNextSibling())
			if (childNode.getNodeType() == Node.ELEMENT_NODE) {
				Element childElem = (Element) childNode;
				String elemName = childElem.getTagName();
				if (elemName.equals("separator"))
					listData.add(MacrosListItem.SEPARATOR);
				else if (elemName.equals("label")) {
					String textAttr = childElem.getAttribute("text");
					if (textAttr.equals("")) continue;
					listData.add(new MacrosListItem.Label(childElem, textAttr));
				}
				String nameAttr = childElem.getAttribute("name");
				if (nameAttr.equals("")) continue;
				if (elemName.equals("group")) {
					MacrosListModel mlm = new MacrosListModel((level + 1));
					MacrosListItem.Group item = new MacrosListItem.Group(nameAttr, childElem, mlm, parentGroup);
					listData.add(item);
					hasGroup = true;
					groupItems.put(item.getName(), item);
					mlm.traverseStructureXMLNodes(dtdHelper, childElem, macrosMap, item, groupItems, groupLinkItems, mlm.level, undefinedMacros);
				} else if (elemName.equals("element")) {
					int insertablePos = -1;
					if (dtdHelper.isElementInitialized())
						try {
							insertablePos = dtdHelper.getInsertIndexForNode(childElem.getFirstChild());
						} catch (Exception e) {
							// ToFix: if the targetElement of DTDHelper is set to null in another thread we got a NPE here
						}
					listData.add(new MacrosListItem.Element(nameAttr, childElem, insertablePos, parentGroup));
				} else if (elemName.equals("macro-ref")) {
					Element macroElem = macrosMap.get(nameAttr);
					if (macroElem != null) {
						int insertablePos = -1;
						if (dtdHelper.isElementInitialized())
							try {
								insertablePos = dtdHelper.getInsertIndexForMacro(macroElem);
							} catch (Exception e) {
								// ToFix: if the targetElement of DTDHelper is set to null in another thread we got a NPE here
							}
						String label = childElem.getAttribute("label").trim();
						if (!label.equals("")) nameAttr = label;
						listData.add(new MacrosListItem.Macro(nameAttr, macroElem, insertablePos, parentGroup));
					} else {
						System.out.println("Macro \"" + nameAttr + "\" referenced in group \"" + parentGroup + "\", but not found!");
						undefinedMacros.add(nameAttr);
					}
				} else if (elemName.equals("group-link")) {
					MacrosListItem.GroupLink item = new MacrosListItem.GroupLink(nameAttr, childElem, parentGroup);
					listData.add(item);
					groupLinkItems.add(item);
				}
			}
		items = new MacrosListItem[listData.size()];
		listData.toArray(items);
		itemCount = items.length;
		fireIntervalAdded(this, 0, items.length - 1);
	}

	void resetMacroListItems(DTDHelper dtdHelper) {
		boolean macroChanged = false;
		for (int i=0; i < itemCount; i++) {
			MacrosListItem item = items[i];
			if (item.isGroup())
				((MacrosListItem.Group)item).macrosListModel.resetMacroListItems(dtdHelper);
			else if (item instanceof MacrosListItem.Element) {
				int insertablePos = -1;
				if ((dtdHelper != null) && (dtdHelper.isElementInitialized()))
					if (item.isElement())
						insertablePos = dtdHelper.getInsertIndexForNode(item.getXMLElement().getFirstChild());
					else
						insertablePos = dtdHelper.getInsertIndexForMacro(item.getXMLElement());
				if (insertablePos != ((MacrosListItem.Element) item).getInsertablePosition()) {
					((MacrosListItem.Element) item).setInsertablePosition(insertablePos);
					macroChanged = true;
				}
			}
		}
		if (macroChanged)
			fireContentsChanged(this, 0, itemCount - 1);
	}

	int indexOf(Object item) {
		if (item instanceof MacrosListItem)
			return listData.indexOf(item);
		return -1;
	}

	int getLevel() {
		return level;
	}

	MacrosListItem getItemAt(int index) {
		return items[index];
	}
	boolean hasGroup() {
		return hasGroup;
	}

// ListModel methods
	public int getSize() {
		return itemCount;
	}

	public Object getElementAt(int index) {
		return items[index];
	}
}
