package at.ac.tuwien.asgaard.delta.gui.macros;

import at.ac.tuwien.asgaard.delta.gui.DeltaView;
import at.ac.tuwien.asgaard.delta.util.GlobalObjects;
import at.ac.tuwien.asgaard.delta.util.Preferences;
import at.ac.tuwien.asgaard.swing.SwingUtils;
import at.ac.tuwien.asgaard.swing.models.EmptyListModel;
import at.ac.tuwien.asgaard.swing.models.SingleListSelectionModel;
import at.ac.tuwien.asgaard.util.FileUtils;
import at.ac.tuwien.asgaard.util.Log;
import at.ac.tuwien.asgaard.util.MiscUtils;
import at.ac.tuwien.asgaard.util.xml.XMLUtils;
import at.ac.tuwien.asgaard.util.xml.dtd.DTDHelper;
import at.ac.tuwien.asgaard.util.xml.dtd.DTDHelperXML4J;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXParseException;
import org.xml.sax.XMLReader;

import java.awt.*;
import java.awt.event.*;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.util.*;

import javax.swing.*;
import javax.swing.border.*;
import javax.swing.event.*;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.stream.StreamSource;

/**
 * Class MacrosView.... todo
 *
 * @version 2.7.3
 * @author Peter Votruba
 */
public final class MacrosView extends JPanel implements DeltaView {
	private MacrosList levelLists[];
	private int levels;
	private boolean loaded;

	private File macrosFile;
	private MacrosListItem currentItem, currentOwnItem;
	private int currentListIndex;
	private int oldListIndex;
	private LinkedList<MacrosListItem> historyBackList, historyForwardList;

	private final MacrosSearchView searchView;
	private final MacrosInsertableView insertableView;

	private Preferences properties;
	private Dimension minListSize;
	private boolean synchronizeStructureView;

	private Transformer macrosTransformer;
	private XMLReader macrosFileXMLReader;
	private ErrorListener errorListener;

	private boolean processItemSelectionFlag = true, addToHistoryFlag = true;

	private Color selectionColor;

	static final Color DEFAULT_LIST_BACKGROUND  = SwingUtils.changeColor(GlobalObjects.DEFAULT_BACKGROUND, -8);
	static final Color SELECTED_LIST_BACKGROUND  = Color.WHITE;

	private Color defaultListTitleColor  = new Color(102, 102, 153);
	private Color selectedListTitleColor = SwingUtils.changeColor(Color.BLUE, -20);

	static final int MACROS_LIST_MARGIN = 4;

	private DTDHelper dtdHelper;
	private boolean isLoading;

	/** Creates new form MacroView */
	public MacrosView(final Preferences properties, final MacrosSearchView searchView, final MacrosInsertableView insertableView) {
		Log.trace("enter: MacrosView()");

		this.properties = properties;
		this.searchView = searchView;
		this.insertableView = insertableView;

		setLayout(new BoxLayout(this, BoxLayout.X_AXIS));
		setBackground(GlobalObjects.DEFAULT_BACKGROUND);
		setBorder(new EmptyBorder(MACROS_LIST_MARGIN, MACROS_LIST_MARGIN, MACROS_LIST_MARGIN, MACROS_LIST_MARGIN));
		setMinimumSize(GlobalObjects.ZERO_SIZE);
		minListSize = new Dimension();

		fetchProperties();

		insertableView.setSelectionColor(selectionColor);

		levels = 0;
		loaded = false;
		isLoading = false;

		searchView.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) { // is fired when the user doubleclicks on a macro in the search view
				currentItem = searchView.getCurrentItem();
				SwingUtils.fireActionEvent(listenerList, MacrosView.this);
			}
		});
		searchView.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent evt) {
				MacrosListItem item = searchView.getCurrentItem();
				MacrosListItem oldItem = currentItem;
				currentItem = item;
				if (oldItem != currentItem)
					SwingUtils.fireChangeEvent(listenerList, searchView);
			}
		});
		searchView.addComponentListener(new ComponentAdapter() {
			public void componentHidden(ComponentEvent evt) {
				if (!loaded) return;
				if (synchronizeStructureView) {
					MacrosListItem item = searchView.getCurrentItem();
					if ((item != null) && (item != currentOwnItem))
						selectItemPath(item, true, false);
					else selectItemPath(currentOwnItem, true, false); 
					repaint();
				} else if (currentItem != currentOwnItem) {
					currentItem = currentOwnItem;
					SwingUtils.fireChangeEvent(listenerList, this);
					if (currentListIndex != -1)
						levelLists[currentListIndex].requestFocusInWindow();
				}
			}
		});

		insertableView.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) { // is fired when the user doubleclicks on a macro in the search view
				currentItem = insertableView.getCurrentItem();
				SwingUtils.fireActionEvent(listenerList, MacrosView.this);
			}
		});
		insertableView.addChangeListener(new ChangeListener() {
			public void stateChanged(ChangeEvent evt) {
				MacrosListItem item = insertableView.getCurrentItem();
				MacrosListItem oldItem = currentItem;
				currentItem = item;
				if (oldItem != currentItem)
					SwingUtils.fireChangeEvent(listenerList, insertableView);
			}
		});
		insertableView.addComponentListener(new ComponentAdapter() {
			public void componentHidden(ComponentEvent evt) {
				if (!loaded) return;
				if (synchronizeStructureView) {
					MacrosListItem item = insertableView.getCurrentItem();
					if ((item != null) && (item != currentOwnItem))
						selectItemPath(item, true, false);
					// else selectItemPath(currentOwnItem, true); // todo ???
					repaint();
				} else if (currentItem != currentOwnItem) {
					currentItem = currentOwnItem;
					SwingUtils.fireChangeEvent(listenerList, this);
					if (currentListIndex != -1)
						levelLists[currentListIndex].requestFocusInWindow();
				}
			}
		});

		Log.trace("exit: MacrosView()");
	}

	private void init() {
		historyBackList = new LinkedList<MacrosListItem>();
		historyForwardList = new LinkedList<MacrosListItem>();

		try { 
			Source source = new StreamSource(new BufferedReader(new InputStreamReader(MacrosView.class.getResourceAsStream("macros.xsl"))), "");
			macrosTransformer = XMLUtils.getTransformerFactory().newTransformer(source);
			macrosFileXMLReader = XMLUtils.getNewXMLReader();
			macrosFileXMLReader.setEntityResolver(new EntityResolver() {
				public InputSource resolveEntity(String publicId, String systemId) /*throws SAXException, IOException*/ {
					return new InputSource(new BufferedReader(new InputStreamReader(MacrosView.class.getResourceAsStream("macros.dtd"))));
				}
			});
		} catch (Throwable t) {
			throw new RuntimeException("Couldn't process internal XSL stylesheet!\n  "+t);
		}
	}

	private void selectItemPath(MacrosListItem item, boolean _addToHistory, boolean selectFirstItemOfGroup) {
		if (item == null) {  // must be the root group-node -> select the first element of the root-group
			if (levelLists[0].getModel().getSize() > 0)
				levelLists[0].setSelectedIndex(0);
		} else {
			if (currentOwnItem.parent != item.parent) {
				processItemSelectionFlag = addToHistoryFlag = false;
				selectItemPath(item.getParent());
				processItemSelectionFlag = true;
			}

			addToHistoryFlag = _addToHistory;
			if (item.level > 0)
				levelLists[item.level].setModel(item.getParent().getMacrosListModel());
			levelLists[item.level].selectAndShowIndex(levelLists[item.level].getMacrosListModel().indexOf(item));
			if (selectFirstItemOfGroup && (item.type == MacrosListItem.TYPE_GROUP) && (levelLists[currentListIndex + 1].getModel().getSize() > 0)) {
				addToHistoryFlag = _addToHistory && !selectFirstItemOfGroup;
				levelLists[currentListIndex + 1].setSelectedIndex(0);
			}
			addToHistoryFlag = true;
		}
		levelLists[currentListIndex].requestFocusInWindow();
	}

	private int selectItemPath(MacrosListItem item) {
		int depth;
		if (item == null) return 0;
		if (item.getParent() == null)
			depth = 0;
		else
			depth = selectItemPath(item.getParent()) + 1;
		if (depth > 0)
			levelLists[depth].setModel(item.getParent().getMacrosListModel());
		levelLists[depth].selectAndShowIndex(levelLists[depth].getMacrosListModel().indexOf(item));
		return depth;
	}

	public int getLevels() {
		return levels;
	}

	public File getFile() {
		return macrosFile;
	}

	public String getFilePath() {
		return ((macrosFile == null) ? null : macrosFile.getAbsolutePath());
	}

	public MacrosListItem getCurrentItem() {
		return currentItem;
	}

	public boolean isMacroSelected() {
		return (currentItem != null) && (currentItem instanceof MacrosListItem.Element);
	}

	public Element getMacroXMLElement() {
		return (isMacroSelected() ? currentItem.getXMLElement() : null);
	}

	public boolean isMacroLink() {
		return ((currentItem != null) && (currentItem.type == MacrosListItem.TYPE_MACRO) && ((MacrosListItem.Macro) currentItem).hasLink());
	}

	public boolean isMacroInsertable() {
		return (isMacroSelected() && currentItem.isInsertable());
	}

	public int getMacroInsertablePosition() {
		return (isMacroSelected() ? ((MacrosListItem.Element) currentItem).getInsertablePosition() : -1);
	}

	public Icon getCurrentItemIcon() {
		return (currentItem == null) ? null : currentItem.getIcon();
	}

	public String getMacroSourcePlain() {
		if (!isMacroSelected()) return "";
		StringBuffer result = new StringBuffer(currentItem.getShortDescription());
		NodeList nl = currentItem.getXMLElement().getChildNodes();
		for (int i = 0; i < nl.getLength(); i++)
			result.append("\n").append(XMLUtils.serializeNode(nl.item(i), true));
		return result.toString();
	}

	public String getMacroSource() {
		return XMLUtils.transformChildren(currentItem.getXMLElement(), currentItem.getShortHTMLDescription());
	}

	public boolean setDTDHelper(DTDHelper dtdHelper) {
		if (this.dtdHelper != dtdHelper) {
			this.dtdHelper = dtdHelper;
			resetMacroListItems();
			return true;
		}
		return false;
	}

	public void resetMacroListItems() {
		if (loaded && !isLoading) {
			levelLists[0].macrosListModel.resetMacroListItems(dtdHelper); // !!! NPE happened, only reason can be that 'loaded' is true but there was an exception during loading
			searchView.resetMacroListItems();
			insertableView.resetMacroListItems();
		}
	}

	// I/O methods
	public void load(File macrosFile) throws Exception {
		load(macrosFile, "delta-link");
	}
	public void load(File macrosFile, String linkElementName) throws Exception {
		Log.trace("enter: MacrosView.load()");
		setEnabled(false);
		isLoading = true;
		if (!loaded)
			init();

		loaded = false; // otherwise: if 'loaded' was already true and there occurs an exception in the try {} below, 'loaded' would stay true, but there might be something set wrong 
		try {
			Document doc = XMLUtils.createNewDocument();
			Result result = new DOMResult(doc);
			InputSource inputSource = new InputSource(new BufferedReader(new FileReader(macrosFile)));
			inputSource.setSystemId(FileUtils.urlifyFile(macrosFile));
			Source source = new SAXSource(macrosFileXMLReader, inputSource);
			macrosTransformer.setErrorListener(getErrorListener());
			macrosTransformer.setParameter("linkElementName", linkElementName);
			macrosTransformer.transform(source, result);

			Element rootNode = doc.getDocumentElement();
			if ((rootNode == null) || (rootNode.getChildNodes() == null) || (rootNode.getChildNodes().getLength() == 0))
				throw new RuntimeException(macrosFile.getName() + " is not a Macros file (document element has no children)!");

			Node macrosRootNode = rootNode.getFirstChild();
			if ((macrosRootNode == null) || !macrosRootNode.getNodeName().equals("macros-def"))
				throw new RuntimeException(macrosFile.getName() + " is not a Macros file (top-level element 'macros-def' not found)!");

			NodeList macros = macrosRootNode.getChildNodes();
			int childrenCount = macros.getLength();
			Map<String, Element> macrosMap = new HashMap<String, Element>(childrenCount);
			for (int i = 0; i < childrenCount; i++) {
				Node node = macros.item(i);
				if (node.getNodeType() == Node.ELEMENT_NODE) {
					Element macro = (Element) node;
					macrosMap.put(macro.getAttribute("name"), macro);
				}
			}

			Node structureRootNode = macrosRootNode.getNextSibling();
			if ((structureRootNode == null) || !structureRootNode.getNodeName().equals("structure-def"))
				throw new RuntimeException(macrosFile.getName() + " is not a Macros file (top-level element 'structure-def' not found)!");

			levels = getMaxLevel(structureRootNode, 1);

			createMacrosView();

			if (dtdHelper == null)
				dtdHelper = new DTDHelperXML4J();

			Collection<String> undefinedMacros = new TreeSet<String>();
			MacrosListModel mlm = new MacrosListModel(dtdHelper, structureRootNode, macrosMap, undefinedMacros);
			if (!undefinedMacros.isEmpty()) {
				String undefinedMacrosString = MiscUtils.wrapString(undefinedMacros.toString(), "\n", 50, 0);
				JComponent messageComp = new JScrollPane(new JTextArea("There are references to following non-existing macros in the macros file:\n\n" + undefinedMacrosString));
				JOptionPane.showMessageDialog(this, messageComp, "Undefined Macros", JOptionPane.INFORMATION_MESSAGE);
				/*  // this would show a JComboBox if undefinedMacros.size < 20 - no way to always use JList
				JOptionPane optionPane = new JOptionPane("There are references to non-existing macros in the macros file:\n", JOptionPane.INFORMATION_MESSAGE, JOptionPane.DEFAULT_OPTION, null, undefinedMacros.toArray(), null);
				optionPane.setWantsInput(false);
				JDialog dialog = optionPane.createDialog(this, "Undefined Macros");
				dialog.setVisible(true);
				dialog.dispose();*/
			}

			levelLists[0].setModel(mlm);

			currentItem = null;
			historyBackList.clear();
			historyForwardList.clear();

			((JViewport) getParent()).setViewPosition(GlobalObjects.TOP_POINT);
			levelLists[0].requestFocusInWindow();
			if (levelLists[0].getModel().getSize() > 0)
				levelLists[0].setSelectedIndex(0);
	//  SwingUtils.fireChangeEvent(listenerList, this);
			this.macrosFile = macrosFile;

			searchView.setRootMacrosListModel(mlm);
			insertableView.setRootMacrosListModel(mlm);
			loaded = true;
		} finally {
			isLoading = false;
			setEnabled(true);
			Log.trace("exit: MacrosView.load()");
		}
	}

	public void reload() throws Exception {
	// JViewport viewPort = ((JViewport) getParent());
	// Point viewPos = viewPort.getViewPosition();
		Object selectedItems[] = new Object[currentListIndex + 1];
		for (int i = 0; i < currentListIndex + 1; i++)
			selectedItems[i] = levelLists[i].getSelectedValue();

		load(macrosFile);

		addToHistoryFlag = false;
		for (int i = 0; i < selectedItems.length; i++) {
			MacrosListModel mlm = levelLists[i].getMacrosListModel();
			if (mlm == null) break;
			int index = mlm.indexOf(selectedItems[i]);
			if (index == -1) break;
			levelLists[i].selectAndShowIndex(index);
		}
		addToHistoryFlag = true;
	// viewPort.setViewPosition(viewPos);
	}

	private int getMaxLevel(Node groupRootNode, int levelNr) {
		int maxLevel = levelNr;
		for (Node node = groupRootNode.getFirstChild(); node != null; node = node.getNextSibling())
			if ((node.getNodeType() == Node.ELEMENT_NODE) && (node.getNodeName().equals("group")))
				maxLevel = Math.max(maxLevel, getMaxLevel(node, levelNr + 1));
		return maxLevel;
	}

	private void createMacrosView() {
		removeAll();
		if ((levelLists == null) || (levelLists.length < levels))
			levelLists = new MacrosList[levels];
		oldListIndex = -1;
		currentListIndex = -1;
		currentItem = null;
		currentOwnItem = null;
		for (int levelIndex = 0; levelIndex < levels; levelIndex++) {
			final MacrosList list = new MacrosList();
			list.setSelectionBackground(selectionColor);
			final int _levelIndex = levelIndex;
			list.getSelectionModel().addListSelectionListener(new ListSelectionListener() {
				public void valueChanged(ListSelectionEvent evt) {
					if (!processItemSelectionFlag) return;
					int itemIndex = list.getMinSelectionIndex();
					if (itemIndex == -1) return; // is a list item selected?
					MacrosListModel model = list.getMacrosListModel();
					MacrosListItem item = model.getItemAt(itemIndex);
					int oldItemIndex = ((SingleListSelectionModel) evt.getSource()).getPreviousIndex();
					if (item.getType() >= MacrosListItem.TYPE_LABEL) { // not clickable
						int newIndex;
						if (oldItemIndex < itemIndex)
							if (itemIndex < model.getSize() - 1)
								newIndex = itemIndex + 1;
							else
								newIndex = itemIndex - 1;
						else if (itemIndex > 0)
							newIndex = itemIndex - 1;
						else
							newIndex = 1;
						list.setSelectedIndex(newIndex);
					} else {
						list.ensureIndexIsVisible(itemIndex);
						if ((oldListIndex != _levelIndex) || (oldItemIndex != itemIndex))
							listItemChanged(_levelIndex, itemIndex, false);
					}
				}
			});
			list.addMouseListener(new MouseAdapter() {
				public void mousePressed(MouseEvent evt) {
					int itemIndex = list.locationToIndex(evt.getPoint());
					if (itemIndex == -1) return;
					if (SwingUtilities.isLeftMouseButton(evt)) {
						if (evt.getClickCount() == 2)
							listItemChanged(_levelIndex, itemIndex, true); // only if double-click
					} else if (evt.getClickCount() == 1)
						list.setSelectedIndex(itemIndex);
				}
			});
			list.addKeyListener(new KeyAdapter() {
				public void keyPressed(KeyEvent evt) {
					switch (evt.getKeyCode()) {
						case KeyEvent.VK_ENTER:
							if (list.getSelectionModel().getMinSelectionIndex() == -1) return;
							if (currentItem instanceof MacrosListItem.Element)
								SwingUtils.fireActionEvent(listenerList, MacrosView.this);
							else if (currentItem.getType() == MacrosListItem.TYPE_GROUP_LINK)
								selectItemPath(((MacrosListItem.GroupLink) currentItem).getLinkToGroup(), true, false); // arg selectFirstItemOfGroup was true
							evt.consume();
							break;
						case KeyEvent.VK_LEFT:
								if (list.getSelectionModel().getMinSelectionIndex() == -1) return;
								if (_levelIndex > 0) {
									levelLists[_levelIndex - 1].requestFocusInWindow();
									levelLists[_levelIndex - 1].setSelectedIndex(levelLists[_levelIndex - 1].getSelectedIndex());
								}
								evt.consume();
								break;
						case KeyEvent.VK_RIGHT:
							if (list.getSelectionModel().getMinSelectionIndex() == -1) return;
							if (_levelIndex < levels-1) {
								if (levelLists[_levelIndex + 1].getModel().getSize() > 0) {
									levelLists[_levelIndex + 1].requestFocusInWindow();
									levelLists[_levelIndex + 1].setSelectedIndex(0);
								}
							}
							evt.consume();
							break;
					}
				}
			});
			ToolTipManager.sharedInstance().registerComponent(list);
			JScrollPane scrollPane = new JScrollPane(list);
			JPanel dummy = new JPanel(new BorderLayout());
			dummy.setOpaque(false);
			dummy.setMinimumSize(minListSize);
			dummy.setPreferredSize(minListSize);

			dummy.setBorder(new TitledBorder(GlobalObjects.EMPTY_BORDER_0, "Level " + (levelIndex + 1), TitledBorder.LEADING, TitledBorder.TOP, null, defaultListTitleColor));
			dummy.add(scrollPane);

			scrollPane.setBackground(GlobalObjects.DEFAULT_BACKGROUND);
		//	sp.setViewportBorder(DEFAULT_LIST_BORDER);
			add(dummy);
			levelLists[levelIndex] = list;
			if (levelIndex < levels - 1)
				add(Box.createHorizontalStrut(MACROS_LIST_MARGIN));
		}   // for
	}

	private void listItemChanged(int currentListIndex, int currentItemIndex, boolean doubleClick) {
	// Log.trace("enter: MacrosView.listItemChanged() - currentListIndex=" + currentListIndex + ", currentItemIndex=" + currentItemIndex);
		MacrosListModel mlm = levelLists[currentListIndex].getMacrosListModel();
		this.currentListIndex = currentListIndex;

		currentItem = mlm.getItemAt(currentItemIndex);

		int oldItemIndex = ((SingleListSelectionModel) levelLists[currentListIndex].getSelectionModel()).getPreviousIndex();
		boolean alreadySelected = ((oldListIndex == currentListIndex) && (currentItemIndex == oldItemIndex));

		if (!( (currentListIndex == oldListIndex) && (currentItemIndex == oldItemIndex) ) &&
				addToHistoryFlag && (currentOwnItem != null)) {
			historyBackList.addFirst(currentOwnItem);
			historyForwardList.clear();
		}
		currentOwnItem = currentItem;

		if (oldListIndex != currentListIndex) {
			if (oldListIndex >= 0)
				levelLists[oldListIndex].setColors(defaultListTitleColor, DEFAULT_LIST_BACKGROUND);
			levelLists[currentListIndex].setColors(selectedListTitleColor, SELECTED_LIST_BACKGROUND);
		}
		for (int i = currentListIndex + 1; i < levels; i++) {
			if (levelLists[i].getModel() == EmptyListModel.sharedInstance) break;
			levelLists[i].setModel(EmptyListModel.sharedInstance);
		}

		JComponent titlePanel = (JComponent) levelLists[currentListIndex].getParent().getParent().getParent();
		int listWidth = titlePanel.getWidth();
		JViewport viewPort = ((JViewport) getParent());
		Point viewPos = viewPort.getViewPosition();

		int viewWidth = viewPort.getExtentSize().width;
		if ((currentListIndex == 0) || (viewWidth < (listWidth + MACROS_LIST_MARGIN)))
			viewPos.x = titlePanel.getX() - MACROS_LIST_MARGIN;  // scroll to leftest pos
		else if (currentListIndex == levels - 1)
			viewPos.x = titlePanel.getX() - viewWidth + listWidth + MACROS_LIST_MARGIN;  // scroll end of current list to right side of view
		else if (mlm.hasGroup()) {
			if (viewWidth < (3 * (listWidth + MACROS_LIST_MARGIN)))
				viewPos.x = titlePanel.getX() - ((viewWidth - listWidth) >> 1);   // center current list
			else if (viewPos.x > titlePanel.getX() - listWidth - 2 * MACROS_LIST_MARGIN)
				viewPos.x = titlePanel.getX() - listWidth - 2 * MACROS_LIST_MARGIN; // scroll to start of previous list
			else if (viewPos.x + viewWidth < titlePanel.getX() + 2 * (listWidth + MACROS_LIST_MARGIN))
				viewPos.x = titlePanel.getX() - viewWidth + 2 * (listWidth + MACROS_LIST_MARGIN); // scroll end of next list to right side of view
		} else
			if ((viewWidth < (2 * (listWidth + MACROS_LIST_MARGIN))) || (viewPos.x + viewWidth < titlePanel.getX() + listWidth + MACROS_LIST_MARGIN))
				viewPos.x = titlePanel.getX() - viewWidth + listWidth + MACROS_LIST_MARGIN;  // scroll end of current list to right side of view
			else if (viewPos.x > titlePanel.getX() - listWidth - 2 * MACROS_LIST_MARGIN)
				viewPos.x = titlePanel.getX() - listWidth - 2 * MACROS_LIST_MARGIN; // scroll to start of previous list
		viewPort.setViewPosition(viewPos);

		switch (currentItem.getType()) {
			case MacrosListItem.TYPE_ELEMENT:
			case MacrosListItem.TYPE_MACRO:
				if (doubleClick)   // insert Macro/Element
					SwingUtils.fireActionEvent(listenerList, this);
				else if (!alreadySelected)
					SwingUtils.fireChangeEvent(listenerList, this);
				break;
			case MacrosListItem.TYPE_GROUP:
				levelLists[currentListIndex + 1].setModel(((MacrosListItem.Group) currentItem).getMacrosListModel());
				if (!alreadySelected)
					SwingUtils.fireChangeEvent(listenerList, this);
				break;
			case MacrosListItem.TYPE_GROUP_LINK:
				if (doubleClick) {
					selectItemPath(((MacrosListItem.GroupLink) currentItem).getLinkToGroup(), true, false); // arg selectFirstItemOfGroup was true
					return;
				} else if (!alreadySelected)
					SwingUtils.fireChangeEvent(listenerList, this);
				break;
		}

		oldListIndex = currentListIndex;
	}


	public boolean goBack() {
		if (historyBackList.isEmpty()) return false;
		historyForwardList.addFirst(currentItem);
		selectItemPath(historyBackList.removeFirst(), false, false);
		return (!historyBackList.isEmpty());
	}

	public boolean goForward() {
		if (historyForwardList.isEmpty()) return false;
		historyBackList.addFirst(currentItem);
		selectItemPath(historyForwardList.removeFirst(), false, false);
		return (!historyForwardList.isEmpty());
	}

	public boolean canGoBack() {
		return (!historyBackList.isEmpty());
	}

	public boolean canGoForward() {
		return (!historyForwardList.isEmpty());
	}


	private ErrorListener getErrorListener() {
		if (errorListener == null)
			errorListener = new ErrorListener() {
				private String getLocationString(TransformerException exc) {
					SourceLocator loc = exc.getLocator();
					if (loc != null) {
						return "(line #" + loc.getLineNumber() + ", col #" + loc.getColumnNumber() + ')';
					}
					Throwable cause = exc.getException();
					if ((cause != null) && (cause instanceof SAXParseException)) {
						SAXParseException spe = (SAXParseException) cause;
						return "(line #" + spe.getLineNumber() + ", col #" + spe.getColumnNumber() + ')';
					}
					return "";
				}
				public void warning(TransformerException exc) /*throws TransformerException*/ {
					System.err.println("Structure-View: XML-Warning " + getLocationString(exc) + "- " + ((exc.getException() != null) ? exc.getException() : exc));
				}
				public void error(TransformerException exc) throws TransformerException {
					System.err.println("Structure-View: XML-Error " + getLocationString(exc) + "- " + ((exc.getException() != null) ? exc.getException() : exc));
					throw exc;
				}
				public void fatalError(TransformerException exc) throws TransformerException {
					System.err.println("Structure-View: fatal XML-Error " + getLocationString(exc) + "- " + ((exc.getException() != null) ? exc.getException() : exc));
					throw exc;
				}
			};
		return errorListener;
	}


// Eventlistener methods
	public void addChangeListener(ChangeListener listener) {
		listenerList.add(ChangeListener.class, listener);
	// searchView.addChangeListener(listener);
	// insertableView.addChangeListener(listener);
	}

	public void removeChangeListener(ChangeListener listener) {
		listenerList.remove(ChangeListener.class, listener);
	// searchView.removeChangeListener(listener);
	// insertableView.removeChangeListener(listener);
	}

	public void addActionListener(ActionListener listener) {
		listenerList.add(ActionListener.class, listener);
	// searchView.addActionListener(listener);
	// insertableView.addActionListener(listener);
	}

	public void removeActionListener(ActionListener listener) {
		listenerList.remove(ActionListener.class, listener);
	// searchView.removeActionListener(listener);
	// insertableView.removeActionListener(listener);
	}


	private void fetchProperties() {
		selectionColor			 = properties.getColor(Preferences.KEY_COLOR_SELECTION);
		selectedListTitleColor	 = SwingUtils.changeColor(Color.BLUE, -40);
	//	selectedListTitleColor	 = SwingUtils.changeColor(properties.getColor(Preferences.KEY_COLOR_SIDE_BORDER_ACTIVE), -10);
		defaultListTitleColor	 = properties.getColor(Preferences.KEY_COLOR_SIDE_BORDER_INACTIVE);
		synchronizeStructureView = properties.getBoolean(Preferences.KEY_MACROS_SEARCH_SYNC_VIEWS);
		int	minListWidth  =	properties.getInt(Preferences.KEY_MACROS_MIN_LIST_WIDTH);
		int	minListHeight =	properties.getInt(Preferences.KEY_MACROS_MIN_LIST_HEIGHT);
		insertableView.setSelectionColor(selectionColor);
		minListSize.setSize(minListWidth, minListHeight);
	}

	public void refetchProperties(Set<String> changedProperties) {
		fetchProperties();
        if (changedProperties.contains(Preferences.KEY_COLOR_SELECTION) || changedProperties.contains(Preferences.KEY_MACROS_MIN_LIST_HEIGHT) || changedProperties.contains(Preferences.KEY_MACROS_MIN_LIST_WIDTH))
            for (int i = 0; i < levels; i++) {
                JList levelList = levelLists[i];
                levelList.setSelectionBackground(selectionColor);
                JScrollPane sp = (JScrollPane) levelList.getParent().getParent();
                sp.setMinimumSize(minListSize);
                sp.setPreferredSize(minListSize);
                sp.revalidate();
                sp.repaint();
            }
        resetMacroListItems();
	}

	public void storeProperties() {
		// currently, MacrosView don't has any properties to store, so lets delegate to MacrosSearchView
	}

	public boolean isLoading() {
		return isLoading;
	}

	public void save() /*throws IOException*/ {
		throw new UnsupportedOperationException("MacrosView doesn't support save - use Macro-Editor instead");
	}

	public void saveAs(File newXMLFile) /*throws IOException*/ {
		// currently, MacrosView doesn't support saveAs
		throw new UnsupportedOperationException("currently, MacrosView doesn't support saveAs");
	}

	public void close() {
		if (loaded) {
			removeAll();
			repaint();
			searchView.setRootMacrosListModel(null);
			insertableView.setRootMacrosListModel(null);
			levelLists = null;
			oldListIndex = -1;
			currentListIndex = -1;
			currentItem = null;
			currentOwnItem = null;
			historyBackList.clear();
			historyForwardList.clear();
			levels = 0;
			loaded = false;
			macrosFile = null;
			SwingUtils.fireChangeEvent(listenerList, this);
		}
	}

	public boolean print() {
		// TODO
		return false;
	}

	public boolean isLoaded() {
		return loaded;
	}
}
