package at.ac.tuwien.asgaard.delta.gui.xml;

import at.ac.tuwien.asgaard.util.xml.XMLUtils;
import at.ac.tuwien.asgaard.util.xml.dtd.DTDHelper;
import at.ac.tuwien.asgaard.util.Log;
import at.ac.tuwien.asgaard.delta.gui.misc.undo.SideAbstractUndoableEdit;
import at.ac.tuwien.asgaard.delta.gui.MainFrame;
import at.ac.tuwien.asgaard.delta.gui.DeltaSideController;
import at.ac.tuwien.asgaard.delta.xmlBeans.xmlLanguage.LinkedAttributes;
import at.ac.tuwien.asgaard.delta.xmlBeans.xmlLanguage.LinkedAttribute;
import at.ac.tuwien.asgaard.delta.xmlBeans.xmlLanguage.XmlLanguage;
import org.w3c.dom.*;

import javax.swing.event.*;
import javax.swing.undo.*;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.awt.*;

public final class XMLAttributesTableModel extends AbstractAttributesTableModel implements Comparator {
	private Element xmlElement;
	private NamedNodeMap attributes;
	private XMLAttributesTable xmlAttributesTable;

	XMLAttributesTableModel() {
		super(new ArrayList(2), new String[]{"  Attribute  ", "Value"});
	}

	Element getXMLElement() {
		return xmlElement;
	}
	void setXMLElement(Element newXMLElement) {
		boolean doRefresh = (((xmlElement != null) && xmlElement.hasAttributes()) ||
							 ((newXMLElement != null) && newXMLElement.hasAttributes()));
		xmlElement = newXMLElement;
		if (doRefresh) refresh();
	}

	void refresh() {
		tableData.clear();
		if (xmlElement != null) {
			attributes = xmlElement.getAttributes();
			for (int i=0; i < attributes.getLength(); i++) {
				Attr attribute = (Attr)attributes.item(i);
				tableData.add(new Object[] {attribute.getName(), attribute.getValue()});
			}
			Collections.sort(tableData, this);
		} else
			attributes = null;
		fireTableChanged(new TableModelEvent(this, 0, tableData.size()-1, 0, TableModelEvent.INSERT));
	}

	int addRow(String attrName, String attrValue, boolean calledFromUndoOrRedo) {
		lastChangedAttributeName = attrName;
		attributes = xmlElement.getAttributes();
		if (!calledFromUndoOrRedo)
			fireUndoableEditHappened(new AttributeAddedUndoableEdit(xmlAttributesTable.xmlTreeNode, xmlElement, (Attr)attributes.getNamedItem(attrName)));
		return addRow(new Object[] {attrName, attrValue});
	}

	boolean removeAttribute(int rowIndex, boolean calledFromUndoOrRedo) {
		lastChangedAttributeName = (String)((Object[])tableData.get(rowIndex))[0];
		Attr attribute = xmlElement.getAttributeNode(lastChangedAttributeName);
		if (!calledFromUndoOrRedo)
			xmlElement.removeAttribute(lastChangedAttributeName);
		if (attributes.getLength() != tableData.size()) {
			removeRow(rowIndex);
			attributes = xmlElement.getAttributes();
		} else {
			String newValue = attribute.getValue();
			setValueAt(newValue, rowIndex, 1);// set default attribute value
			lastChangedAttributeName = null;
		}
		if (!calledFromUndoOrRedo)
			fireUndoableEditHappened(new AttributeRemovedUndoableEdit(xmlAttributesTable.xmlTreeNode, xmlElement, attribute));
		return (lastChangedAttributeName != null);
	}

	public void setValueAt(Object newValue, int rowIndex, int columnIndex) {
		setValueAt(newValue, rowIndex, columnIndex, true);
	}
	protected void setValueAt(Object newValue, int rowIndex, int columnIndex, boolean fireUndoEvent) {
		Object[] rowData = (Object[])tableData.get(rowIndex);
		Attr attribute = (Attr)attributes.getNamedItem((String)rowData[0]);
		Object oldValue = rowData[1];
		if (!oldValue.equals(newValue)) {
			String newStringValue = newValue.toString();
			if (!(newValue instanceof Integer))
				newStringValue = XMLUtils.escapeCharacters(newStringValue);
			rowData[columnIndex] = newValue;
			attribute.setValue(newStringValue);
			fireTableChanged(new TableModelEvent(this, rowIndex, rowIndex, columnIndex, TableModelEvent.UPDATE));
			if (fireUndoEvent)
				fireUndoableEditHappened(new AttributeChangedUndoableEdit(xmlAttributesTable, attribute, oldValue.toString(), newStringValue, rowIndex));
		}
	}

	Attr getAttribute(int rowIndex) {
		//return (Attr)attributes.item(rowIndex); // NamedNodeMap attributes may not have the same ordering as tableModel
		return (Attr) attributes.getNamedItem((String) ((Object[]) tableData.get(rowIndex))[0]);
	}

	String getElementName(int rowIndex) {
		return xmlElement.getNodeName();
	}

	/**
	 * Adds an undo listener for notification of any changes.
	 *
	 * @param listener the <code>UndoableEditListener</code> to add
	 * @see #removeUndoableEditListener
	 */
	void addUndoableEditListener(UndoableEditListener listener) {
		listenerList.add(UndoableEditListener.class, listener);
	}

	/**
	 * Removes an undo listener.
	 *
	 * @param listener the <code>UndoableEditListener</code> to remove
	 * @see #addUndoableEditListener
	 */
	void removeUndoableEditListener(UndoableEditListener listener) {
		listenerList.remove(UndoableEditListener.class, listener);
	}

	private void fireUndoableEditHappened(UndoableEdit undoableEdit) {
		UndoableEditEvent evt = new UndoableEditEvent(this, undoableEdit);
		// Guaranteed to return a non-null array
		Object[] listeners = listenerList.getListenerList();
		// Process the listeners last to first, notifying those that are interested in this event
		for (int i = listeners.length - 2; i >= 0; i -= 2) {
			if (listeners[i] == UndoableEditListener.class)
				((UndoableEditListener) listeners[i + 1]).undoableEditHappened(evt);
		}
	}

	void postInit(XMLAttributesTable xmlAttributesTable) {
		this.xmlAttributesTable = xmlAttributesTable;
	}

	private String lastChangedAttributeName;
	public String getLastChangedAttributeName() {
		return lastChangedAttributeName;
	}

	void changeAttributeName(int rowIndex, String newAttrName) {
		Object[] rowData = (Object[]) tableData.get(rowIndex);
		lastChangedAttributeName = newAttrName;
		rowData[0] = newAttrName;
		removeRow(rowIndex);
		addRow(rowData, rowIndex);
	}


//	public boolean equals(Object obj) {
//		return false;
//	}

	public int compare(Object o1, Object o2) {
		Object[] row1 = (Object[])o1, row2 = (Object[])o2;
		String attrName1 = (String)row1[0], attrName2 = (String)row2[0];
		int attrType1 = xmlAttributesTable.dtdHelper.getAttributeDefaultType(xmlElement, attrName1);
		int attrType2 = xmlAttributesTable.dtdHelper.getAttributeDefaultType(xmlElement, attrName2);
		if (attrType1 != attrType2) {
		 if (attrType1 == DTDHelper.REQUIRED) return -1;
		 if (attrType2 == DTDHelper.REQUIRED) return 1;
		 if (attrType1 == DTDHelper.FIXED) return 1;
			if (attrType2 == DTDHelper.FIXED) return -1;
		}
		return attrName1.compareToIgnoreCase(attrName2);
	}

	public static class AttributeChangedUndoableEdit extends SideAbstractUndoableEdit {
		public String oldValue, newValue;
		public final int rowIndex;
		public Attr attribute;
		public XMLTreeNode xmlTreeNode;
		private XMLAttributesTable xmlAttributesTable;
		private final XmlLanguage currentXMLLanguage;

		AttributeChangedUndoableEdit(XMLAttributesTable xmlAttributesTable, Attr attribute, String oldValue, String newValue, int rowIndex) {
			super(false);
			this.xmlAttributesTable = xmlAttributesTable;
			this.xmlTreeNode = xmlAttributesTable.xmlTreeNode;
			this.attribute = attribute;
			this.newValue = newValue;
			this.oldValue = oldValue;
			this.rowIndex = rowIndex;
			Component comp = xmlAttributesTable;
			while (!(comp instanceof MainFrame.DeltaSide))
				comp = comp.getParent();
			DeltaSideController thisSideController = ((MainFrame.DeltaSide) comp).sideController;
			this.currentXMLLanguage = thisSideController.currentXMLLanguage;
			undoPresentationName = "Undo - Attribute value change";
			redoPresentationName = "Redo - Attribute value change";
			updateLinkedAttributes(false);
		}

		public void undo() throws CannotUndoException {
			super.undo();
			attribute.setValue(oldValue);
			updateLinkedAttributes(true);
			xmlAttributesTable.xmlView.centerAndSelectTreeNode(xmlTreeNode);
			xmlAttributesTable.xmlAttributesTableModel.setValueAt(oldValue, rowIndex, 1, false);
			xmlAttributesTable.scrollToAndSelectRow(rowIndex);
			xmlAttributesTable.xmlView.xmlTreeNodeChanged(xmlTreeNode);
		}
		public void redo() throws CannotRedoException {
			super.redo();
			attribute.setValue(newValue);
			updateLinkedAttributes(false);
			xmlAttributesTable.xmlView.centerAndSelectTreeNode(xmlTreeNode);
			xmlAttributesTable.xmlAttributesTableModel.setValueAt(newValue, rowIndex, 1, false);
			xmlAttributesTable.scrollToAndSelectRow(rowIndex);
			xmlAttributesTable.xmlView.xmlTreeNodeChanged(xmlTreeNode);
		}

		public void die() {
			super.die();
			oldValue = null;
			newValue = null;
			attribute = null;
			xmlTreeNode = null;
		}

		private void updateLinkedAttributes(boolean undoing) {
			if (currentXMLLanguage != null) {
				LinkedAttributes linkedAttributes = currentXMLLanguage.getLinkedAttributes();
				if (linkedAttributes != null) {
					List<LinkedAttribute> linkedAttributeList = linkedAttributes.getLinkedAttributeList();
					Document ownerDocument = attribute.getOwnerDocument();
					for (LinkedAttribute linkedAttribute : linkedAttributeList) {
						if (XMLUtils.matchNodeWithXPath(attribute, linkedAttribute.getSourceXPath())) {
							List list = XMLUtils.evaluateXPathSilently(linkedAttribute.getTargetAttributeXPath(), ownerDocument);
							for (Object item : list)
								if (item instanceof Attr) {
									Attr targetAttribute = (Attr) item;
									if (targetAttribute.getValue().equals(undoing ? newValue : oldValue)) {
										targetAttribute.setValue(undoing ? oldValue : newValue);
										XMLTreeNode targetTreeNode =
												xmlAttributesTable.xmlView.getTreeNode4XMLNode(targetAttribute.getOwnerElement());
										xmlAttributesTable.xmlView.xmlTreeNodeChanged(targetTreeNode);
									}
								}
						}
					}
				}
			}
		}
	}

	private class AttributeAddedUndoableEdit extends SideAbstractUndoableEdit {
		private XMLTreeNode xmlTreeNode;
		private Element element;
		private Attr attribute;

		AttributeAddedUndoableEdit(XMLTreeNode xmlTreeNode, Element element, Attr attribute) {
			super(false);
			this.xmlTreeNode = xmlTreeNode;
			this.attribute = attribute;
			this.element = element;
			undoPresentationName = "Undo - remove previously added attribute \"" + attribute.getName() + "\"";
			redoPresentationName = "Redo - recover attribute \"" + attribute.getName() + "\"";
		}

		public void undo() throws CannotUndoException {
			super.undo();
			element.removeAttributeNode(attribute);
			xmlAttributesTable.xmlView.centerAndSelectTreeNode(xmlTreeNode);
			int row = getRowCount() - 1;
			boolean removed = removeAttribute(row, true);
			if (removed && !isEmpty()) {
				row = Math.min(row, getRowCount() - 1);
				xmlAttributesTable.scrollToAndSelectRow(row);
			}
			xmlAttributesTable.xmlView.xmlTreeNodeChanged(xmlTreeNode);
		}

		public void redo() throws CannotRedoException {
			super.redo();
			element.setAttributeNode(attribute);
			xmlAttributesTable.xmlView.centerAndSelectTreeNode(xmlTreeNode);
			int newIndex = addRow(attribute.getName(), attribute.getValue(), true);
			xmlAttributesTable.scrollToAndSelectRow(newIndex);
			xmlAttributesTable.xmlView.xmlTreeNodeChanged(xmlTreeNode);
		}

		public void die() {
			super.die();
			Log.trace("enter: AttributeAddedUndoableEdit.die()");
			xmlTreeNode = null;
			element = null;
			attribute = null;
		}
	}

	private class AttributeRemovedUndoableEdit extends SideAbstractUndoableEdit {
		private XMLTreeNode xmlTreeNode;
		private Element element;
		private Attr attribute;

		AttributeRemovedUndoableEdit(XMLTreeNode xmlTreeNode, Element element, Attr attribute) {
			super(false);
			this.xmlTreeNode = xmlTreeNode;
			this.attribute = attribute;
			this.element = element;
			undoPresentationName = "Undo - recover removed attribute \"" + attribute.getName() + "\"";
			redoPresentationName = "Redo - remove attribute \"" + attribute.getName() + "\" again";
		}

		public void undo() throws CannotUndoException {
			super.undo();
			element.setAttributeNode(attribute);
			xmlAttributesTable.xmlView.centerAndSelectTreeNode(xmlTreeNode);
			int newIndex = addRow(attribute.getName(), attribute.getValue(), true);
			xmlAttributesTable.scrollToAndSelectRow(newIndex);
			xmlAttributesTable.xmlView.xmlTreeNodeChanged(xmlTreeNode);
		}

		public void redo() throws CannotRedoException {
			super.redo();
			element.removeAttributeNode(attribute);
			xmlAttributesTable.xmlView.centerAndSelectTreeNode(xmlTreeNode);
			int row = getRowCount() - 1;
			boolean removed = removeAttribute(row, true);
			if (removed && !isEmpty()) {
				row = Math.min(row, getRowCount() - 1);
				xmlAttributesTable.scrollToAndSelectRow(row);
			}
			xmlAttributesTable.xmlView.xmlTreeNodeChanged(xmlTreeNode);
		}

		public void die() {
			super.die();
			Log.trace("enter: AttributeRemovedUndoableEdit.die()");
			xmlTreeNode = null;
			element = null;
			attribute = null;
		}
	}
}
