package at.ac.tuwien.asgaard.delta.util;

import at.ac.tuwien.asgaard.util.Log;

import java.io.*;
import java.util.Date;
import java.util.Iterator;
import java.util.Properties;
import java.util.TreeSet;

public abstract class AbstractDELTAProperties extends Properties {
    protected File propertiesFile;
	protected String propertiesFileName;
	protected Properties defaultProperties;

	public final boolean loadPreferences;

	protected AbstractDELTAProperties(String propertiesFileName, boolean useDefaultProperties, boolean loadFile, boolean createIfFileNotExists) {
		if (useDefaultProperties) {
			defaultProperties = createDefaultProperties();
			putAll(defaultProperties); // don't use "default" properties, because they're not saved
		}
		this.propertiesFileName = propertiesFileName;
		this.propertiesFile = new File(propertiesFileName);
        loadPreferences = loadFile;
		if (loadFile)
			try {
				load(propertiesFileName);
			} catch (IOException e) {
				if (createIfFileNotExists)
					try {
						save();
					} catch (IOException e1) {
						Log.error("Couldn't write default properties to \""+ propertiesFileName +"\"!");
					}
				else
					Log.error("Couldn't read from \"" + propertiesFileName + "\"!");
			}
	}

	protected abstract Properties createDefaultProperties();
	protected abstract String getPropertiesHeader();

	public final void reload() throws IOException {
		load(propertiesFileName);
	}

	public final void load(String propertiesFileName) throws IOException {
        File propertiesFile = new File(propertiesFileName);
		FileInputStream in = new FileInputStream(propertiesFile);
		load(in);
		in.close();
		this.propertiesFileName = propertiesFileName;
        this.propertiesFile = propertiesFile;
		Log.trace("\"" + propertiesFile.getAbsolutePath() + "\" loaded.\n"+
			"  properties = " + this + "\n");
	}

	public final void save() throws IOException {
		saveAs(propertiesFileName);
	}

	public final void saveAs(String propertiesFileName) throws IOException {
        File propertiesFile = new File(propertiesFileName);
		FileOutputStream out = new FileOutputStream(propertiesFile);
		store(out, getPropertiesHeader());
		out.close();
		this.propertiesFileName = propertiesFileName;
        this.propertiesFile = propertiesFile;
		Log.trace("\"" + propertiesFile.getAbsolutePath() + "\" saved.\n" +
		          "  properties = " + this + "\n");
	}

	public File getPropertiesFile() {
		return propertiesFile;
	}

	public final int getInt(String key) {
		return getInt(key, 0);
	}
	/**
	 * Gets a Integer Property.
	 * @param key Property-Key.
	 * @param defaultValue Default-Value
	 * @param loadProperty if <code>true</code> looks in the preferences, else returns <code>defaultValue</code>
	 * @return the Integer Property
	 */
	public final int getInt(String key, int defaultValue, boolean loadProperty) {
		return (loadProperty ? getInt(key, defaultValue) : defaultValue);
	}
	public final int getInt(String key, int defaultValue) {
		if (key == null) return defaultValue;
		String s = getProperty(key);
		if (s == null) return defaultValue;
		int value = defaultValue;
		try {
			value = Integer.parseInt(s);
		} catch (NumberFormatException e) {
			try {
				s = defaultProperties.getProperty(key);
				if (s == null) return defaultValue;
				value = Integer.parseInt(s);
			} catch (Exception e1) { }
		}
		return value;
	}

	public final void putInt(String key, int value) {
		if (key == null) return;
		setProperty(key, String.valueOf(value));
	}

	public final char getChar(String key, char defaultValue) {
		if (key == null) return defaultValue;
		String s = getProperty(key);
		if ((s == null) || (s.length() == 0)) return defaultValue;
		return s.charAt(0);
	}

	public final boolean getBoolean(String key) {
		if (key == null) return false;
		return Boolean.valueOf(getProperty(key)).booleanValue();
	}

	public final void putBoolean(String key, boolean value) {
		if (key == null) return;
		setProperty(key, String.valueOf(value));
	}

	public final String getString(String key) {
		if (key == null) return null;
		String result = getProperty(key);
		return ("".equals(result) ? null : result); 
	}

	public final void putString(String key, String value) {
		if ((key == null) || (value == null)) return;
		setProperty(key, value);
	}

	public final Properties getProperties() {
		return this;
	}

	public final String getPropertiesFileName() {
		return propertiesFileName;
	}

	public synchronized void store(OutputStream out, String comments)
			throws IOException {
		BufferedWriter awriter;
		awriter = new BufferedWriter(new OutputStreamWriter(out, "8859_1"));
		if (comments != null)
			writeln(awriter, "#" + comments);
		writeln(awriter, "#" + new Date().toString());
		for (Iterator iter = new TreeSet(keySet()).iterator(); iter.hasNext();) {
			String key = (String) iter.next();
			String val = (String) get(key);
			key = saveConvert(key, true);
			/* No need to escape embedded and trailing spaces for value, hence
			* pass false to flag. */
			val = saveConvert(val, false);
			writeln(awriter, key + "=" + val);
		}
		awriter.flush();
	}

	private static void writeln(BufferedWriter bw, String s) throws IOException {
		bw.write(s);
		bw.newLine();
	}

	/** Converts unicodes to encoded &#92;uxxxx and escapes
	* special characters with a preceding slash
	*/
	private static String saveConvert(String theString, boolean escapeSpace) {
		int len = theString.length();
		int bufLen = len * 2;
		if (bufLen < 0) {
			bufLen = Integer.MAX_VALUE;
		}
		StringBuffer outBuffer = new StringBuffer(bufLen);

		for (int x = 0; x < len; x++) {
			char aChar = theString.charAt(x);
			// Handle common case first, selecting largest block that
			// avoids the specials below
			if ((aChar > 61) && (aChar < 127)) {
				if (aChar == '\\') {
					outBuffer.append('\\');
					outBuffer.append('\\');
					continue;
				}
				outBuffer.append(aChar);
				continue;
			}
			switch (aChar) {
				case ' ':
					if (x == 0 || escapeSpace)
						outBuffer.append('\\');
					outBuffer.append(' ');
					break;
				case '\t':
					outBuffer.append('\\');
					outBuffer.append('t');
					break;
				case '\n':
					outBuffer.append('\\');
					outBuffer.append('n');
					break;
				case '\r':
					outBuffer.append('\\');
					outBuffer.append('r');
					break;
				case '\f':
					outBuffer.append('\\');
					outBuffer.append('f');
					break;
				case '=': // Fall through
				case ':': // Fall through
				case '#': // Fall through
				case '!':
					outBuffer.append('\\');
					outBuffer.append(aChar);
					break;
				default:
					if ((aChar < 0x0020) || (aChar > 0x007e)) {
						outBuffer.append('\\');
						outBuffer.append('u');
						outBuffer.append(toHex((aChar >> 12) & 0xF));
						outBuffer.append(toHex((aChar >> 8) & 0xF));
						outBuffer.append(toHex((aChar >> 4) & 0xF));
						outBuffer.append(toHex(aChar & 0xF));
					} else {
						outBuffer.append(aChar);
					}
			}
		}
		return outBuffer.toString();
	}

	private static char toHex(int nibble) {
		return hexDigit[(nibble & 0xF)];
	}

	/**
	 * A table of hex digits
	 */
	private static final char[] hexDigit = {
		'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
	};
}
