package at.ac.tuwien.asgaard.delta.xmlBeans;

import at.ac.tuwien.asgaard.delta.gui.images.MainImages;
import at.ac.tuwien.asgaard.util.FileUtils;
import at.ac.tuwien.asgaard.util.Log;
import at.ac.tuwien.asgaard.util.MiscUtils;
import at.ac.tuwien.asgaard.util.xml.XMLUtils;

import net.sf.saxon.xpath.XPathExpressionImpl;
import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import java.io.*;

import javax.swing.*;
import javax.xml.transform.TransformerException;

public final class Project extends AbstractXMLBean {
	private static final EntityResolver dtdEntityResolver = new EntityResolver() {
		public InputSource resolveEntity(String publicId, String systemId) {
			return new InputSource(new BufferedInputStream(Project.class.getResourceAsStream("project.dtd")));
		}
	};

	private File projectFile, leftSideFile, rightSideFile, macrosFile;
	protected final XPathExpressionImpl xpathLeftSideFile, xpathRightSideFile, xpathMacrosFile;
// private static org.apache.xpath.XPath xpathLeftSideFile, xpathRightSideFile,  xpathMacrosFile;

	public Project(File projectFile, String name, File leftSideFile, File rightSideFile, File macrosFile) {
		this(projectFile, XMLUtils.createNewDocument("-//Asgaard//DTD DELT/A-Project//EN", "http://www.asgaard.tuwien.ac.at/~peter/DELTA/dtd/project.dtd", "project"));

		this.leftSideFile = leftSideFile;
		this.rightSideFile = rightSideFile;
		this.macrosFile = macrosFile;

		rootElement.setAttribute("name", name);
		Element element;
		if (leftSideFile != null) {
			element = documentNode.createElement("leftSideFile");
			element.setAttribute("path", FileUtils.toRelativeFile(leftSideFile, this.projectFile));
			rootElement.appendChild(element);
		}
		if (rightSideFile != null) {
			element = documentNode.createElement("rightSideFile");
			element.setAttribute("path", FileUtils.toRelativeFile(rightSideFile, this.projectFile));
			rootElement.appendChild(element);
		}
		if (macrosFile != null) {
			element = documentNode.createElement("macrosFile");
			element.setAttribute("path", FileUtils.toRelativeFile(macrosFile, this.projectFile));
			rootElement.appendChild(element);
		}
	}

	private Project(File projectFile, Document documentNode) {
		super(documentNode);
		this.projectFile = projectFile;

		try {
			xpathLeftSideFile = xpeval.createExpression(
					"/project/leftSideFile"); //new org.apache.xpath.XPath("/project/leftSideFile", null, null, org.apache.xpath.XPath.SELECT, null);//new DOMXPath("/project/leftSideFile");
			xpathRightSideFile = xpeval.createExpression(
					"/project/rightSideFile"); //new org.apache.xpath.XPath("/project/rightSideFile", null, null, org.apache.xpath.XPath.SELECT, null);//new DOMXPath("/project/rightSideFile");
			xpathMacrosFile = xpeval.createExpression(
					"/project/macrosFile"); //new org.apache.xpath.XPath("/project/macrosFile", null, null, org.apache.xpath.XPath.SELECT, null);//new DOMXPath("/project/macrosFile");
		} catch (TransformerException exc) {
			throw new RuntimeException(exc); // should not happen
		}
	}

	public static Project load(File projectFile) throws IOException, SAXException {
		DOMParser domParser = XMLUtils.getValidatingDOMParser();
		domParser.setEntityResolver(dtdEntityResolver);
		InputSource inputSource = new InputSource(new BufferedReader(new FileReader(projectFile)));
		inputSource.setSystemId(projectFile.getPath());
		domParser.parse(inputSource);
		return new Project(projectFile, domParser.getDocument());
	}

	public static Project loadSilent(File projectFile) {
		if (projectFile == null) return null;
		try {
			return load(projectFile);
		} catch (IOException exc) {
			Log.error("I/O-Error while loading DELT/A-Project XML-File \"" + projectFile.getAbsolutePath() + "\": " +
					  exc.getMessage());
		//	exc.printStackTrace();
			return null;
		} catch (SAXException exc) {
			Log.error("Couldn't parse DELT/A-Project XML-File \"" + projectFile.getAbsolutePath() + "\": " + exc.getMessage());
		//	exc.printStackTrace();
			return null;
		}
	}

	public void save() throws IOException {
		BufferedWriter bufferedWriter = new BufferedWriter(new FileWriter(projectFile));
		XMLUtils.serializeNode(documentNode, bufferedWriter, true);
		bufferedWriter.flush();
		bufferedWriter.close();
	}

	public void saveAs(File newProjectFile) throws IOException {
		// if (!MiscUtils.equals(projectFile, newProjectFile)) {
		boolean reset = !projectFile.getParent().equals(newProjectFile.getParent());
		projectFile = newProjectFile;
		if (reset) {
			// if newProjectFile is in an other directory, the relative path to the left/right/macros file have to be reset
			File f = leftSideFile;
			leftSideFile = null;
			setLeftSideFile(f);
			f = rightSideFile;
			rightSideFile = null;
			setRightSideFile(f);
			f = macrosFile;
			macrosFile = null;
			setMacrosFile(f);
		}
		save();
		// }
	}

	public File getProjectFile() {
		return projectFile;
	}

	public Icon getIcon() {
		return getIcon(projectFile);
	}

	public Icon getDefaultIcon() {
		return MainImages.PROJECT_ICON;
	}

	public File getLeftSideFile() {
		if (leftSideFile == null)
			leftSideFile = FileUtils.toExistingAbsoluteFile(getLeftSideFileName(), projectFile);
		return leftSideFile;
	}

	public void setLeftSideFile(File leftSideFile) {
		if (!MiscUtils.equals(this.leftSideFile, leftSideFile)) {
			this.leftSideFile = leftSideFile;
			try {
				Element element = (Element) xpathLeftSideFile
						.evaluateSingle(); //selectSingleNode(documentNode, xpathLeftSideFile);//xpathLeftSideFile.selectSingleNode(documentNode);
				if (leftSideFile != null) {
					if (element == null) {
						element = documentNode.createElement("leftSideFile");
						Node beforeNode = rootElement.getFirstChild();
						if ((beforeNode != null) && beforeNode.getNodeName().equals("description"))
							beforeNode = beforeNode.getNextSibling();
						rootElement.insertBefore(element, beforeNode);
					}
					element.setAttribute("path", FileUtils.toRelativeFile(leftSideFile, projectFile));
				} else if (element != null)
					rootElement.removeChild(element);
			} catch (TransformerException exc) {
				throw new RuntimeException(exc); // should not happen
			}
		}
	}

	public String getLeftSideFileName() {
		try {
			Element element = (Element) xpathLeftSideFile
					.evaluateSingle(); //selectSingleNode(documentNode, xpathLeftSideFile);//xpathLeftSideFile.selectSingleNode(documentNode);
			if (element == null) return null;
			return element.getAttribute("path");
		} catch (TransformerException exc) {
			return null;
		}
	}

	public File getRightSideFile() {
		if (rightSideFile == null)
			rightSideFile = FileUtils.toExistingAbsoluteFile(getRightSideFileName(), projectFile);
		return rightSideFile;
	}

	public void setRightSideFile(File rightSideFile) {
		if (!MiscUtils.equals(this.rightSideFile, rightSideFile)) {
			this.rightSideFile = rightSideFile;
			try {
				Element element = (Element) xpathRightSideFile
						.evaluateSingle(); //selectSingleNode(documentNode, xpathRightSideFile);//xpathRightSideFile.selectSingleNode(documentNode);
				if (rightSideFile != null) {
					if (element == null) {
						element = documentNode.createElement("rightSideFile");
						Node beforeNode = rootElement.getLastChild();
						if ((beforeNode != null) && !beforeNode.getNodeName().equals("macrosFile"))
							beforeNode = null;
						rootElement.insertBefore(element, beforeNode);
					}
					element.setAttribute("path", FileUtils.toRelativeFile(rightSideFile, projectFile));
				} else if (element != null)
					rootElement.removeChild(element);
			} catch (TransformerException exc) {
				throw new RuntimeException(exc); // should not happen
			}
		}
	}

	public String getRightSideFileName() {
		try {
			Element element = (Element) xpathRightSideFile
					.evaluateSingle(); //selectSingleNode(documentNode, xpathRightSideFile);//xpathRightSideFile.selectSingleNode(documentNode);
			if (element == null) return null;
			return element.getAttribute("path");
		} catch (TransformerException exc) {
			return null;
		}
	}


	public File getMacrosFile() {
		if (macrosFile == null)
			macrosFile = FileUtils.toExistingAbsoluteFile(getMacrosFileName(), projectFile);
		return macrosFile;
	}

	public void setMacrosFile(File macrosFile) {
		if (!MiscUtils.equals(this.macrosFile, macrosFile)) {
			this.macrosFile = macrosFile;
			try {
				Element element = (Element) xpathMacrosFile
						.evaluateSingle(); //selectSingleNode(documentNode, xpathMacrosFile);//xpathMacrosFile.selectSingleNode(documentNode);
				if (macrosFile != null) {
					if (element == null) {
						element = documentNode.createElement("macrosFile");
						rootElement.appendChild(element);
					}
					element.setAttribute("path", FileUtils.toRelativeFile(macrosFile, projectFile));
				} else if (element != null)
					rootElement.removeChild(element);
			} catch (TransformerException exc) {
				throw new RuntimeException(exc); // should not happen
			}
		}
	}

	public String getMacrosFileName() {
		try {
			Element element = (Element) xpathMacrosFile
					.evaluateSingle(); //selectSingleNode(documentNode, xpathMacrosFile);//xpathMacrosFile.selectSingleNode(documentNode);
			if (element == null) return null;
			return element.getAttribute("path");
		} catch (TransformerException exc) {
			return null;
		}
	}

	// copy constructor
	public Project(Project otherProject) {
		this(otherProject.projectFile, otherProject.documentNode);
		otherProject.description = description;
		otherProject.icon = icon;
		otherProject.iconFileName = iconFileName;
		otherProject.leftSideFile = leftSideFile;
		otherProject.macrosFile = macrosFile;
		otherProject.name = name;
		otherProject.rightSideFile = rightSideFile;
	}
}
