package at.ac.tuwien.asgaard.swing.autoComplete;

import at.ac.tuwien.asgaard.swing.models.SimpleComboBoxModel;
import at.ac.tuwien.asgaard.util.DefaultComparator;

import java.awt.event.*;
import java.lang.reflect.Field;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import javax.swing.*;
import javax.swing.plaf.basic.BasicComboBoxUI;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.PlainDocument;

/**
 * Class AutoCompleteComboBox ... modified code from http://www.orbital-computer.de/JComboBox/
 * @author peter
 */
public class AutoCompleteComboBox extends JComboBox {
    private JTextField editorComp;
    
    //  flag to indicate if setSelectedItem has been called
    // subsequent calls to remove/insertString should be ignored
    private boolean selecting = false;
    private boolean hidePopupOnFocusLoss;
    private boolean hitBackspace = false;
    private boolean hitBackspaceOnSelection;
    
//  needed for "Non-strict matching"
    boolean listContainsSelectedItem; 
    
    private AutoCompleteComboBoxDocument document;
    
    // if false -> "Non-strict matching"; if true -> don't allow to enter any items that are not in the Model -> fires ActionEvent only when current Value is in Model
    private boolean restrictValues = false;
    
    // makes only sense if restrictValues=false; adds current value to Model if it's new (model have to be sorted afterwards)
    private boolean autoAddValue = false;
    
    // makes only sense if restrictValues=true; if true, skips characters to next undecideable 
    private boolean maximumMatch = false;

	// if true, shows popup on any key pressed
	private boolean showPopupOnAnyKey = false;

    private boolean ignoreCase = true;
    
    // Binary Lookup & Performance (Prototype Value) is also implemented

	private Comparator comparator;
	private SimpleComboBoxModel simpleComboBoxModel;

	/**
     * @param items
     */
    public AutoCompleteComboBox(Collection items) {
        super(new SimpleComboBoxModel(items));
        init();
    }

    /**
     * @param items
     */
    public AutoCompleteComboBox(Object[] items) {
        super(new SimpleComboBoxModel(items));
        init();
    }

    /**
     * 
     */
    public AutoCompleteComboBox() {
        super(new SimpleComboBoxModel());
        init();
    }
    
    private void init() {
		comparator = (ignoreCase ? String.CASE_INSENSITIVE_ORDER : DefaultComparator.STRING_COMPARATOR);
		simpleComboBoxModel = (SimpleComboBoxModel) dataModel;
		setEditable(true);
		setActionCommand("comboBoxEdited");
		document = new AutoCompleteComboBoxDocument();

		// Bug 5100422 on Java 1.5: Editable JComboBox won't hide popup when tabbing out
		hidePopupOnFocusLoss = System.getProperty("java.version").startsWith("1.5");

		editorComp = (JTextField) getEditor().getEditorComponent();
		editorComp.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent evt) {
				if (autoAddValue)
					storeInput();
				AutoCompleteComboBox.super.fireActionEvent();
			}
		});
		editorComp.addKeyListener(new KeyAdapter() {
			public void keyPressed(KeyEvent evt) {
				if (showPopupOnAnyKey && Character.isLetterOrDigit(evt.getKeyChar()) && isDisplayable())
					setPopupVisible(true);
				hitBackspace = false;
				switch (evt.getKeyCode()) {
					// determine if the pressed key is backspace (needed by the remove method)
					case KeyEvent.VK_BACK_SPACE:
						hitBackspaceOnSelection = (editorComp.getSelectionStart() != editorComp.getSelectionEnd());
					case KeyEvent.VK_DELETE:
						hitBackspace = true;
						break;
						// ignore delete key
					/*case KeyEvent.VK_DELETE:
						e.consume();
						UIManager.getLookAndFeel().provideErrorFeedback(AutoCompleteComboBox.this);
					//	System.out.println("input has been received, but can not be accepted");
						break;*/
				}
			}
		});
		editorComp.addFocusListener(new FocusAdapter() {
			// Highlight whole text when gaining focus
			public void focusGained(FocusEvent e) {
				highlightCompletedText(0);
			}
			public void focusLost(FocusEvent e) {
				// Workaround for Bug 5100422 - Hide Popup on focus loss
				if (hidePopupOnFocusLoss && isDisplayable())
					setPopupVisible(false);
			}
		});
		editorComp.addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent evt) {
				if (isDisplayable()/* && !SwingUtilities.isLeftMouseButton(evt)*/)
					setPopupVisible(!isPopupVisible());
			}
		});
		editorComp.setDocument(document);

		setPrototypeValue();

		// Handle initially selected object
		Object selected = getSelectedItem();
		if (selected != null)
			document.setText(selected.toString());

		highlightCompletedText(0);
	}

	protected void fireActionEvent() { }

	private void highlightCompletedText(int start) {
		editorComp.setCaretPosition(document.getLength());
		editorComp.moveCaretPosition(start);
	}
    
    // the same as the "private void setSelectedItem(Object item)" method in AutoCompletion.java, but since I extends JComboBox and JComboBox also has a method with this signature ...
    private void _setSelectedItem(Object item) {
        selecting = true;
        dataModel.setSelectedItem(item);
        selecting = false;
    }
    
    private Object lookupItem(String pattern) {
        Object selectedItem = dataModel.getSelectedItem();
        
        // only search for a different item if the currently selected does not match
        if ((selectedItem != null) && startsWith(selectedItem.toString(), pattern, ignoreCase)
                && (restrictValues || listContainsSelectedItem)) 
            return selectedItem;

        return binaryLookup(pattern);
    }
    
    private Object binaryLookup(String pattern) {
        int bottom = 0, top = dataModel.getSize()-1;
        int pos = 0;
        Object item = null;
        
        // search for a matching item
        while (bottom <= top) {
            pos = (bottom + top) >> 1;
            item = dataModel.getElementAt(pos);
            int compare = compareStarts(item.toString(), pattern, ignoreCase);
            if (compare == 0) 
                break;
            if (compare > 0) 
                bottom = pos + 1;
            else 
                top = pos - 1;
        }
        
        // if no item matches bottom is greater than top
        if (bottom > top) 
            return null;
        
        // search for the _first_ matching item
        for (int i=bottom; i<pos; i++) {
            Object anItem = dataModel.getElementAt(i);
            if (startsWith(anItem.toString(), pattern, ignoreCase))
                return anItem;
        }
        return item;
    }

    public void configureEditor(ComboBoxEditor anEditor, Object anItem) {
        hitBackspace = false; // bugfix; reset "hitBackspace" on selection to correctly remove old value
        super.configureEditor(anEditor, anItem);
    }
    
    
    public void setPrototypeValue() {
        JList list = getListBox();
        if (list == null) return; // should not happen
        Object value = getPrototypeValue(list);
        setPrototypeDisplayValue(value);
        list.setPrototypeCellValue(value);
    }
    private JList getListBox() {
        try {
			Field field = BasicComboBoxUI.class.getDeclaredField("listBox");
            field.setAccessible(true);
            return (JList) field.get(getUI());
        } catch (Exception exc) {
            return null;
        }
    }
    private Object getPrototypeValue(JList list) {
        Object prototypeValue = null;
        double prototypeWidth = 0;
        ListCellRenderer renderer = getRenderer();
        for (int i=0, n=dataModel.getSize(); i<n; i++) {
            Object value = dataModel.getElementAt(i);
            java.awt.Component c = renderer.getListCellRendererComponent(list, value, i, false, false);
            double width = c.getPreferredSize().getWidth();
            if (width > prototypeWidth) {
                prototypeWidth = width;
                prototypeValue = value;
            }
        }
        return prototypeValue;
    }
    
    
    // setText() does not fire an ActionEvent in contrast to setSelectedItem()
    public void setText(String text) {
        document.setText(text);
        editorComp.setCaretPosition(document.getLength());
     //   editorComp.requestFocusInWindow();
    }
	public String getText() {
		return editorComp.getText();
	}

    // Getter / Setter
	public List getElements() {
		return simpleComboBoxModel.getElements();
	}

    /**
     * @return Returns the autoAddValue.
     */
    public boolean isAutoAddValue() {
        return autoAddValue;
    }
    /**
     * @param autoAddValue The autoAddValue to set.
     */
    public void setAutoAddValue(boolean autoAddValue) {
        this.autoAddValue = autoAddValue;
        restrictValues = false;
    }

    /**
     * @return Returns the maximumMatch.
     */
    public boolean isMaximumMatch() {
        return maximumMatch;
    }
    /**
     * @param maximumMatch The maximumMatch to set.
     */
    public void setMaximumMatch(boolean maximumMatch) {
        this.maximumMatch = maximumMatch;
        restrictValues = true;
    }
    /**
     * @return Returns the restrictValues.
     */
    public boolean isRestrictValues() {
        return restrictValues;
    }
    /**
     * @param restrictValues The restrictValues to set.
     */
    public void setRestrictValues(boolean restrictValues) {
        this.restrictValues = restrictValues;
        if (restrictValues) autoAddValue = false;
        else maximumMatch = false; 
    }

	public boolean isShowPopupOnAnyKey() {
		return showPopupOnAnyKey;
	}
	public void setShowPopupOnAnyKey(boolean showPopupOnAnyKey) {
		this.showPopupOnAnyKey = showPopupOnAnyKey;
	}

	public boolean isIgnoreCase() {
		return ignoreCase;
	}
	public void setIgnoreCase(boolean ignoreCase) {
		this.ignoreCase = ignoreCase;
		comparator = (ignoreCase ? String.CASE_INSENSITIVE_ORDER : DefaultComparator.STRING_COMPARATOR);
		List elements = simpleComboBoxModel.getElements();
		Collections.sort(elements, comparator);
	}

	public void storeInput() {
		addNewItem(getText());
	}
	public void addNewItem(Object newItem) {
		List elements = simpleComboBoxModel.getElements();
		int size = elements.size();
		if (size == 0) elements.add(newItem);
		else
			if (comparator.compare(newItem, elements.get(0)) < 0)
				elements.add(0, newItem);
			else if (comparator.compare(newItem, elements.get(size-1)) > 0)
				elements.add(newItem);
			else {
				int startPos = 0, endPos = size;
				while (startPos < endPos) {
					int pos = (startPos + endPos) >> 1;
					Object item = elements.get(pos);
					int diff = comparator.compare(newItem, item);
					if (diff == 0) return;
					if (diff < 0) endPos = pos - 1;
					else startPos = pos + 1;
				}
				Object item = elements.get(startPos);
				int diff = comparator.compare(newItem, item);
				if (diff == 0) return;
				if (diff < 0)
					elements.add(startPos, newItem);
				else
					elements.add(startPos+1, newItem);
			}
		setPrototypeValue();
	}

	// delegates to SimpleComboBoxModel
	public void init(Object[] items, Object selectedItem) {
		simpleComboBoxModel.init(items, selectedItem);
	}
	public void init(Collection items, Object selectedItem) {
		simpleComboBoxModel.init(items, selectedItem);
	}

	private class AutoCompleteComboBoxDocument extends PlainDocument {
        public void remove(int offs, int len) throws BadLocationException {
            // return immediately when selecting an item
            if (selecting) return;

            if (hitBackspace) {
            	if (restrictValues || listContainsSelectedItem) {
	                // user hit backspace => move the selection backwards
	                // old item keeps being selected
	                if (offs > 0)
	                    if (hitBackspaceOnSelection) 
	                        offs--;
						else {
							// User hit backspace with the cursor positioned on the start => beep
							UIManager.getLookAndFeel().provideErrorFeedback(AutoCompleteComboBox.this);
						//	System.out.println("input has been received, but can not be accepted");
						}
	                highlightCompletedText(offs);
					if (showPopupOnAnyKey && isDisplayable())
						setPopupVisible(offs > 0);
	                return;
            	}
            	
            	// in case of "Non-strict matching" and !listContainsSelectedItem
            	
            	super.remove(offs, len);
                String currentText = getText(0, getLength());
                // lookup if a matching item exists
                Object item = lookupItem(currentText);
                if (item != null) {
                //    System.out.println("'" + item + "' matches."); // debug output
                    setSelectedItem(item);
                    listContainsSelectedItem = true;
                } else {
                //    System.out.println("No match. Selecting '" + currentText + "'."); // debug output
                    // no item matches => use the current input as selected item
                    item = currentText;
                    setSelectedItem(item);
                    listContainsSelectedItem = false;
                }
                // display the completed string
				if (listContainsSelectedItem) {	// so the cursor is not moved if !listContainsSelectedItem and since the text stays the same
					setText(item.toString());
                    highlightCompletedText(offs);
				}
            } else {
                super.remove(offs, len);
                if (!restrictValues) {
                    setSelectedItem(getText(0, getLength()));
                	listContainsSelectedItem = false;
                }
            }       
        }
        
        public void insertString(int offs, String str, AttributeSet a) throws BadLocationException {
            // return immediately when selecting an item
            if (selecting) return;
            
            // insert the string into the document
            super.insertString(offs, str, a);
            
            // lookup and select a matching item
			String oldTextValue = getText(0, getLength());
			Object item = lookupItem(oldTextValue);
            
            boolean match=false;
            listContainsSelectedItem = true;
            
            if (item == null) 
                if (restrictValues) {
	                // keep old item selected if there is no match
	                item = getSelectedItem();
	                // imitate no insert (later on offs will be incremented by str.length(): selection won't move forward)
	                offs = offs-str.length();
	                // provide feedback to the user that his input has been received but can not be accepted
	                UIManager.getLookAndFeel().provideErrorFeedback(AutoCompleteComboBox.this);
				//	System.out.println("input has been received, but can not be accepted");
	            } else {
	                // no item matches => use the current input as selected item
	                item = oldTextValue;
	                listContainsSelectedItem = false;
	            }
            if (!restrictValues || (item != null)) { // in case of "Non-strict matching" always select item 
                match = true;
                _setSelectedItem(item);
            }
            
            if (maximumMatch && match) offs = getMaximumMatchingOffset(oldTextValue, item);
            else offs += str.length();

			if (listContainsSelectedItem) {	// so the cursor is not moved if !listContainsSelectedItem and since the text stays the same
				String newTextValue = item.toString();
				if (item instanceof String && ignoreCase) {
					int sameStartCount = equalStartLength(oldTextValue, newTextValue, true);
					newTextValue = oldTextValue.substring(0, sameStartCount) + newTextValue.substring(sameStartCount);
				}
				setText(newTextValue);
			}

            // select the completed part
            if (restrictValues || listContainsSelectedItem) { // in case of "Non-strict matching" only highligth if listContainsSelectedItem
                highlightCompletedText(offs);
				if (showPopupOnAnyKey && isDisplayable())
					setPopupVisible(offs > 0);
			}
        }
        
        private void setText(String text) {
            try {
                // remove all text and insert the completed string
                super.remove(0, getLength());
                super.insertString(0, text, null);
            } catch (BadLocationException e) {
                throw new RuntimeException(e.toString());
            }
        }
        
//      calculates how many characters are predetermined by the given pattern.
        private int getMaximumMatchingOffset(String pattern, Object selectedItem) {
            String selectedAsString = selectedItem.toString();
            int match = selectedAsString.length();
            // look for items that match the given pattern
            for (int i=0, n=dataModel.getSize(); i < n; i++) {
                String itemAsString = dataModel.getElementAt(i).toString();
                if (startsWith(itemAsString, pattern, ignoreCase)) {
                    // current item matches the pattern
                    // how many leading characters have the selected and the current item in common?
                    int tmpMatch = equalStartLength(itemAsString, selectedAsString, ignoreCase);
                    if (tmpMatch < match) match = tmpMatch;
                }
            }
            return match;
        }
    }


	// checks if str1 starts with str2 - ignores case
	public static boolean startsWith(String str1, String str2, boolean ignoreCase) {
		if (ignoreCase){
			str1 = str1.toUpperCase();
			str2 = str2.toUpperCase();
		}
		return str1.startsWith(str2);
	}


    /**
     * returns how many leading characters two strings have in common?
     * @param str1
	 * @return how many leading characters two strings have in common
	 * @param str2
	 * @param ignoreCase
	 */
    public static int equalStartLength(String str1, String str2, boolean ignoreCase) {    // is used by getMaximumMatchingOffset() for "Maximum Match"
        char[] ch1 = str1.toCharArray();
        char[] ch2 = str2.toCharArray();
        int n = (ch1.length > ch2.length) ? ch2.length : ch1.length;
        for (int i=0; i<n; i++) {
			char c1 = ch1[i], c2 = ch2[i];
			if (ignoreCase) {
				c1 = Character.toUpperCase(c1);
				c2 = Character.toUpperCase(c2);
			}
			if (c1 != c2) return i;
		}
        return n;
    }
    
    // similar to String#compareToIgnoreCase(), but works on char arrays and treats the case differently when one string starts with the other
    public static int compareStarts(String str1, String str2, boolean ignoreCase) {
        char[] ch1 = str1.toCharArray();
        char[] ch2 = str2.toCharArray();
        for (int i=0, n = (ch2.length < ch1.length) ? ch2.length : ch1.length; i<n; i++) {
            char c1 = ch2[i], c2 = ch1[i];
            if (c1 != c2)
				if (ignoreCase) {
					int diff = Character.toUpperCase(c1) - Character.toUpperCase(c2);
					if (diff != 0)
						return diff;
				} else
					return (c1 - c2);
        }

        if (ch1.length < ch2.length) return 1;
        return 0; // if the str1 startsWith str2 return also 0, e.g. str1="jordina",str2="j" 
    }
}
