/*
 * Project "Guideline Markup Tool"
 * Author: Peter
 * Date: Sep 22, 2002
 * Time: 3:51:48 AM
 */
package at.ac.tuwien.asgaard.swing.models;

import javax.swing.tree.*;
import java.util.*;

/**
 * Class SimpleMutableTreeNode. ... todo
 *
 * @version 2.7.3
 * @author Peter Votruba
 * @since JDK 1.3.1
 */
public class SimpleMutableTreeNode implements MutableTreeNode {
	/** this node's parent, or null if this node has no parent */
	protected SimpleMutableTreeNode parent;
	/** array of children, may be null if this node has no children */
	protected List<TreeNode> children;
	protected ListEnumeration childEnumeration;
	/** optional user object */
	protected transient Object userObject;

    protected SimpleMutableTreeNode() { }

    public SimpleMutableTreeNode(Object userObject) {
        this.userObject = userObject;
    }

    public void insert(MutableTreeNode newChild, int childIndex) {
		if (newChild == null)
			throw new IllegalArgumentException("new child is null");
		else if (isNodeAncestor(newChild))
			throw new IllegalArgumentException("new child is an ancestor");

		MutableTreeNode oldParent = (MutableTreeNode)newChild.getParent();
		if (oldParent != null)
			oldParent.remove(newChild);

		newChild.setParent(this);

		if (children == null) {
			children = new ArrayList<TreeNode>();
			childEnumeration = new ListEnumeration(children);
		}
        if (childIndex == children.size())
            children.add(newChild);
        else
		    children.add(childIndex, newChild);
	}

	public void remove(int childIndex) {
		if (children == null)
			throw new ArrayIndexOutOfBoundsException("node has no children");
		MutableTreeNode child = (MutableTreeNode) children.remove(childIndex);
		child.setParent(null);
	}

	public void remove(MutableTreeNode child) {
		if (child == null)
			throw new IllegalArgumentException("argument is null");
		if (child.getParent() != this)
			throw new IllegalArgumentException("argument is not a child");
		if (children == null)
			throw new ArrayIndexOutOfBoundsException("node has no children");
		children.remove(children.indexOf(child));
		child.setParent(null);
	}

	public void removeFromParent() {
		if (parent != null)
			parent.remove(this);
	}

	public TreeNode getParent() {
		return parent;
	}
    public SimpleMutableTreeNode getSimpleMutableTreeNodeParent() {
        return parent;
    }
	public void setParent(MutableTreeNode newParent) {
		parent = (SimpleMutableTreeNode)newParent;
	}
	public void setParent(SimpleMutableTreeNode newParent) {
		parent = newParent;
	}

	/**
	 * Returns this node's user object.
	 *
	 * @return the Object stored at this node by the user
	 * @see #setUserObject
	 * @see #toString
	 */
	public Object getUserObject() {
		return userObject;
	}
	public void setUserObject(Object userObject) {
		this.userObject = userObject;
	}

	public TreeNode getChildAt(int childIndex) {
		if (children == null)
			throw new ArrayIndexOutOfBoundsException("node has no children");
		return children.get(childIndex);
	}

	public int getChildCount() {
		if (children == null)
			return 0;
		 else
			return children.size();
	}

	public int getIndex(TreeNode aChild) {
		if (aChild == null)
			throw new IllegalArgumentException("argument is null");
		if (aChild.getParent() != this)
			return -1;
		return children.indexOf(aChild); // linear search
	}

	public boolean getAllowsChildren() {
		return true;
	}

	public boolean isLeaf() {
		if (children == null)
			return true;
		 else
			return (children.size() == 0);
	}

	public Enumeration<TreeNode> children() {
		if (children == null)
			return DefaultMutableTreeNode.EMPTY_ENUMERATION;
		else
			return childEnumeration.reset();
	}

	public List<TreeNode> getChildList() {
		return children;
	}

	/**
	 * Removes <code>newChild</code> from its parent and makes it a child of
	 * this node by adding it to the end of this node's child array.
	 *
	 * @param newChild node to add as a child of this node
	 * @exception java.lang.IllegalArgumentException if <code>newChild</code> is null
	 * @exception java.lang.IllegalStateException if this node does not allow children
	 */
	public void add(SimpleMutableTreeNode newChild) {
		if (newChild == null)
			throw new IllegalArgumentException("new child is null");
		else if (isNodeAncestor(newChild))
			throw new IllegalArgumentException("new child is an ancestor");

		MutableTreeNode oldParent = newChild.parent;
		if (oldParent != null)
			oldParent.remove(newChild);
		newChild.parent = this;

		if (children == null) {
			children = new ArrayList<TreeNode>();
			childEnumeration = new ListEnumeration(children);
		}
		children.add(newChild);
	}

	/**
	 * Returns true if <code>anotherNode</code> is an ancestor of this node
	 * -- if it is this node, this node's parent, or an ancestor of this
	 * node's parent.  (Note that a node is considered an ancestor of itself.)
	 * If <code>anotherNode</code> is null, this method returns false.  This
	 * operation is at worst O(h) where h is the distance from the root to
	 * this node.
	 *
	 * @param anotherNode node to test as an ancestor of this node
	 * @return true if this node is a descendant of <code>anotherNode</code>
	 */
	public boolean isNodeAncestor(TreeNode anotherNode) {
		if (anotherNode == null) return false;
		TreeNode ancestor = this;
		do {
			if (ancestor == anotherNode) return true;
		} while ((ancestor = ancestor.getParent()) != null);
		return false;
	}

    public void removeAll() {
		int childCount = getChildCount();
		if (childCount > 0) {
			for (int i = 0; i < childCount; i++)
				((MutableTreeNode) children.get(i)).setParent(null);
			children.clear();
		}
    }

    public void replaceChild(MutableTreeNode newChild, MutableTreeNode oldChild) {
        int index = children.indexOf(oldChild);
        remove(index);
        insert(newChild, index);
    }

	public boolean hasSiblings() {
		if (parent == null) return false;
		return (parent.getChildCount() > 1);
	}
	public boolean hasPreviousSiblings() {
		if (parent == null) return false;
		return getIndex() > 0;
	}
	public boolean hasNextSiblings() {
		if (parent == null) return false;
		return getIndex() < parent.getChildCount()-1;
	}
	// get the index of this node in the list of children of this node's parent
	public int getIndex() {
		if (parent == null) return -1;
		return parent.getChildList().indexOf(this);
	}

	public TreePath getTreePath() {
		if (parent == null) return new TreePath(this);
		return parent.getTreePath().pathByAddingChild(this);
	}
	public TreePath getTreePathTo(TreeNode toTreeNode) {
		if (this == toTreeNode) return new TreePath(this);
		return parent.getTreePathTo(toTreeNode).pathByAddingChild(this);
	}

	/**
	 * Returns the result of sending <code>toString()</code> to this node's
	 * user object, or null if this node has no user object.
	 *
	 * @see #getUserObject
	 */
	public String toString() {
		if (userObject == null) return super.toString();
		return userObject.toString();
	}

    protected static final class ListEnumeration implements Enumeration<TreeNode> {
		private int index;
        private List<TreeNode> children;

        ListEnumeration(List<TreeNode> children) {
            this.children = children;
        }
		public boolean hasMoreElements() {
			return (index < children.size());
		}
		public TreeNode nextElement() {
			return children.get(index++);
		}
		public Enumeration<TreeNode> reset() {
			index = 0;
			return this;
		}
	}
}
