package at.ac.tuwien.asgaard.swing.models;

import java.beans.PropertyVetoException;
import java.beans.VetoableChangeListener;
import java.beans.VetoableChangeSupport;
import java.util.ArrayList;
import java.util.List;

import javax.swing.*;
import javax.swing.event.*;

public class SingleListSelectionModel implements ListSelectionModel {
    public static final int RESELECTION_DISABLED = 3;
    public static final int RESELECTION_ENABLED  = 4;

	private int oldIndex, currentIndex, anchorSelectionIndex, leadSelectionIndex;
	private boolean allowReselection;
	private final List listenerList;
    private VetoableChangeSupport vetoableSupport;

    public SingleListSelectionModel() {
		this(false);
	}
	public SingleListSelectionModel(boolean allowReselection) {
		this.allowReselection = allowReselection;
		currentIndex = oldIndex = anchorSelectionIndex = leadSelectionIndex = -1; // start with "empty" selection
		listenerList = new ArrayList(4);
	}

	public int getCurrentIndex() {
		return currentIndex;
	}

	public int getPreviousIndex() {
		return oldIndex;
	}

	public boolean selectionChanged() {
		return (currentIndex != oldIndex);
	}

	public void setSelectionIndex(int newSelectionIndex) {
		anchorSelectionIndex = newSelectionIndex;
		leadSelectionIndex = newSelectionIndex;
		fireSelectionChanged(newSelectionIndex);
	}
	public void selectPreviousIndex() {
		anchorSelectionIndex = oldIndex;
		leadSelectionIndex = oldIndex;
		fireSelectionChanged(oldIndex);
	}
	public void forceReselection() {
		ListSelectionEvent evt;
		if (currentIndex == -1)
			evt = new ListSelectionEvent(this, oldIndex, oldIndex, false);
		else
			evt = new ListSelectionEvent(this, Math.min(currentIndex, oldIndex), Math.max(currentIndex, oldIndex), false);
		for (int i = listenerList.size()-1; i >= 0; i--)
			((ListSelectionListener) listenerList.get(i)).valueChanged(evt);
	}

	private void fireSelectionChanged(int newIndex) {
		if (allowReselection || (newIndex != currentIndex)) {
            if (vetoableSupport != null)
                try {
                    vetoableSupport.fireVetoableChange("selectionIndex", currentIndex, newIndex);
                } catch (PropertyVetoException e) {
                    return; // cancel selection change
                }

            oldIndex = currentIndex;
			currentIndex = newIndex;

			ListSelectionEvent evt;
			if (newIndex == -1)
				evt = new ListSelectionEvent(this, oldIndex, oldIndex, false);
			else
				evt = new ListSelectionEvent(this, Math.min(currentIndex, oldIndex), Math.max(currentIndex, oldIndex), false);
			for (int i = listenerList.size()-1; i >= 0; i--)
				((ListSelectionListener) listenerList.get(i)).valueChanged(evt);
		}
	}

	public void addSelectionInterval(int index0, int index1) {
		anchorSelectionIndex = index0;
		leadSelectionIndex = index1;
		fireSelectionChanged(index0);
	}

	public void clearSelection() {
		setSelectionIndex(-1);
	}

	public int getAnchorSelectionIndex() {
		return anchorSelectionIndex;
	}

	public int getLeadSelectionIndex() {
		return leadSelectionIndex;
	}

	public int getMaxSelectionIndex() {
		return currentIndex;
	}

	public int getMinSelectionIndex() {
		return currentIndex;
	}

	public int getSelectionMode() {
		return SINGLE_SELECTION;
	}

	public boolean getValueIsAdjusting() {
		return false;
	}

	public void insertIndexInterval(int index, int length, boolean before) {
		int startIndex = (before ? index : index+1);
		if (currentIndex >= startIndex)
			fireSelectionChanged(currentIndex + length);
	}

	public boolean isSelectedIndex(int index) {
		return (currentIndex == index);
	}

	public boolean isSelectionEmpty() {
		return (currentIndex == -1);
	}

	public void removeIndexInterval(int index0, int index1) {
		int rmMinIndex = Math.min(index0, index1);
		int rmMaxIndex = Math.max(index0, index1);
		int gapLength = (rmMaxIndex - rmMinIndex) + 1;
		if (currentIndex >= rmMinIndex)
			if (currentIndex <= rmMaxIndex)
				fireSelectionChanged(-1);
			else
				fireSelectionChanged(currentIndex-gapLength);
	}

	public void removeSelectionInterval(int index0, int index1) {
		anchorSelectionIndex = index0;
		leadSelectionIndex = index1;
		int rmMinIndex = Math.min(index0, index1);
		int rmMaxIndex = Math.max(index0, index1);
		if ((currentIndex >= rmMinIndex) && (currentIndex <= rmMaxIndex))
			fireSelectionChanged(-1);
	}

	public void setAnchorSelectionIndex(int index) {
		anchorSelectionIndex = index;
		fireSelectionChanged(index);
	}

	public void setLeadSelectionIndex(int index) {
		leadSelectionIndex = index;
		fireSelectionChanged(index);
	}

	public void setSelectionInterval(int index0, int index1) {
		anchorSelectionIndex = index0;
		leadSelectionIndex = index1;
		fireSelectionChanged(index0);
	}


	public void addListSelectionListener(ListSelectionListener listener) {
		listenerList.add(listener);
	}
	public void removeListSelectionListener(ListSelectionListener listener) {
		listenerList.remove(listener);
	}

    public void addVetoableChangeListener(VetoableChangeListener listener) {
        if (vetoableSupport == null)
            vetoableSupport = new VetoableChangeSupport(this);
        vetoableSupport.addVetoableChangeListener(listener);
    }
    public void removeVetoableChangeListener(VetoableChangeListener listener) {
        if (vetoableSupport != null)
            vetoableSupport.removeVetoableChangeListener(listener);
    }

    public boolean isAllowReselection() {
        return allowReselection;
    }

    public void setAllowReselection(boolean allowReselection) {
        this.allowReselection = allowReselection;
    }

	public void setSelectionMode(int selectionMode) {
        if (selectionMode == RESELECTION_ENABLED) allowReselection = true;
        else if (selectionMode == RESELECTION_DISABLED) allowReselection = false;
    }

	public void reselect() {
		allowReselection = true;
		fireSelectionChanged(currentIndex);
		allowReselection = false;
	}

	public void setValueIsAdjusting(boolean valueIsAdjusting) { }
}

