/*
 * Project "Guideline Markup Tool"
 * Author: Peter
 * Date: Sep 22, 2002
 * Time: 2:05:00 PM
 */
package at.ac.tuwien.asgaard.swing.models;

import javax.swing.event.EventListenerList;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.RowMapper;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;
import java.beans.PropertyChangeListener;

/**
 * Class SingleTreeSelectionModel. ... todo
 *
 * @version 2.7.3
 * @author Peter Votruba
 * @since JDK 1.3.1
 */
public class SingleTreeSelectionModel implements TreeSelectionModel {
	public static final int RESELECTION_DISABLED = 0;
	public static final int RESELECTION_TEMPORARILY_ENABLED = 1;

	private int reselectionMode;

	private int selectedRow;
	private TreePath selectedPath;
	private RowMapper rowMapper;
	private final EventListenerList listenerList;

	private boolean isCurrentlySelecting;

	public SingleTreeSelectionModel() {
		this(RESELECTION_DISABLED);
	}

	public SingleTreeSelectionModel(int reselectionMode) {
		this.reselectionMode = reselectionMode;
		selectedRow = -1;
		isCurrentlySelecting = false;
		listenerList = new EventListenerList();
	}

    public void reselect() {
        setSelectionMode(RESELECTION_TEMPORARILY_ENABLED);
        setSelectionPath(selectedPath);
    }

	private void changeSelection(TreePath newSelection) {
		if ((reselectionMode == RESELECTION_DISABLED) &&
				((selectedPath != null) && (newSelection != null) &&
								(selectedPath.getLastPathComponent() == newSelection.getLastPathComponent()))
                || ((selectedPath == null) && (newSelection == null)))
			return;
		isCurrentlySelecting = true;
		reselectionMode = RESELECTION_DISABLED; // disable re-selection again to avoid hang-up

		TreePath oldSelection = selectedPath;
		selectedPath = newSelection;
		resetRowSelection();

		Object[] listeners = listenerList.getListenerList();
		if (listeners.length > 0) {
			//TreeSelectionEvent evt = new TreeSelectionEvent(this, newSelection, true, oldSelection, newSelection);
            TreeSelectionEvent evt = new TreeSelectionEvent(this, new TreePath[]{newSelection, oldSelection},
                    new boolean[]{true, false}, oldSelection, newSelection);
			for (int i = listeners.length - 2; i >= 0; i -= 2)
                ((TreeSelectionListener) listeners[i + 1]).valueChanged(evt);
		}
		isCurrentlySelecting = false;
	}

	public boolean isCurrentlySelecting() {
		return isCurrentlySelecting;
	}

	public void setSelectionMode(int reselectionMode) {
		this.reselectionMode = reselectionMode;
	}

	public int getSelectionMode() {
		return TreeSelectionModel.SINGLE_TREE_SELECTION;
	}

	public void setSelectionPath(TreePath path) {
		changeSelection(path);
	}

	public void setSelectionPaths(TreePath[] paths) {
		changeSelection((paths == null) ? null : paths[0]);
	}

	public void addSelectionPath(TreePath path) {
		changeSelection(path);
	}

	public void addSelectionPaths(TreePath[] paths) {
		changeSelection(((paths == null) || (paths.length == 0)) ? null : paths[0]);
	}

	public void removeSelectionPath(TreePath path) {
		if ((selectedPath != null) && (path != null) && path.equals(selectedPath))
			changeSelection(null);
	}

	public void removeSelectionPaths(TreePath[] paths) {
		if ((paths != null) && (paths.length != 0))
			removeSelectionPath(paths[0]);
	}

	public TreePath getSelectionPath() {
		return selectedPath;
	}

	public TreePath[] getSelectionPaths() {
		return ((selectedPath == null) ? null : new TreePath[]{selectedPath});
	}

	public int getSelectionCount() {
		return ((selectedPath == null) ? 0 : 1);
	}

	public boolean isPathSelected(TreePath path) {
		return ((selectedPath != null) && (path != null) && selectedPath.equals(path));
	}

	public boolean isSelectionEmpty() {
		return (selectedPath == null);
	}

	public void clearSelection() {
		changeSelection(null);
	}

	public void setRowMapper(RowMapper newMapper) {
		rowMapper = newMapper;
		resetRowSelection();
	}

	public RowMapper getRowMapper() {
		return rowMapper;
	}

	public int[] getSelectionRows() {
		if (rowMapper != null && selectedPath != null) {
			int[] rows = rowMapper.getRowsForPaths(new TreePath[] {selectedPath});
			if ((rows != null) && (rows.length > 0) && (rows[0] == -1))
				return null;
			return rows;
		}
		return null;
	}

	public int getMinSelectionRow() {
		return selectedRow;
	}

	public int getMaxSelectionRow() {
		return selectedRow;
	}

	public boolean isRowSelected(int row) {
		return (selectedRow == row);
	}

	public void resetRowSelection() {
		if (selectedPath != null && rowMapper != null) {
			int[] rows = rowMapper.getRowsForPaths(new TreePath[]{selectedPath});
			if ((rows != null) && (rows.length > 0))
				selectedRow = rows[0];
			else
				selectedRow = -1;
		} else
			selectedRow = -1;
	}

	public int getLeadSelectionRow() {
		return selectedRow;
	}

	public TreePath getLeadSelectionPath() {
		return selectedPath;
	}

	public void addTreeSelectionListener(TreeSelectionListener listener) {
		listenerList.add(TreeSelectionListener.class, listener);
	}

	public void removeTreeSelectionListener(TreeSelectionListener listener) {
		listenerList.remove(TreeSelectionListener.class, listener);
	}

	// does nothing
	public void addPropertyChangeListener(PropertyChangeListener listener) { }

	// does nothing
	public void removePropertyChangeListener(PropertyChangeListener listener) { }
}
