package at.ac.tuwien.asgaard.util;

import at.ac.tuwien.asgaard.delta.util.EnvironmentProperties;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.text.*;
import java.util.Date;

/**
 * Class FileUtils. ... todo
 *
 * @version 2.7.3
 * @author Peter Votruba
 */ 
public final class FileUtils {
	public static final String BACKUP_EXTENSION = ".bak";
	public static final String currentDir;
	public static final File currentDirFile;
// public static final File defaultBackupDirFile  = new File(currentDir, "backup");

	static {
		if (EnvironmentProperties.isRunUnderWebStart())
			currentDir = System.getProperty("user.home");
		else
			currentDir = System.getProperty("user.dir");  // new File(".").getAbsolutePath();
		currentDirFile = new File(currentDir);
	}

	/** this class shouldn't be instantiated */
	private FileUtils() { }

	public static String correctFileSeparator(String path) {
		char other = (File.separatorChar == '/') ? '\\' : '/';
		return path.replace(other, File.separatorChar);
	}

	public static boolean backupExists(File file) {
		return ((file != null) && new File(file.getAbsolutePath() + FileUtils.BACKUP_EXTENSION).exists());
	}

	public static boolean revertToBackup(File currentFile) {
		try {
			File bakFile = new File(currentFile.getAbsolutePath() + FileUtils.BACKUP_EXTENSION),
				tempFile = new File(currentFile.getAbsolutePath() + ".tmp");
			if (tempFile.exists()) tempFile.delete();
			return (currentFile.renameTo(tempFile) &&
			        bakFile.renameTo(currentFile) &&
			        tempFile.renameTo(bakFile));
		} catch (Throwable t) {
			Log.error("Couldn't revert \""+currentFile+"\" to backup: "+t);
			return false;
		}
	}

	public static void createBackup(File file) {
		File backupFile = new File(file.getAbsolutePath() + BACKUP_EXTENSION);
		try {
			copyFileToFile(file, backupFile);
		} catch (IOException exc) {
			try {
				backupFile.delete();  // if backup failed, delete any old backup
			} catch (Throwable t) { }
			Log.error("Couldn't create Backup \""+backupFile+"\" of \""+file+"\": "+exc);
		}
	}
	
	private static Format backupFileNameFormat = new MessageFormat("{0}_{1}-{2}_{3}.{4}");
	private static final Format DATE_FORMAT = DateFormat.getDateInstance(DateFormat.SHORT);
	private static final Format TIME_FORMAT = DateFormat.getTimeInstance(DateFormat.SHORT);
	
	/**
	 * Copies <code>file</code> to <code>backupDir</code> using following pattern 
	 * for the filename "${old_filename}_${date}-${time}_${username}" retaining the old file extension.
	 * @param file file to backup
	 * @param backupDir directory where new the new file will be put. if this is null the parent directory of the old file will be used
	 * @param userName user's name is used in the pattern of the new file name 
	 */
	public static void createAdvancedBackup(File file, File backupDir, String userName) throws IOException {
		if (backupDir == null) backupDir = file.getParentFile();
		String oldFileName = getFileName(file);
		String oldFileExt = getFileExt(file);
		Date now = new Date();
		String date = DATE_FORMAT.format(now);
		String time = TIME_FORMAT.format(now);
		String newFileName = backupFileNameFormat.format(new Object[] {oldFileName, date, time, userName, oldFileExt});
		File newFile = new File(backupDir, newFileName);
		copyFileToFile(file, newFile);
	}

	public static void copyFileToFile(File source, File target) throws IOException {
		writeInputStreamToFile(new FileInputStream(source), target);
	}

	public static void writeInputStreamToFile(InputStream source, File target) throws IOException {
		InputStream in = null;
		OutputStream out = null;
		try {
			in = new BufferedInputStream(source);
			out = new BufferedOutputStream(new FileOutputStream(target));
			while (true) {
				int data = in.read();
				if (data == -1)
					break;
				out.write(data);
			}
		} finally {
			if (in != null)
				in.close();
			if (out != null)
				out.close();
		}
	}

	public static void pipeReaderToWriter(Reader in, Writer out) throws IOException {
		int data = in.read();
		while (data != -1) {
			out.write(data);
			data = in.read();
		}
		in.close();
		out.flush();
		out.close();
	}

	public static String getDirPart(String filePath) {
		if (filePath == null) return null;
		filePath = correctFileSeparator(filePath);
		int lastSepPos = filePath.lastIndexOf(File.separatorChar);
		return filePath.substring(0, lastSepPos + 1);
	}
	public static File getDir(String filePath) {
		if (filePath == null) return null;
		filePath = correctFileSeparator(filePath);
		File dir = new File(filePath).getParentFile();
		return (dir!=null ? dir : new File(getDirPart(filePath)));
	}

	public static String getFileNameExt(String filePath) {
		if (filePath == null) return null;
		filePath = correctFileSeparator(filePath);
		int lastSepPos = filePath.lastIndexOf(File.separatorChar);
		if ((lastSepPos == -1) && (File.separatorChar != '/')) lastSepPos = filePath.lastIndexOf('/');
		if ((lastSepPos == -1)) lastSepPos = filePath.lastIndexOf(':');
		return filePath.substring(lastSepPos + 1);
	}
	public static String getFileName(String filePath) {
		String fileNameExt = getFileNameExt(filePath);
		int dotPos = fileNameExt.lastIndexOf('.');
		if (dotPos < 1) return fileNameExt;
		return fileNameExt.substring(0, dotPos);
	}
	public static String getFileName(File file) {
		return getFileName(file.getPath());
	}
	// returns .ext, e.g. ".xml"
	public static String getFileExt(File file) {
		return getFileExt(file.getName());
	}

	public static String getFileExt(String fileName) {
		int dotPos = fileName.lastIndexOf('.');
		if ((dotPos == -1) || (dotPos == fileName.length()-1)) return "";
		return fileName.substring(dotPos);
	}

	public static String getAbsolutePath(File file) {
		String filePath;
		try {
			filePath = file.getCanonicalPath(); // resolves xxx/../xxx/yyy.zzz to xxx/yyy.zzz
		} catch (IOException exc) {
			filePath = file.getAbsolutePath();
		}
		return filePath;
	}

	public static File getAbsoluteFile(File file) {
		try {
			file = file.getCanonicalFile();
		} catch (IOException exc) {
			file = file.getAbsoluteFile();
		}
		return file;
	}

	public static String toRelativeFile(File file) {
		return toRelativeFile(file, currentDirFile);
	}

	public static String toRelativeFile(File file, File relativeTo) {
		if (file == null) return null;
		String filePath = getAbsolutePath(file);
		return toRelativeFile(filePath, relativeTo);
	}

	public static String toRelativeFile(String filePath, File relativeTo) {
		if ((filePath == null) || filePath.equals("")) return null;
		filePath = correctFileSeparator(filePath);
		if ((relativeTo == null) || !relativeTo.isAbsolute()) return filePath;
		if (!relativeTo.isDirectory()) relativeTo = relativeTo.getParentFile();
		String localDir = relativeTo.getAbsolutePath() + File.separatorChar;

		String result = "";
		int pos = localDir.length()-1;
		while (pos > -1) {
			if (filePath.startsWith(localDir.substring(0, pos+1)))
				return result + filePath.substring(pos + 1);
			pos = localDir.lastIndexOf(File.separatorChar, pos-1);
			result = result + ".." + File.separatorChar;
		}
		return filePath;
	}

	public static File toAbsoluteFile(String filePath) {
		return toAbsoluteFile(filePath, currentDirFile);
	}

	public static File toAbsoluteFile(String filePath, File relativeTo) {
		if (filePath == null)
			return null;
		filePath = correctFileSeparator(filePath);
		filePath = filePath.trim();
		if (filePath.equals("")) return null;
		File file = new File(filePath);
		if (!file.isAbsolute()) {
			if (relativeTo == null) return null;
			if (!relativeTo.isDirectory())
				relativeTo = relativeTo.getParentFile();
			file = new File(relativeTo, filePath);
			file = getAbsoluteFile(file);
		}
		return file;
	}

	public static File toExistingAbsoluteFile(String filePath, File relativeTo) {
		File file = toAbsoluteFile(filePath, relativeTo);
		if ((file != null) && !file.exists()) return null;
		return file;
	}

	public static String toAbsoluteFilePath(String fileOrURL, File relativeTo) {
		if ((fileOrURL == null) || (fileOrURL.equals(""))) return null; 
			try {
				return new URL(fileOrURL).toString();
			} catch (MalformedURLException exc) {
				File file = toAbsoluteFile(fileOrURL, relativeTo);
				if (file == null) return null;
				return getAbsolutePath(file);
			}
	}

	/**
	 *
	 * @param fileOrURL either a relative path or absolute path or URL to a file
	 * @param relativeTo
	 * @return URL object representing this file
	 */
	public static URL toURL(String fileOrURL, File relativeTo) {
        if (fileOrURL == null) return null;
		try {
			return new URL(fileOrURL);
		} catch (MalformedURLException exc) {
			try {
				return toAbsoluteFile(fileOrURL, relativeTo).toURI().toURL();
			} catch (Exception e1) {
				return null; // shouldn't happen
			}
		}
	}
	public static URL toExistingURL(String fileOrURL, File relativeTo) {
		URL url = toURL(fileOrURL, relativeTo);
		try {
			InputStream in = url.openStream();
			if (in.read() == -1) { in.close(); return null; }
			in.close();
			return url;
		} catch (Exception exc) {
			return null;
		}
	}

	public static File getTemporaryFile(File dir, String prefix, String suffix) {
		int i = 1;
		File file;
		do {
			file = new File(dir, prefix + (i++) + suffix);
		} while (file.exists());
		return file;
	}
	public static File getTemporaryFile(String dir, String prefix, String suffix) {
		dir = correctFileSeparator(dir);
		int i = 1;
		File file;
		do {
			file = new File(dir, prefix + (i++) + suffix);
		} while (file.exists());
		return file;
	}

	public static boolean isURL(String url) {
		try {
			new URL(url);
			return true;
		} catch (MalformedURLException exc) {
			return false;
		}
	}

	public static String urlifyFile(File file) {
        try {   // new URL(file.toURL(), file.getName()).toExternalForm()
            return file.toURI().toURL().toExternalForm();
        } catch (MalformedURLException exc) {
            return file.getPath();
        }
    }

	public static String urlifyFile(String fileName, File baseFile) {
        // fileName = fileName.substring(fileName.lastIndexOf('/') + 1);
        try {
            URL u = baseFile.toURI().toURL();
            return new URL(u, correctFileSeparator(fileName)).toExternalForm();
        } catch (MalformedURLException exc) {
            return urlifyFile(new File(baseFile, correctFileSeparator(fileName)));
        }
    }
}
