/**
 * Copyright (c) 2003 Markus Krtzig. All Rights Reserved.
 *
 * Markus Krtzig makes no representations or warranties about
 * the fitness of this software for any particular purpose,
 * including the implied warranty of merchantability.
 */
package at.ac.tuwien.asgaard.util;

import java.awt.*;
import java.awt.event.ActionEvent;
import java.net.URL;

import javax.help.*;
import javax.swing.*;

/**
 * This class provides an <code>AbstractAction</code>
 * that is geared towards supporting the creation and display of
 * <code>JavaHelp</code> sets. It should be constructed via the
 * <code>get*Instance</code> static factory methods that create
 * instances of this class with slightly different behaviour. 
 * <p>
 * <b>Usage note:</b><br>
 * Because usually there is only one helpset loaded in an application, 
 * the <code>HelpBroker</code> is <code>static</code> and needs to be 
 * constructed first. Therefore the <code>startHelpsetLoader</code>
 * method must be invoked before any help action can be carried out. 
 * It takes the name of the helpset as an argument and constructs the 
 * <code>HelpBroker</code> in a worker thread. This should be done at
 * program start.
 * 
 *
 * @author M. Kraetzig
 */
public class JHelpUtils {
	private static String helpSetName = null;
	private static final ActionEvent cshAction = new ActionEvent(new JLabel(), ActionEvent.ACTION_FIRST, null);
	private static HelpBroker helpBroker = null;
	private static HelpsetLoader worker = null;

	private static class HelpsetLoader extends at.ac.tuwien.asgaard.swing.SwingWorker {
		private static Point loc = new Point(220, 0);
		private static Dimension dimension = new Dimension(700, 600);
		private final String helpSetName;
		private HelpBroker mainHelpBroker = null;

		public HelpsetLoader(String helpSetName) {
			this.helpSetName = helpSetName;
		}

		public synchronized Object construct() {
			if (mainHelpBroker == null) {
				try {
					URL hsURL = HelpSet.findHelpSet(HelpsetLoader.this.getClass().getClassLoader(), helpSetName);
					HelpSet mainHelpSet = new HelpSet(null, hsURL);
					mainHelpBroker = mainHelpSet.createHelpBroker();
				} catch (Throwable ee) {
					System.out.println("\nHelpSet " + helpSetName + " not created.\n  " + ee);
					return null;
				}
				mainHelpBroker.setLocation(loc);
				mainHelpBroker.setSize(dimension);
			}
			return mainHelpBroker;
		}
	}


	/**
	 * Gets instance of broker object and ensures that it is
	 * initialized.
	 *
	 * @return help broker object or <code>null</code> if it could not be
	 *         initialized
	 * @throws IllegalStateException if <code>HelpBroker</code> was not initialized via
	 *         <code>startHelpsetLoader</code> before
	 */
	public static HelpBroker getHelpBroker() {
		if (helpSetName == null)
			throw new IllegalStateException("Name of helpset has not yet been set.");

		if (helpBroker == null) {
			Object o = worker.get();
			if (o instanceof HelpBroker)
				helpBroker = (HelpBroker) o;
		}
		return helpBroker;
	}


	/**
	 * Shows the helpset that has been initialized via <code>startHelpsetLoader</code>.
	 *
	 * @throws IllegalStateException if <code>HelpBroker</code> was not initialized via
	 *         <code>startHelpsetLoader</code> before
	 */
	public static void showHelp() {
		if (helpBroker == null) {
			if (getHelpBroker() != null)
				new CSH.DisplayHelpFromSource(helpBroker).actionPerformed(cshAction);
		} else
			helpBroker.setDisplayed(true);
	}

	/**
	 * Shows the topic associated with <code>target</code> in the helpset that has
	 * been initialized via <code>startHelpsetLoader</code>.
	 *
	 * @param target the help ID to display
	 *
	 * @throws IllegalStateException if <code>HelpBroker</code> was not initialized via
	 *         <code>startHelpsetLoader</code> before
	 */
	public static void showHelp(String target) {
		showHelp();
		if (target != null)
			try {
				helpBroker.setCurrentID(Map.ID.create(target, helpBroker.getHelpSet()));
			} catch (InvalidHelpSetContextException ee) {
				System.out.println(ee.getMessage());
			}
	}
	public static void showHelpHome() {
		showHelp();
		try {
			helpBroker.setCurrentID(helpBroker.getHelpSet().getHomeID());
		} catch (InvalidHelpSetContextException ee) {
	//  System.out.println(ee.getMessage());   // tothink
		}
	}
	public static void showHelpOutOfModalDialog(String target, Object source) {
		if (helpBroker == null) {
			if (getHelpBroker() != null)
				new CSH.DisplayHelpFromSource(helpBroker).actionPerformed(new ActionEvent(source, ActionEvent.ACTION_FIRST, null));
		} else {
			if (helpBroker instanceof DefaultHelpBroker)
				((DefaultHelpBroker) helpBroker).setActivationObject(source);
			helpBroker.setDisplayed(true);
		}
		if (target != null)
			try {
				helpBroker.setCurrentID(Map.ID.create(target, helpBroker.getHelpSet()));
			} catch (Exception ee) {
	//  System.out.println(ee.getMessage());  // tothink
			}
	}

	/**
	 * Shows the topic associated with the component that currently has focus
	 * in the helpset that has  been initialized via <code>startHelpsetLoader</code>.
	 * If the component does not have a valid help ID, the default topic is displayed.
	 *
	 * @throws IllegalStateException if <code>HelpBroker</code> was not initialized via
	 *         <code>startHelpsetLoader</code> before
	 */
	public static void showHelpFromFocus() {
		if (getHelpBroker() != null)
			new CSH.DisplayHelpFromFocus(helpBroker).actionPerformed(cshAction);
	}

	/**
	 * Creates a worker thread to initialize a <code>HelpBroker</code>
	 * object for a helpset specified by <code>hSetName</code>. This method
	 * should be called at program start to prepare the helpset for display.
	 * Usually it should also only called once. All other methods rely on the
	 * availability of the <code>HelpBroker</code> object created by
	 * this method.
	 *
	 * @param hSetName the name of the helpset to statically initialize
	 *                 <code>JHelpUtils</code> with,
	 *                 must be a valid filename that can be transformed to a
	 *                 <code>URL</code>, usually stated relative to the
	 *                 resource directory of this class
	 */
	public static void startHelpsetLoader(String hSetName) {
		helpSetName = hSetName;
		worker = new HelpsetLoader(helpSetName);
		worker.start();
	}

	/**
	 * Shows the topic associated with the component that is clicked on with the mouse
	 * in the helpset that has  been initialized via <code>startHelpsetLoader</code>.
	 * If the component does not have a valid help ID, the default topic is displayed.
	 * This operation is useful for implementing context sensitive help.
	 *
	 * @throws IllegalStateException if <code>HelpBroker</code> was not initialized via
	 *         <code>startHelpsetLoader</code> before
	 */
	public static void trackFieldHelp() {
		if (getHelpBroker() != null)
			new CSH.DisplayHelpAfterTracking(helpBroker).actionPerformed(cshAction);
	}
}