package at.ac.tuwien.asgaard.util.xml;

import at.ac.tuwien.asgaard.util.FileUtils;
import at.ac.tuwien.asgaard.util.Log;

import net.sf.saxon.FeatureKeys;
import net.sf.saxon.TransformerFactoryImpl;
import net.sf.saxon.event.Builder;
import net.sf.saxon.event.SaxonOutputKeys;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.xpath.XPathEvaluator;
import net.socialchange.doctype.Doctype;
import net.socialchange.doctype.DoctypeChangerReader;
import net.socialchange.doctype.DoctypeGenerator;
import org.apache.avalon.framework.logger.ConsoleLogger;
import org.apache.avalon.framework.logger.Logger;
import org.apache.fop.apps.Driver;
import org.apache.fop.apps.FOPException;
import org.apache.fop.messaging.MessageHandler;
import org.apache.fop.render.awt.AWTRenderer;
import org.apache.fop.viewer.PreviewDialog;
import org.apache.fop.viewer.SecureResourceBundle;
import org.apache.fop.viewer.Translator;
import org.apache.fop.viewer.UserMessage;
import org.apache.xerces.dom.DOMImplementationImpl;
import org.apache.xerces.dom.DocumentImpl;
import org.apache.xerces.impl.Constants;
import org.apache.xerces.parsers.DOMParser;
import org.apache.xerces.parsers.DTDConfiguration;
import org.apache.xerces.parsers.NonValidatingConfiguration;
import org.apache.xerces.parsers.SAXParser;
import org.apache.xml.serialize.Method;
import org.apache.xml.serialize.OutputFormat;
import org.apache.xml.serialize.XMLSerializer;
import org.w3c.dom.*;
import org.xml.sax.*;

import java.io.*;
import java.net.URL;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;
import java.util.regex.Pattern;

import javax.swing.*;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.transform.*;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.sax.SAXResult;
import javax.xml.transform.sax.SAXSource;
import javax.xml.transform.sax.SAXTransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

/**
 * Class XMLUtils provides static XML utilitiy methods.
 *
 * @version 2.7.3
 * @author Peter Votruba
 * @since JDK 1.4.0
 */
public final class XMLUtils {
	private static XMLSerializer serializer;
	private static OutputFormat format;
	private static SAXTransformerFactory transformerFactory;
	private static Transformer xml2htmlTransformer, dtdToHTMLSrcTransformer, dtdToHTMLDocTransformer;
	private static XMLReader htmlParser;
	private static DOMImplementation domImpl;
	private static DOMParser domParser, nonValidatingdomParser;

	public static final String HTML_ERROR_MSG = "<html><body>Couldn't transform source code to HTML!</body></html>";
	public static final String[] XPATH_NODE_NAMES = {null, "", null, "text()", null, null, null, "processing-instruction()", "comment()", null, null, null, null};

	/** this class shouldn't be instantiated */
	private XMLUtils() {  }

	public static Document createNewDocument() {
		return new DocumentImpl();
	}
	public static Document createNewDocument(String systemID, String rootElement) throws DOMException {
		if (domImpl == null) domImpl = DOMImplementationImpl.getDOMImplementation();
		DocumentType dt = domImpl.createDocumentType(rootElement, null, systemID);
		return domImpl.createDocument(null, rootElement, dt);
	}
	public static Document createNewDocument(String publicID, String systemID, String rootElement) throws DOMException {
		if (domImpl == null) domImpl = DOMImplementationImpl.getDOMImplementation();
		DocumentType dt = domImpl.createDocumentType(rootElement, publicID, systemID);
		return domImpl.createDocument(null, rootElement, dt);
	}
	public static DOMParser getValidatingDOMParser() {
		if (domParser == null) {
			domParser = new DOMParser(new DTDConfiguration());
			try {
			// domParser.setProperty(Constants.XERCES_PROPERTY_PREFIX + Constants.DOCUMENT_CLASS_NAME_PROPERTY, "org.apache.xerces.dom.DocumentImpl");
				domParser.setFeature(Constants.SAX_FEATURE_PREFIX + Constants.NAMESPACES_FEATURE, false); // we dont use namespaces
				domParser.setFeature(Constants.XERCES_FEATURE_PREFIX + Constants.INCLUDE_IGNORABLE_WHITESPACE, false); // we dont need "ignorable whitespace"
				domParser.setFeature(Constants.SAX_FEATURE_PREFIX + Constants.VALIDATION_FEATURE, true);
				domParser.setErrorHandler(new ErrorHandler() {
					public void warning(SAXParseException exc)
							throws SAXException {
						throw new SAXException("Parser Warning: " + exc.getColumnNumber() + "/" + exc.getLineNumber() + " - " + exc.toString());
					}
					public void error(SAXParseException exc)
							throws SAXException {
						throw new SAXException("Parser Error: " + exc.getColumnNumber() + "/" + exc.getLineNumber() + " - " + exc.toString());
					}
					public void fatalError(SAXParseException exc)
							throws SAXException {
						throw new SAXException("Fatal Parser Error: " + exc.getColumnNumber() + "/" + exc.getLineNumber() + " - " + exc.toString());
					}
				});
			} catch (Exception exc) {
				Log.trace("Couldn't set features of DOMParser: " + exc.getMessage());
			}
		}
		return domParser;
	}
	public static DOMParser getNonValidatingDOMParser() {
		if (nonValidatingdomParser == null) {
			nonValidatingdomParser = new DOMParser(new NonValidatingConfiguration());
			try {
				// nonValidatingdomParser.setProperty(Constants.XERCES_PROPERTY_PREFIX + Constants.DOCUMENT_CLASS_NAME_PROPERTY, "org.apache.xerces.dom.DocumentImpl");
				nonValidatingdomParser.setFeature(Constants.SAX_FEATURE_PREFIX + Constants.NAMESPACES_FEATURE, false); // we dont use namespaces
				nonValidatingdomParser.setFeature(Constants.XERCES_FEATURE_PREFIX + Constants.INCLUDE_IGNORABLE_WHITESPACE, false); // we dont need "ignorable whitespace"
				nonValidatingdomParser.setFeature(Constants.SAX_FEATURE_PREFIX + Constants.VALIDATION_FEATURE, false);
				nonValidatingdomParser.setErrorHandler(new ErrorHandler() {
					public void warning(SAXParseException exc)
							throws SAXException {
						throw new SAXException("Parser Warning: " + exc.getColumnNumber() + "/" + exc.getLineNumber() + " - " + exc.toString());
					}
					public void error(SAXParseException exc)
							throws SAXException {
						throw new SAXException("Parser Error: " + exc.getColumnNumber() + "/" + exc.getLineNumber() + " - " + exc.toString());
					}
					public void fatalError(SAXParseException exc)
							throws SAXException {
						throw new SAXException("Fatal Parser Error: " + exc.getColumnNumber() + "/" + exc.getLineNumber() + " - " + exc.toString());
					}
				});
			} catch (Exception exc) {
				Log.trace("Couldn't set features of DOMParser: " + exc.getMessage());
			}
		}
		return nonValidatingdomParser;
	}

	public static SAXTransformerFactory getTransformerFactory() {
		if (transformerFactory == null) {
			transformerFactory = new TransformerFactoryImpl();  // Saxon 8
			transformerFactory.setAttribute(FeatureKeys.TREE_MODEL, new Integer(Builder.TINY_TREE));  //STANDARD_TREE)); //
		}
		return transformerFactory;
	}
	public static XMLReader getNewXMLReader() {
		XMLReader reader = new SAXParser();
		try {
			reader.setFeature("http://xml.org/sax/features/namespaces", false);
			reader.setFeature("http://xml.org/sax/features/validation", true);
		} catch (Exception exc) { }
		return reader;
	}

	/**
	 * Replaces characters that may be confused by a XML
	 * parser with their equivalent character entity references.
	 *
	 * @param s String to be escaped
	 * @return escaped String or null if s = null
	 */
	public static String escapeCharacters(String s) {
		if (s == null) return null;
		char[] chars = s.toCharArray();
		int length = chars.length;
		StringBuffer sb = new StringBuffer(length);
		for (int i = 0; i < length; i++) {
			char c = chars[i];
			switch (c) {
				case '\"':
					sb.append("&quot;");
					break;
				case '\'':
					sb.append("&apos;");
					break;
				case '&':
					sb.append("&amp;");
					break;
				case '<':
					sb.append("&lt;");
					break;
				case '>':
					sb.append("&gt;");
					break;
				default:
					if (c > '\u007f') {
						sb.append("&#");
						sb.append(Integer.toString(c));
						sb.append(';');
					} else
						sb.append(c);
			}
		}
		return sb.toString();
	}

	private static final Pattern ltPattern = Pattern.compile("&lt;");
	private static final Pattern quotPattern = Pattern.compile("&quot;");
	private static final Pattern aposPattern = Pattern.compile("&apos;");
	private static final Pattern gtPattern = Pattern.compile("&gt;");
	private static final Pattern ampPattern = Pattern.compile("&amp;");
	public static String unescapeCharacters(String s) {
		if (s == null) return null;
		s = ltPattern.matcher(s).replaceAll("<");
		s = gtPattern.matcher(s).replaceAll(">");
		s = quotPattern.matcher(s).replaceAll("\"");
		s = aposPattern.matcher(s).replaceAll("'");
		s = ampPattern.matcher(s).replaceAll("&");
		return s;
	}

	public static String serializeNode(Node node, boolean indent) {
		return serializeNode(node, indent, false);
	}
	public static String serializeNode(Node node, boolean indent, boolean writeLineNumbers) {
		if (node == null) return "";
		StringWriter sw = new StringWriter();
		try {
			serializeNode(node, sw, indent, writeLineNumbers);
			return sw.toString();
		} catch (IOException exc) {
			return "Couldn't serialize node!";
		}
	}

	public static void serializeNode(Node node, Writer writer, boolean indent) throws IOException {
		serializeNode(node, writer, indent, false);
	}
	public static void serializeNode(Node node, Writer writer, boolean indent, boolean writeLineNumbers) throws IOException {
		if (writeLineNumbers)
			writer = new LineNumberWriter(writer);

		if (serializer == null) {
			format = new OutputFormat(Method.XML, "UTF-8", true);
			format.setMediaType("text/xml");
			serializer = new XMLSerializer(format);
		}
		serializer.setOutputCharStream(writer);
		format.setIndenting(indent);
		format.setLineWidth(0);
		if (node instanceof Element) {
			format.setOmitDocumentType(true);
			format.setOmitXMLDeclaration(true);
			serializer.serialize((Element) node);
		} else if (node instanceof Document) {
			Document doc = (Document) node;
		// format.setDoctype(OutputFormat.whichDoctypePublic(doc), OutputFormat.whichDoctypeSystem(doc));
			format.setOmitXMLDeclaration(false);
			format.setOmitDocumentType(false);
			serializer.serialize(doc);
		}
	}

	public static String transformDOM2String(Document docNode) { // used in XMLView
		if (docNode == null) return "";
		return transform2HTMLSource(new DOMSource(docNode, ""));
	}

	public static String transformChildren(Element element, String title) { // used in MacrosView
		if (element == null) return "";
		StringBuffer result = new StringBuffer(HTML_HEADER + title + "<dl class=\"root\">");
		NodeList nl = element.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++)
			result.append(transform2HTMLSource(new DOMSource(nl.item(i), "")));
		result.append("</dl></body></html>");
		return result.toString();
	}

	public static String transformHTML2HTMLSource(Reader in) { // used in HTMLView
		if (in == null) return HTML_ERROR_MSG;
		if (htmlParser == null)
			try {
				htmlParser = new org.cyberneko.html.parsers.SAXParser();
				htmlParser.setProperty("http://cyberneko.org/html/properties/names/elems", "lower");
				htmlParser.setProperty("http://cyberneko.org/html/properties/names/attrs", "lower");
			} catch (Throwable t) {
				Log.trace("nekohtml.jar not found ... using default SAX-Parser instead.");
				htmlParser = getNewXMLReader();
			}
		return transform2HTMLSource(new SAXSource(htmlParser, new InputSource(in)));
	}

	public static String transform2HTMLSource(Source source) {
		try {
			if (xml2htmlTransformer == null) {
				StreamSource xslSource = new StreamSource(
						new BufferedReader(new InputStreamReader(XMLUtils.class.getResourceAsStream("xml2html.xsl"))), "");
				xml2htmlTransformer = getTransformerFactory().newTransformer(xslSource);
			}
			StringWriter sw = new StringWriter();
			xml2htmlTransformer.transform(source, new StreamResult(sw));
			return sw.toString();
		} catch (Throwable t) {
			Log.error("Couldn't transform source code to HTML: " + t);
			t.printStackTrace(System.err);
			return HTML_ERROR_MSG;
		}
	}

	public static String transformDTD2Src(URL dtdurl) {
		try {
			StringWriter sw = new StringWriter();
			transformDTD2Source(dtdurl, sw);
			return sw.toString();
		} catch (Throwable t) {
			Log.error("Couldn't transform DTD source code to HTML: " + t);
			t.printStackTrace(System.err);
			return HTML_ERROR_MSG;
		}
	}
	public static void transformDTD2Source(URL dtdurl, Writer writer) throws TransformerException, IOException {
		if (dtdToHTMLSrcTransformer == null) {
			StreamSource xslSource = new StreamSource(new BufferedReader(new InputStreamReader(XMLUtils.class.getResourceAsStream("dtd2html-src.xsl"))), "");
			dtdToHTMLSrcTransformer = getTransformerFactory().newTransformer(xslSource);
		}
		transformDTD(dtdurl, writer, dtdToHTMLSrcTransformer);
	}

	public static String transformDTD2Report(URL dtdurl) {
		try {
			StringWriter sw = new StringWriter();
			transformDTD2Report(dtdurl, sw);
			return sw.toString();
		} catch (Throwable t) {
			Log.error("Couldn't transform DTD source code to HTML: " + t);
			t.printStackTrace(System.err);
			return HTML_ERROR_MSG;
		}
	}

	public static void transformDTD2Report(URL dtdurl, Writer writer) throws TransformerException, IOException {
		if (dtdToHTMLDocTransformer == null) {
			StreamSource xslSource = new StreamSource(
					new BufferedReader(new InputStreamReader(XMLUtils.class.getResourceAsStream("dtd2html-doc.xsl"))), "");
			dtdToHTMLDocTransformer = getTransformerFactory().newTransformer(xslSource);
		}
		transformDTD(dtdurl, writer, dtdToHTMLDocTransformer);
	}

	public static String transformDOM2String(Document documentNode, File xslFile) throws IOException, TransformerException {
		BufferedReader reader = new BufferedReader(new FileReader(xslFile));
		Transformer transformer = getTransformerFactory().newTransformer(new StreamSource(reader));
		StringWriter sw = new StringWriter();
		String documentURI = ((org.apache.xerces.dom.CoreDocumentImpl)documentNode).getDocumentURI(); // documentNode.getDocumentURI(); // DOM 3 - not available in Java1.4
		transformer.transform(new DOMSource(documentNode, documentURI), new StreamResult(sw));
		reader.close();
		return sw.toString();
	}

	public static void transformDTD(URL dtdURL, Writer writer, Transformer transformer) throws TransformerException, IOException {
		BufferedReader reader = new BufferedReader(new InputStreamReader(dtdURL.openStream()));
		InputSource inputSource = new InputSource(reader);
		inputSource.setSystemId(dtdURL.toExternalForm());
		XMLReader xmlReader = new org.cyberneko.dtd.parsers.SAXParser();
		Source xmlSource = new SAXSource(xmlReader, inputSource);
		Result xmlResult = new StreamResult(writer);
		transformer.transform(xmlSource, xmlResult);
		reader.close();
		writer.close();
	}

	public static void transformDOM2File(Document documentNode, File xslFile, File targetFile) throws IOException, TransformerException {
		BufferedReader reader = new BufferedReader(new FileReader(xslFile));
		Transformer transformer = getTransformerFactory().newTransformer(new StreamSource(reader));
		BufferedWriter writer = new BufferedWriter(new FileWriter(targetFile));
		transformer.transform(new DOMSource(documentNode, ""), new StreamResult(writer));
		reader.close();
		writer.close();
	}

		public static void transform(Reader xmlFile, Reader xslFile, Writer targetFile) throws IOException, TransformerException {
			  BufferedReader xslReader = new BufferedReader(xslFile);
		Transformer transformer = getTransformerFactory().newTransformer(new StreamSource(xslReader));
		BufferedWriter writer = new BufferedWriter(targetFile);
		BufferedReader xmlReader = new BufferedReader(xmlFile);
		transformer.transform(new StreamSource(xmlReader), new StreamResult(writer));
		xmlReader.close();
		xslReader.close();
		writer.close();
	}

	public static void transform(final File xmlFile, File xslFile, File targetFile)
			throws IOException, TransformerException, SAXException, ParserConfigurationException {
		BufferedReader xslReader = new BufferedReader(new FileReader(xslFile));
		final Transformer transformer = getTransformerFactory().newTransformer(new StreamSource(xslReader));
		BufferedWriter writer = new BufferedWriter(new FileWriter(targetFile));
		SAXParserFactory spf = SAXParserFactory.newInstance();
		spf.setValidating(false);
		final String dtSysId[] = new String[1];
	//	Reader xmlFileReader = new BufferedReader(new FileReader(xmlFile));
		Reader xmlFileReader = new DoctypeChangerReader(new BufferedReader(new FileReader(xmlFile)), new DoctypeGenerator() {
			  public Doctype generate(Doctype doctype) {
				  dtSysId[0] = doctype.getSystemId();
			//	  transformer.setOutputProperty(OutputKeys.DOCTYPE_SYSTEM, dtSysId[0]);
				  return doctype;
			  }
		  });
		transformer.setOutputProperty(SaxonOutputKeys.INDENT_SPACES, "2"); // TODO fix indentation
		XMLReader xmlReader = spf.newSAXParser().getXMLReader();
		xmlReader.setEntityResolver(new EntityResolver() {
			public InputSource resolveEntity(String publicId, String systemId) {
				return new InputSource(FileUtils.toAbsoluteFilePath(dtSysId[0], FileUtils.getAbsoluteFile(xmlFile)));
			}
		});
		Source source = new SAXSource(xmlReader, new InputSource(xmlFileReader));
		transformer.transform(source, new StreamResult(writer));
		xmlFileReader.close();
		xslReader.close();
		writer.close();
	}

	public static final class FOPUtils {
		private static PreviewDialog FO_PREVIEWDIALOG = null;
		private static AWTRenderer FO_AWT_RENDERER = null;
		private static Translator FO_TRANSLATOR = null;
		private static final String FO_TRANSLATION_PATH = "/org/apache/fop/viewer/resources/";

		public static void transformDOM2FO(Document documentNode, File xslFile, File foFile) throws IOException, TransformerException {
			Writer out = new BufferedWriter(new FileWriter(foFile));
			Reader in = null;
			try {
				in = new BufferedReader(new FileReader(xslFile));
				Transformer transformer = getTransformerFactory().newTransformer(new StreamSource(in));

				//Setup input for XSLT transformation
				Source src = new DOMSource(documentNode, "");

				//Resulting SAX events (the generated FO) must be piped through to FOP
				Result res = new StreamResult(out);

				//Start XSLT transformation and FOP processing
				transformer.transform(src, res);
			} finally {
				out.close();
				if (in != null) in.close();
			}
		}

		public static void transformFO2PDF(File foFile, File pdfFile) throws IOException, FOPException {
			OutputStream out = new BufferedOutputStream(new FileOutputStream(pdfFile));
			Reader in = null;
			try {
				in = new BufferedReader(new FileReader(foFile));
				Driver driver = new Driver(new InputSource(in), out);
				//Setup logger
				Logger logger = new ConsoleLogger(ConsoleLogger.LEVEL_INFO);
				driver.setLogger(logger);
				MessageHandler.setScreenLogger(logger);

				//Setup Renderer (output format)
				driver.setRenderer(Driver.RENDER_PDF);
				//Process FO
				driver.run();
			} finally {
				if (in != null) in.close();
				out.close();
			}
		}

		public static void transformDOM2PDF(Document documentNode, File xslFile, File pdfFile) throws IOException, TransformerException {
			//Construct driver
			Driver driver = new Driver();

			//Setup logger
			Logger logger = new ConsoleLogger(ConsoleLogger.LEVEL_INFO);
			driver.setLogger(logger);
			MessageHandler.setScreenLogger(logger);

			//Setup Renderer (output format)
			driver.setRenderer(Driver.RENDER_PDF);

			//Setup output
			OutputStream out = new BufferedOutputStream(new FileOutputStream(pdfFile));
			Reader in = null;
			try {
				driver.setOutputStream(out);

				//Setup XSLT
				in = new BufferedReader(new FileReader(xslFile));
				Transformer transformer = getTransformerFactory().newTransformer(new StreamSource(in));

				//Setup input for XSLT transformation
				Source src = new DOMSource(documentNode, "");

				//Resulting SAX events (the generated FO) must be piped through to FOP
				Result res = new SAXResult(driver.getContentHandler());

				//Start XSLT transformation and FOP processing
				transformer.transform(src, res);
			} finally {
				if (in != null) in.close();
				out.close();
			}
		}

		private static SecureResourceBundle getResourceBundle(String path) throws IOException {
			URL url = FO_TRANSLATION_PATH.getClass().getResource(path);
			if (url == null) {
				// if the given resource file not found, the english resource uses as default
				path = path.substring(0, path.lastIndexOf(".")) + ".en";
				url = FO_TRANSLATION_PATH.getClass().getResource(path);
			}
			return new SecureResourceBundle(url.openStream());
		}

		public static JFrame getFOPreviewDialog() throws IOException {
			if (FO_PREVIEWDIALOG == null) {
				//Setup l18n
				String language = System.getProperty("user.language");
				FO_TRANSLATOR = getResourceBundle(FO_TRANSLATION_PATH + "resources." + language);
				FO_TRANSLATOR.setMissingEmphasized(false);
				UserMessage.setTranslator(getResourceBundle(FO_TRANSLATION_PATH + "messages." + language));

				//Setup renderer
				FO_AWT_RENDERER = new AWTRenderer(FO_TRANSLATOR); // translator is not used in AWTRenderer
				FO_PREVIEWDIALOG = new PreviewDialog(FO_AWT_RENDERER, FO_TRANSLATOR);
				FO_PREVIEWDIALOG.validate();
				FO_AWT_RENDERER.setProgressListener(FO_PREVIEWDIALOG);
				MessageHandler.addListener(FO_PREVIEWDIALOG);
				FO_AWT_RENDERER.setComponent(FO_PREVIEWDIALOG);
			}
			return FO_PREVIEWDIALOG;
		}

		public static void renderFO2PreviewDialog(File foFile) throws IOException, FOPException {
			getFOPreviewDialog();

			//Setup Driver
			Driver driver = new Driver();
			driver.setLogger(new ConsoleLogger(ConsoleLogger.LEVEL_INFO));
			driver.setRenderer(FO_AWT_RENDERER);

			Reader in = null;
			try {
				// build FO tree: time
				FO_PREVIEWDIALOG.progress(FO_TRANSLATOR.getString("Build FO tree") + " ...");

				//Load XSL-FO file (you can also do an XSL transformation here)
				TransformerFactory factory = TransformerFactory.newInstance();
				Transformer transformer = factory.newTransformer();
				in = new BufferedReader(new FileReader(foFile));
				Source src = new StreamSource(in);
				Result res = new SAXResult(driver.getContentHandler());
				transformer.transform(src, res);

				//Show page
				FO_PREVIEWDIALOG.progress(FO_TRANSLATOR.getString("Show"));
				FO_PREVIEWDIALOG.showPage();
			} catch (Exception e) {
				FO_PREVIEWDIALOG.reportException(e);
				if (e instanceof FOPException)
					throw (FOPException) e;
				throw new FOPException(e);
			} finally {
				if (in != null) in.close();
			}
		}
	}

	private static final String HTML_HEADER = "<html><head><style type = \"text/css\">" +
											  "dl { margin-left:20px; margin-top:-3px; margin-bottom:-3px; }\n" +
											  "dl.root { margin-left:5px; margin-top:-3px; margin-bottom:-3px; }\n dt { white-space:nowrap; }\n" +
											  ".element { color:#800000; font-weight:bold }\n .attr_name { color:#900000 }\n" +
											  ".attr_value { color:#800080; font-style:italic }\n .text { color:#000000 }\n" +
											  ".comment { color:#808080 }\n .pi { color:#008080 }\n .symbol { color:#0000FF }" +
											  "</style></head><body>";

	/** removes empty (trimmed) TEXT-Nodes */
	public static void normalize(Element element) {
		element.normalize();
		for (Node child = element.getFirstChild(); child != null; ) {
			Node next = child.getNextSibling();
			if (child.getNodeType() == Node.TEXT_NODE) {
				Text textNode = (Text)child;
				String text = textNode.getData().trim();
				if (text.length() == 0)
					element.removeChild(child);
			}
			child = next;
		}
	}

	public static void insertNodeAt(Node parent, Node child, int pos) {
		NodeList childNodes = parent.getChildNodes();
		if (childNodes.getLength() == pos)
			parent.appendChild(child);
		else {
			Node beforeXMLNode;
			if (pos == 0) beforeXMLNode = parent.getFirstChild();
			else beforeXMLNode = childNodes.item(pos);
			parent.insertBefore(child, beforeXMLNode);
		}
	}

	public static int getIndexOfChildNode(Node child) {
		if (child == null) return -1;
		Node parent = child.getParentNode();
		if (parent == null) return -1;
		int index = 0;
		Node node = parent.getFirstChild();
		while (node != null) {
			if (node == child) return index;
			node = node.getNextSibling();
			index++;
		}
		return -1;
	}

	public static boolean isDocumentElement(Node xmlNode) {
		Node parentNode = xmlNode.getParentNode();
		return ((parentNode != null) && (xmlNode.getNodeType() == Node.ELEMENT_NODE) && (parentNode.getNodeType() == Node.DOCUMENT_NODE));
	// return (xmlNode.getOwnerDocument().getDocumentElement() == xmlNode);
	}

	// simple method that does not consider information from DTD
	public static boolean isRemoveable(Node xmlNode) {
		if (xmlNode == null) return false;
		switch (xmlNode.getNodeType()) {
			case Node.ELEMENT_NODE:
				return (!isDocumentElement(xmlNode));
			case Node.TEXT_NODE:
			case Node.CDATA_SECTION_NODE:
			case Node.COMMENT_NODE:
			case Node.PROCESSING_INSTRUCTION_NODE:
				return true;
		}
		return false;
	}

	public static boolean isBranchNode(Node node) {
		if (node == null) return false;
		int nodeType = node.getNodeType();
		return ((nodeType == Node.DOCUMENT_NODE) || (nodeType == Node.ELEMENT_NODE));
	}

	public static String getShortDescriptionForNode(Node node) {
		String shortDescription = "???";
		switch (node.getNodeType()) {
			case Node.DOCUMENT_NODE:
				shortDescription = "DOCUMENT node";
				break;
			case Node.ELEMENT_NODE:
				shortDescription = "ELEMENT '" + node.getNodeName() + "'";
				break;
			case Node.TEXT_NODE:
				shortDescription = "TEXT node";
				break;
			case Node.CDATA_SECTION_NODE:
				shortDescription = "CDATA node";
				break;
			case Node.COMMENT_NODE:
				shortDescription = "COMMENT node";
				break;
			case Node.PROCESSING_INSTRUCTION_NODE:
				shortDescription = "PROCESSING-INSTRUCTION node";
				break;
			case Node.DOCUMENT_TYPE_NODE:
				shortDescription = "DOCTYPE node";
				break;
		}
		return shortDescription;
	}

	/**
	 * Returns unsorted set of Attr of <code>element</code>
	 * @param element
	 * @return unsorted set of attributes
	 */
	public static Set getAttributes(Element element) {
		Set result = new HashSet(); // change to TreeSet if ordered set is needed
		NamedNodeMap attributes = element.getAttributes();
		for (int i = 0, size = attributes.getLength(); (i < size); i++)
			result.add(attributes.item(i));
		return result;
	}
	/**
	 * Returns unsorted set of attribute names of <code>element</code>
	 *
	 * @param element
	 * @return unsorted set of attribute names
	 */
	public static Set getAttributeNames(Element element) {
		Set result = new HashSet(); // change to TreeSet if ordered set is needed
		NamedNodeMap attributes = element.getAttributes();
		for (int i = 0, size = attributes.getLength(); (i < size); i++)
			result.add(attributes.item(i).getNodeName());
		return result;
	}

	public static Element stringToElement(String xmlString) {
		try {
			DOMParser parser = getNonValidatingDOMParser();
			parser.parse(new InputSource(new StringReader(xmlString)));
			return parser.getDocument().getDocumentElement();
		} catch (Exception exc) {
			return null;
		}
	}

	/**
	 * Generic (model-independent) method to determine the relative position of two
	 * node in document order. The nodes must be in the same tree.
	 *
	 * @param first  The first node
	 * @param second The second node, whose position is to be compared with the first node
	 * @return -1 if this node precedes the other node, +1 if it follows the other
	 *         node, or 0 if they are the same node. (In this case, isSameNode() will always
	 *         return true, and the two nodes will produce the same result for generateId())
	 */
	public static int compareNodes(Node first, Node second) {
		if (first.getOwnerDocument() != second.getOwnerDocument())
			throw new IllegalArgumentException("Nodes must be from same document!");

		Node ow = second;

		// are they the same node?
		if (first == second)
			return 0;
		// are they siblings (common case)
		if (first.getParentNode() == second.getParentNode())
			return getIndexOfChildNode(first) - getIndexOfChildNode(second);
		// find the depths of both nodes in the tree

		int depth1 = 0;
		int depth2 = 0;
		Node p1 = first;
		Node p2 = second;
		while (p1 != null) {
			depth1++;
			p1 = p1.getParentNode();
		}
		while (p2 != null) {
			depth2++;
			p2 = p2.getParentNode();
		}
		// move up one branch of the tree so we have two nodes on the same level

		p1 = first;
		while (depth1 > depth2) {
			p1 = p1.getParentNode();
			if (p1 == second)
				return +1;
			depth1--;
		}

		p2 = ow;
		while (depth2 > depth1) {
			p2 = p2.getParentNode();
			if (p2 == first)
				return -1;
			depth2--;
		}

		// now move up both branches in sync until we find a common parent
		while (true) {
			Node par1 = p1.getParentNode();
			Node par2 = p2.getParentNode();
			if (par1 == null || par2 == null)
				return -1;
			// throw new NullPointerException("DOM tree compare - internal error");
			if (par1 == par2)
				return getIndexOfChildNode(p1) - getIndexOfChildNode(p2);
			p1 = par1;
			p2 = par2;
		}
	}

	// XPath Methods
	public static String getXPathForNode(Node currentXMLNode) {
		String xpath = "";
		short nodeType = currentXMLNode.getNodeType();
		if (nodeType == Node.DOCUMENT_NODE) return "/";
		if (nodeType == Node.DOCUMENT_TYPE_NODE) return "";
		while (nodeType != Node.DOCUMENT_NODE) {
			String currentXPath;
			if (nodeType == Node.ELEMENT_NODE)
				currentXPath = currentXMLNode.getNodeName(); // if element node -> xpath = node-name
			else
				currentXPath = XPATH_NODE_NAMES[nodeType];
			Node parentNode = currentXMLNode.getParentNode();
			try {
				List xmlNodes = new XPathEvaluator(new DOMSource(parentNode)).evaluate(currentXPath);
			//	if (xmlNodes.size() > 1)
				// also include [1] when it's the only child of its parent, but not if it's the root element to be Saxon compliant
				if (currentXMLNode.getParentNode().getNodeType() != Node.DOCUMENT_NODE)
					currentXPath = currentXPath + "[" + (xmlNodes.indexOf(currentXMLNode) + 1) + "]";
			} catch (XPathException exc) {  }  // should not happen

			xpath = "/" + currentXPath + xpath;
			currentXMLNode = parentNode;
			nodeType = currentXMLNode.getNodeType();
		}
		return xpath;
	}

	public static String evaluateXPathAsString(String xpathExpression, Node contextNode) {
		try {
			Object result = new XPathEvaluator(new DOMSource(contextNode)).evaluateSingle(xpathExpression);
			if (result == null) return null;
			if (result instanceof Element)
				return ((Element)result).getNodeName();
			if (result instanceof Node)
				return ((Node)result).getNodeValue().trim();
			return result.toString().trim();
		} catch (Throwable exc) {
			return null;
			// throw new RuntimeException(exc);
		}
	}

	public static Object evaluateXPathSingleSilently(String xpathExpression, Node contextNode) {
		try {
			return new XPathEvaluator(new DOMSource(contextNode)).evaluateSingle(xpathExpression);
		} catch (Throwable exc) {
			return null;
		}
	}

	private static XPathEvaluator XPATH_EVALUATOR;
	public static void checkXPathExpression(String xpath) throws XPathException {
		if (XPATH_EVALUATOR == null) {
			XPATH_EVALUATOR = new XPathEvaluator(new DOMSource(createNewDocument()));
			// hack to enable also XSLT-Functions, especially for "current()"
			net.sf.saxon.functions.FunctionLibraryList fl = (net.sf.saxon.functions.FunctionLibraryList) XPATH_EVALUATOR.getStaticContext().getFunctionLibrary();
			fl.libraryList.set(0, new net.sf.saxon.functions.SystemFunctionLibrary(net.sf.saxon.functions.SystemFunctionLibrary.FULL_XSLT));
		}
		// XPathReaderFactory.createReader().parse(xpathExpr);
		XPATH_EVALUATOR.createExpression(xpath);
	}

	public static boolean isValidXPathExpression(String xpath) {
		try {
			checkXPathExpression(xpath);
			return true;
		} catch (XPathException e) {
			return false;
		}
	}

	public static boolean matchNodeWithXPath(Node node, String xpathExpr) {
		return matchNodeWithXPath(node, xpathExpr, node.getOwnerDocument());
	}
	public static boolean matchNodeWithXPath(Node node, String xpathExpr, Node contextNode) {
		if (node == null) return false;
		return evaluateXPathSilently(xpathExpr, contextNode).contains(node);
	}

	public static List<String> evaluateXPathSilentlyAsStringList(String xpathExpr, Node contextNode) {
		List list = evaluateXPathSilently(xpathExpr, contextNode);
		return convertXPathResultListToStringList(list);
	}
	public static List<String> convertXPathResultListToStringList(List list) {
		List<String> result = new LinkedList<String>();
		convertXPathResultListToCollection(list, result);
		return result;
	}
	public static void convertXPathResultListToCollection(List list, Collection result) {
		for (Object item : list) {
			String s;
			if (item instanceof Node)
				s = ((Node) item).getNodeValue();
			else if (item instanceof NodeInfo)
				s = ((NodeInfo) item).getStringValue();
			else
				s = item.toString();
		//	s = s.trim();
		//	if (s.length() > 0)
				result.add(s);
		}
	}

	public static List evaluateXPathSilently(String xpathExpr, Node contextNode) {
		try {
			return evaluateXPath(xpathExpr, contextNode);
		} catch (Exception exc) {
			System.err.println("error evaluating "+ xpathExpr);
		//	exc.printStackTrace();
			return new LinkedList();
		}
	}

	public static List evaluateXPath(String xpathExpr, Node contextNode) throws XPathException {
		XPathEvaluator xPathEvaluator = new XPathEvaluator(new DOMSource(contextNode));

		// hack to enable also XSLT-Functions, especially for "current()"
		net.sf.saxon.functions.FunctionLibraryList fl = (net.sf.saxon.functions.FunctionLibraryList) xPathEvaluator.getStaticContext().getFunctionLibrary();
		fl.libraryList.set(0, new net.sf.saxon.functions.SystemFunctionLibrary(net.sf.saxon.functions.SystemFunctionLibrary.FULL_XSLT));//xPathEvaluator.getStaticContext() .getConfiguration(), true));//

		return xPathEvaluator.evaluate(xpathExpr);
	}

/*	public static Set evaluateXPathAsSet(String xpathExpr, Node contextNode) throws XPathException {
	   Source source = new DOMSource(contextNode);
	   Configuration config = new Configuration();
	   NodeInfo _contextNode = Builder.build(source, null, config);
	   StandaloneContext staticContext = new StandaloneContext(config);
	   Expression exp = ExpressionTool.make(xpathExpr, staticContext, 0, -1, 1);
	   exp = exp.analyze(staticContext, Type.ITEM_TYPE);
	   SlotManager map = staticContext.getConfiguration().makeSlotManager();
	   ExpressionTool.allocateSlots(exp, 0, map);
	   XPathContextMajor context = new XPathContextMajor(_contextNode, staticContext.getConfiguration());
	   context.openStackFrame(map);
	   SequenceIterator iterator = exp.iterate(context);

	   // hack to enable also XSLT-Functions, especially for "current()"
	   net.sf.saxon.functions.FunctionLibraryList fl = (net.sf.saxon.functions.FunctionLibraryList) staticContext.getFunctionLibrary();
	   fl.libraryList.set(0, new net.sf.saxon.functions.SystemFunctionLibrary(net.sf.saxon.functions.SystemFunctionLibrary.FULL_XSLT));//config, true));

	   Set set = new HashSet();
	   Item item = iterator.next();
	   while (item != null) {
		   set.add(Value.convert(item));
		   item = iterator.next();
	   }
	   return set;
   }*/

	public static final class NodeComparator implements Comparator {
		public static final Comparator SHARED_INSTANCE = new NodeComparator();
		public int compare(Object o1, Object o2) {
			if (!((o1 instanceof Node) && (o2 instanceof Node)))
				throw new IllegalArgumentException("Arguments must be DOM Nodes!");
			return compareNodes((Node)o1, (Node)o2);
		}
	}

	public static final class LineNumberWriter extends Writer {
		private final Writer writer;
		private int lineNumber;
		private final NumberFormat format;

		public LineNumberWriter(Writer writer) {
			this(writer, null);
		}

		public LineNumberWriter(Writer writer, String formatPattern) {
			this.writer	= writer;
			this.format	= new DecimalFormat(formatPattern == null ?	"000: "	: formatPattern);
			lineNumber	= 0;
		}

		public void close() throws IOException {
			writer.close();
		}

		public void flush() throws IOException {
			writer.flush();
		}

		public void write(char cbuf[], int off, int len) throws IOException {
			StringBuffer line = new StringBuffer();
			line.append(cbuf, off, len);
			boolean inserted = false;
			if (lineNumber == 0) {
				line.insert(0, format.format(++lineNumber));
				inserted = true;
			}
			int pos=-1;
			while ((pos = line.indexOf("\n", pos+1)) != -1) {
				line.insert(pos + 1, format.format(++lineNumber));
				inserted = true;
			}

			if (inserted)
				line.getChars(0, line.length(), cbuf, 0);

			writer.write(cbuf, off, line.length());
		}
	}
}
