package at.ac.tuwien.asgaard.util.xml.dtd;

import at.ac.tuwien.asgaard.util.xml.XMLUtils;
import com.ibm.xml.parser.AttDef;
import com.ibm.xml.parser.DTD;
import org.w3c.dom.*;

import java.util.*;

/**
 * Class DTDHelper. ... todo
 *
 * @version 2.0-Beta
 * @author Peter Votruba
 * @since JDK 1.3.1
 */
public abstract class DTDHelper {
	protected Element targetElement;
	protected int state;

	/** the DTD {will not be | was not} loaded. */
	public static final int NO_DTD = 0;
	/** the DTD {will be | was} loaded, but no checks on insertion will be done. */
	public static final int NO_CHECK = 1;
	/** the DTD {will be | was} loaded and checks on insertion will be done, but the document need not be valid AFTER insertion. */
	public static final int CHECK = 2;
	/** the DTD {will be | was} loaded and strict checks on insertion will be done. */
	public static final int STRICT_CHECK = 3;

	// returned by getAttributeDefaultType()
    public static final int FIXED    = AttDef.FIXED;
    public static final int REQUIRED = AttDef.REQUIRED;
    public static final int IMPLIED  = AttDef.IMPLIED;
    public static final int ENUM     = AttDef.NOFIXED;
    public static final int UNKNOWN  = AttDef.UNKNOWN;

	// returned by getAttributeValueType()
    public static final int CDATA            = AttDef.CDATA;
    public static final int ID               = AttDef.ID;
    public static final int IDREF            = AttDef.IDREF;
    public static final int IDREFS           = AttDef.IDREFS;
    public static final int ENTITY           = AttDef.ENTITY;
    public static final int ENTITIES         = AttDef.ENTITIES;
    public static final int NMTOKEN          = AttDef.NMTOKEN;
    public static final int NMTOKENS         = AttDef.NMTOKENS;
    public static final int NOTATION         = AttDef.NOTATION;
    public static final int ENUMERATION   = AttDef.NAME_TOKEN_GROUP;

	public static final String[] ATTRIBUTE_VALUE_TYPES = AttDef.S_TYPESTR;
	public static final String SPECIAL_NODE_NAME_COMMENT = "#COMMENT";
	public static final String SPECIAL_NODE_NAME_TEXT = DTD.CM_PCDATA;

    public static final int INSERT_MODE_INTO    = 0;
    public static final int INSERT_MODE_BEFORE  = 1;
    public static final int INSERT_MODE_AFTER   = 2;
    public static final int INSERT_MODE_REPLACE = 3;
	protected int insertMode = INSERT_MODE_INTO;

	protected Node targetNode, parentNode;

	public Element getTargetElement() { return targetElement; }

	public abstract void setTargetElement(Node targetNode);

	public final boolean isElementInitialized() {
		return (targetNode != null); // don't use targetElement, because using INSERT_MODE_BEFORE, INSERT_MODE_AFTER and INSERT_MODE_REPLACE also non-element nodes can be target
	}

	public abstract void setState(int state);

	public final int getState() {
		return state;
	}

	public int getInsertMode() { return insertMode; }

	public void setInsertMode(int insertMode) {
		this.insertMode = insertMode;
	}

	protected static String getNodeName(Node node) {
		String nodeName = null;
		switch (node.getNodeType()) {
			case Node.COMMENT_NODE:
				nodeName = SPECIAL_NODE_NAME_COMMENT;
				break;
			case Node.TEXT_NODE:
				nodeName = SPECIAL_NODE_NAME_TEXT;
				break;
			case Node.ELEMENT_NODE:
				nodeName = node.getNodeName();
				break;
		}
		return nodeName;
	}

	public final int getLastInsertIntoIndex(Node newNode) {
		return getLastInsertIntoIndex(getNodeName(newNode));
	}

	public abstract int getLastInsertIntoIndex(String newNodeName);

	public int getInsertBeforeIndex(Node newNode) {
		return getInsertBeforeIndex(getNodeName(newNode));
	}

	public abstract int getInsertBeforeIndex(String newNodeName);

	public int getInsertAfterIndex(Node newNode) {
		return getInsertAfterIndex(getNodeName(newNode));
	}

	public abstract int getInsertAfterIndex(String newNodeName);

	public boolean canBeReplacedByNode(Node newNode) {
		if (targetNode == newNode) return false;
		return canBeReplacedByNode(getNodeName(newNode));
	}

	public abstract boolean canBeReplacedByNode(String newNodeName);

	public int getInsertIndexForNode(Node newNode) {
		switch (insertMode) {
			case INSERT_MODE_INTO:
				return getLastInsertIntoIndex(newNode);
			case INSERT_MODE_BEFORE:
				return getInsertBeforeIndex(newNode);
			case INSERT_MODE_AFTER:
				return getInsertAfterIndex(newNode);
			case INSERT_MODE_REPLACE:
				return (canBeReplacedByNode(newNode) ? XMLUtils.getIndexOfChildNode(targetNode) : -1);
		}
		return -1;
	}

	/**
	 * Tries each InsertMode...
	 * @param newNode
	 * @return
	 */
	public boolean isNodeInsertable(Node newNode) {
		if (getLastInsertIntoIndex(newNode) != -1) return true;
		if (getInsertBeforeIndex(newNode) != -1) return true;
		if (getInsertAfterIndex (newNode) != -1) return true;
		return canBeReplacedByNode(newNode);
	}
	public boolean isNodeInsertable(String newNodeName) {
		if (getLastInsertIntoIndex(newNodeName) != -1) return true;
		if (getInsertBeforeIndex(newNodeName) != -1) return true;
		if (getInsertAfterIndex (newNodeName) != -1) return true;
		return canBeReplacedByNode(newNodeName);
	}
	public int getInsertModeForNode(String newNodeName) {
		if (getLastInsertIntoIndex(newNodeName) != -1) return INSERT_MODE_INTO;
		if (getInsertBeforeIndex(newNodeName) != -1) return INSERT_MODE_BEFORE;
		if (getInsertAfterIndex(newNodeName) != -1) return INSERT_MODE_AFTER;
		return canBeReplacedByNode(newNodeName) ? INSERT_MODE_REPLACE : -1;
	}
	public boolean isMacroInsertable(Element macroElement) {
		if (getLastInsertIntoIndexForMacro(macroElement) != -1) return true;
		if (getInsertAfterIndexForMacro(macroElement) != -1) return true;
		if (getInsertBeforeIndexForMacro(macroElement) != -1) return true;
		return canBeReplacedByMacro(macroElement);
	}


	public boolean isNodeInsertableBefore(Node newNode) {
		return (getInsertBeforeIndex(newNode) > -1);
	}

	public boolean isNodeInsertableBefore(String newNodeName) {
		return (getInsertBeforeIndex(newNodeName) > -1);
	}

	public boolean isNodeInsertableAfter(Node newNode) {
		return (getInsertAfterIndex(newNode) > -1);
	}

	public boolean isNodeInsertableAfter(String newNodeName) {
		return (getInsertAfterIndex(newNodeName) > -1);
	}

	public boolean isNodeInsertableInto(Node newNode) {
		return (getLastInsertIntoIndex(newNode) != -1);
	}
	public boolean isNodeInsertableInto(String newNodeName) {
		return (getLastInsertIntoIndex(newNodeName) != -1);
	}

	public abstract boolean isNodeInsertableAt(Node newNode, int index);
	public abstract boolean isNodeInsertableAt(String newNodeName, int index);
	public boolean isNodeDuplicatable() {
		return (getInsertAfterIndex(targetNode) ==  XMLUtils.getIndexOfChildNode(targetNode) + 1);
	}

	public abstract Set getNewAttributes();
	public abstract Set getAllAttributesForElementSet(String elementName);
	public String[] getAllAttributesForElement(String elementName) {
		Set set = getAllAttributesForElementSet(elementName);
		String[] result = new String[set.size()];
		set.toArray(result);
		return result;
	}

	public abstract List getInsertablePositions(Node newNode);

	public abstract List getInsertablePositions(String newNodeName);

	public final boolean acceptsAttributes(Element element, Collection attributes) {
		return acceptsAttributes(element.getNodeName(), attributes);
	}

	public abstract boolean acceptsAttributes(String elemName, Collection attributes);

	public final int getAttributeDefaultType(Attr attribute) {
		return getAttributeDefaultType(attribute.getOwnerElement().getNodeName(), attribute.getName());
	}

	public final int getAttributeDefaultType(Element element, String attrName) {
		return getAttributeDefaultType(element.getNodeName(), attrName);
	}

	public abstract int getAttributeDefaultType(String elemName, String attrName);

	public final int getAttributeValueType(Attr attribute) {
		return getAttributeValueType(attribute.getOwnerElement().getNodeName(), attribute.getName());
	}

	public final int getAttributeValueType(Element element, String attrName) {
		return getAttributeValueType(element.getNodeName(), attrName);
	}

	public abstract int getAttributeValueType(String elemName, String attrName);

	public final String[] getEnumValues(Attr attribute) {
		return getEnumValues(attribute.getOwnerElement().getNodeName(), attribute.getName());
	}

	public final String[] getEnumValues(Element element, String attrName) {
		return getEnumValues(element.getNodeName(), attrName);
	}

	public abstract String[] getEnumValues(String elemName, String attrName);

	public final int getEnumValuesSize(Attr attribute) {
		return getEnumValuesSize(attribute.getOwnerElement().getNodeName(), attribute.getName());
	}

	public final int getEnumValuesSize(Element element, String attrName) {
		return getEnumValuesSize(element.getNodeName(), attrName);
	}

	public abstract int getEnumValuesSize(String elemName, String attrName);

	public final String getAttributeDefaultValue(Attr attribute) {
		return getAttributeDefaultValue(attribute.getOwnerElement().getNodeName(), attribute.getName());
	}

	public final String getAttributeDefaultValue(Element element, String attrName) {
		return getAttributeDefaultValue(element.getNodeName(), attrName);
	}

	public abstract String getAttributeDefaultValue(String elemName, String attrName);

	public final boolean isAttributeRemovable(Attr attribute) {
		return isAttributeRemovable(attribute.getOwnerElement().getNodeName(), attribute.getName(),
									attribute.getValue());
	}

	public final boolean isAttributeRemovable(Element element, String attrName) {
		return isAttributeRemovable(element.getNodeName(), attrName, element.getAttribute(attrName));
	}

	public final boolean isAttributeRemovable(String elemName, String attrName) {
		return isAttributeRemovable(elemName, attrName, "");
	}

	public abstract boolean isAttributeRemovable(String elemName, String attrName, String attrValue);

	public abstract Node validate(Node node);

	public abstract Node validateDeep(Node node);

	public boolean isValid(Node node) {
		return (validate(node) == null);
	}

	public boolean isValidDeep(Node node) {
		return (validateDeep(node) == null);
	}

	public abstract List getAllInvalidNodes(Node node);

	public abstract Set getAllElementsSet();
	public String[] getAllElements() {
		Set set = getAllElementsSet();
		String[] result = new String[set.size()];
		set.toArray(result);
		return result;
	}

    public abstract SortedSet getInsertableIntoElementSet();
	public abstract SortedSet getInsertableBeforeElementSet();
	public abstract SortedSet getInsertableAfterElementSet();
	public abstract SortedSet getInsertableReplaceElementSet();
	public abstract SortedSet getInsertableConvertElementSet();
	public abstract SortedSet getInsertableWrapElementSet();

	public String[] getInsertableIntoElements() {
		Set set = getInsertableIntoElementSet();
		String[] result = new String[set.size()];
		set.toArray(result);
		return result;
	}
	public String[] getInsertableBeforeElements() {
		Set set = getInsertableBeforeElementSet();
		String[] result = new String[set.size()];
		set.toArray(result);
		return result;
	}
	public String[] getInsertableAfterElements() {
		Set set = getInsertableAfterElementSet();
		String[] result = new String[set.size()];
		set.toArray(result);
		return result;
	}
	public String[] getInsertableReplaceElements() {
		Set set = getInsertableReplaceElementSet();
		String[] result = new String[set.size()];
		set.toArray(result);
		return result;
	}


	public abstract Element createMinimalValidElement(String newElementName, String defaultAttributeValue);

	public abstract void addRequiredAttributes(Element element, String defaultAttributeValue);

	public abstract boolean addRequiredContent(Element currentElement, String defaultAttributeValue);

	public int getInsertIndexForMacro(Element macroElement) {
		switch (insertMode) {
			case INSERT_MODE_INTO:
				return getLastInsertIntoIndexForMacro(macroElement);
			case INSERT_MODE_BEFORE:
				return getInsertBeforeIndexForMacro(macroElement);
			case INSERT_MODE_AFTER:
				return getInsertAfterIndexForMacro(macroElement);
			case INSERT_MODE_REPLACE:
				return (canBeReplacedByMacro(macroElement) ? XMLUtils.getIndexOfChildNode(targetNode) : -1);
		}
		return -1;
	}

	public abstract int getLastInsertIntoIndexForMacro(Element macroElement);

	public abstract int getInsertBeforeIndexForMacro(Element macroElement);

	public abstract int getInsertAfterIndexForMacro(Element macroElement);

	public abstract boolean canBeReplacedByMacro(Element macroElement);

	public abstract boolean isNodeRemoveable();

	public abstract boolean isLoaded();

	public abstract boolean canBeConvertedToElement(String newElementName);

	public abstract boolean canBeWrappedByElement(String newElementName);


// public abstract void addRequiredContentRecursive(Element element);
	public abstract Set getRequiredAttributeNames(String elemName);

	public abstract String guessRootElement();

    public abstract void removeIdentifier(String id);  // todo use it

    public abstract void addIdentifier(String id, Element element); // todo

	public boolean isValidAttribute(Attr attr, String newAttrValue) {
		return isValidAttribute(attr.getOwnerElement().getNodeName(), attr.getName(), newAttrValue);
	}
	public boolean isValidAttribute(Element element, String attrName, String newAttrValue) {
		return isValidAttribute(element.getNodeName(), attrName, newAttrValue);
	}
	public boolean isValidAttribute(String elementName, String attrName, String newAttrValue) {
        return isValidAttribute(elementName, attrName, newAttrValue, null);
    }
	public abstract boolean isValidAttribute(String elementName, String attrName, String newAttrValue, String oldAttrValue);

    public String checkAttribute(Attr attr, String newAttrValue) {
        return checkAttribute(attr.getOwnerElement().getNodeName(), attr.getName(), newAttrValue, attr.getValue());
    }
    public String checkAttribute(Element element, String attrName, String newAttrValue) {
        return checkAttribute(element.getNodeName(), attrName, newAttrValue, element.getAttribute(attrName));
    }
    public String checkAttribute(String elementName, String attrName, String newAttrValue) {
        return checkAttribute(elementName, attrName, newAttrValue, null);
    }
    public abstract String checkAttribute(String elementName, String attrName, String newAttrValue, String oldAttrValue);
}

